/*
 * Copyright (c) 2014, 2015 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.Flushable;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.PrintWriter;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.nio.file.AccessDeniedException;
import java.nio.file.Files;
import java.nio.file.InvalidPathException;
import java.nio.file.NoSuchFileException;
import java.nio.file.NotDirectoryException;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.NoSuchElementException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

import org.imca_cat.prgm.Program;
import org.imca_cat.prgm.ProgramArgumentsParseException;
import org.imca_cat.prgm.ProgramDieException;
import org.imca_cat.prgm.ProgramEnvironment;
import org.imca_cat.prgm.ProgramExits;
import org.imca_cat.procrun.ProcessRunner;
import org.imca_cat.procrun.ProcessRunnerResult;

/**
 * The <tt>aap</tt> program.
 */
class Aap extends Program {
  private static final String NAME = "aap";
  private static final Charset CHARSET = StandardCharsets.UTF_8;

  private final AaprocSystem sys;
  private final AtomicBoolean verbose;
  private DatasetWatchService watcher;
  private DatasetProcessService processor;
  private ExecutorService watcherToProcessorPipe;

  public Aap(String[] args) {
    this(ProgramEnvironment.args(args, CHARSET), AaprocSystem.getDefault());
  }

  public Aap(ProgramEnvironment pe, AaprocSystem sys) {
    super(NAME, pe);
    this.sys = sys;
    verbose = new AtomicBoolean(false);
  }

  protected boolean verbose() {
    return verbose.get();
  }

  protected void verbose(boolean newVerbose) {
    verbose.set(newVerbose);
  }

  @Override
  public Integer basicCall() {
    AapArguments a = parseArgs();
    handleArgs(a);
    AapConfiguration c = parseConf(a);
    handleConf(c);
    try {
      watcher = newDatasetWatchService(c);
      watcher.start();
      Path cwd = toPath(ProgramExits.SOFTWARE, sys.userCurrentWorkingDirectory()).toAbsolutePath();
      List<Path> dirs;
      List<String> dirsAsStrings = a.directories();
      if (dirsAsStrings.isEmpty()) {
        dirs = Arrays.asList(cwd);
      } else {
        dirs = new ArrayList<>(dirsAsStrings.size());
        for (String each : dirsAsStrings) {
          Path d = toPath(ProgramExits.USAGE, each).toAbsolutePath();
          if (Files.isDirectory(d)) {
            dirs.add(d);
          } else {
            printErrorln(each + ": not a directory");
          }
        }
      }
      List<Path> resultsRoots = Collections.emptyList();
      if (c.perTreeResultsEnabled()) {
        resultsRoots = new ArrayList<>(dirs.size());
        for (Path each : dirs) {
          if (!isSameFile(ProgramExits.IOERR, each, cwd)) {
            resultsRoots.add(each);
          }
        }
      }
      for (Path each : dirs) {
        try {
          watcher.register(each);
        } catch (NotDirectoryException e) {
          printErrorln(each + ": not a directory");
        }
      }
      processor = newDatasetProcessService(c);
      processor.start();
      watcherToProcessorPipe = Executors.newSingleThreadExecutor();
      watcherToProcessorPipe.execute(newProcessServiceEnqueuer(watcher, processor));
      List<DatasetProcessResult> results = new ArrayList<>();
      for (;;) {
        DatasetProcessResult r = processor.take();
        removeResult(results, r.dataset());
        results.add(r);
        writeResults(results, cwd, resultsRoots);
      }
    } catch (InterruptedException e) {
      Thread.currentThread().interrupt();
      throw new ProgramDieException(ProgramExits.OK);
    } finally {
      Utilities.shutdownNowSilently(watcherToProcessorPipe);
      Utilities.awaitTerminationSilently(watcherToProcessorPipe);
      Utilities.closeSilently(processor);
      Utilities.closeSilently(watcher);
    }
  }

  private AapArguments parseArgs() {
    AapArguments result;
    try {
      result = AapArgumentsParser.parse(pe().args());
    } catch (ProgramArgumentsParseException e) {
      printErrorln(uncapitalize(e.getMessage()));
      printUsage();
      throw new ProgramDieException(ProgramExits.USAGE);
    }
    return result;
  }

  private void handleArgs(AapArguments a) {
    stackTraceOnExitEnabled(a.hasStackTraceOption());
    verbose(a.hasVerboseOption());
    if (a.hasVersionOption()) {
      printVersion();
      throw new ProgramDieException(ProgramExits.OK);
    }
  }

  private void printUsage() {
    PrintWriter w = pe().err();
    synchronized (printLock()) {
      AapArgumentsParser.printUsage(w, NAME);
      if (w.checkError()) {
        throw new ProgramDieException(ProgramExits.SOFTWARE);
      }
    }
  }

  private void printVersion() {
    PrintWriter w = pe().out();
    synchronized (printLock()) {
      w.println(format("%s %s", NAME, sys.version()));
      if (w.checkError()) {
        throw new ProgramDieException(ProgramExits.SOFTWARE);
      }
    }
  }

  private AapConfiguration parseConf(AapArguments a) {
    AapConfiguration cliConf = a.toConfiguration();
    AapConfiguration otherConf = null;
    AapConfiguration userConf = null;
    AapConfiguration sysConf = null;
    if (a.hasConfFile()) {
      otherConf = parseConfFile(toPath(ProgramExits.USAGE, a.confFile()));
    } else {
      Path userConfFile = findUserConfFile();
      if (userConfFile != null) {
        userConf = parseConfFile(userConfFile);
      }
      Path sysConfFile = toPath(sys.sysConfFilePath());
      if (Files.exists(sysConfFile)) {
        sysConf = parseConfFile(sysConfFile);
      }
    }
    LinkedList<AapConfiguration> confs = new LinkedList<>();
    confs.add(cliConf);
    if (otherConf != null) {
      confs.add(otherConf);
    } else {
      if (userConf != null) confs.add(userConf);
      if (sysConf != null) confs.add(sysConf);
    }
    Collections.reverse(confs);
    AapConfiguration result = confs.removeFirst();
    for (AapConfiguration each : confs) {
      result = each.defaults(result);
    }
    return result;
  }

  private void handleConf(AapConfiguration c) {
    if (verbose()) {
      printConf(c);
    }
  }

  private Path findUserConfFile() {
    if (pe().envVars().containsKey("AAPROC_CONFIG")) {
      return toPath(pe().envVars().get("AAPROC_CONFIG"));
    }
    Path result = toPath(sys.userHomeDirectory(), ".aaproc.conf");
    if (Files.exists(result)) return result;
    if (pe().envVars().containsKey("XDG_CONFIG_HOME")) {
      result = toPath(pe().envVars().get("XDG_CONFIG_HOME"), "aaproc", "config");
      if (Files.exists(result)) return result;
    } else {
      result = toPath(sys.userHomeDirectory(), ".config", "aaproc", "config");
      if (Files.exists(result)) return result;
    }
    return null;
  }

  private Path toPath(String first, String... more) {
    return toPath(ProgramExits.NOINPUT, first, more);
  }

  private Path toPath(int exitStatus, String first, String... more) {
    try {
      return pe().fs().getPath(first, more);
    } catch (InvalidPathException e) {
      String pathString = JavaUtilities.join(pe().fs().getSeparator(), first, more);
      printErrorln(format("%s: invalid path", pathString));
      throw new ProgramDieException(exitStatus);
    }
  }

  private boolean isSameFile(int exitStatus, Path p1, Path p2) {
    try {
      return Files.isSameFile(p1, p2);
    } catch (IOException e) {
      String msg = uncapitalize(e.getMessage());
      if (msg == null || msg.trim().isEmpty()) {
        msg = "I/O error";
      }
      printErrorln(format("%s,%s: %s", p1, p2, msg));
      throw new ProgramDieException(exitStatus);
    }
  }

  private AapConfiguration parseConfFile(Path p) {
    try {
      return AapConfigurationParser.parse(p);
    } catch (ConfigurationException e) {
      printErrorln(format("%s: %s", p, uncapitalize(e.getMessage())));
      throw new ProgramDieException(ProgramExits.DATAERR);
    } catch (AccessDeniedException e) {
      printErrorln(format("%s: access denied", p));
      throw new ProgramDieException(ProgramExits.NOINPUT);
    } catch (FileNotFoundException | NoSuchFileException e) {
      printErrorln(format("%s: file not found", p));
      throw new ProgramDieException(ProgramExits.NOINPUT);
    } catch (IOException e) {
      String msg = uncapitalize(e.getMessage());
      if (msg == null || msg.trim().isEmpty()) {
        msg = "I/O error";
      }
      printErrorln(format("%s: %s", p, msg));
      throw new ProgramDieException(ProgramExits.NOINPUT);
    }
  }

  private void printConf(AapConfiguration c) {
    BufferedWriter w = null;
    try {
      w = newBufferedWriter(pe().outStream(), CHARSET);
      w.write("Configuration:");
      w.newLine();
      (new AapConfigurationWriter(c, w)).write();
    } catch (IOException e) {
      throw new ProgramDieException(ProgramExits.SOFTWARE);
    } finally {
      /*
       * We intentionally do not close the writer since that would close the
       * program environment's output stream. Our only option is to flush it
       * and hope that it does not hold any system resources. It is likely that
       * it does not hold any system resources, but the documentation does not
       * state this.
       */
      flushSilently(w);
    }
  }

  private DatasetWatchService newDatasetWatchService(AapConfiguration c) {
    try {
      DatasetWatchService result = new DatasetWatchService(c, pe().fs(),
          newDatasetWatchServiceExceptionHandler(), newDatasetWatchServiceInfoHandler());
      result.addListener(newDatasetWatchServiceListener());
      return result;
    } catch (IOException e) {
      String msg = uncapitalize(e.getMessage());
      if (msg == null || msg.trim().isEmpty()) {
        msg = "I/O error";
      }
      printErrorln("failed to create file system watch service", msg);
      throw new ProgramDieException(ProgramExits.UNAVAILABLE);
    }
  }

  private DatasetWatchServiceExceptionHandler newDatasetWatchServiceExceptionHandler() {
    return new DatasetWatchServiceExceptionHandler() {
      @Override
      public void exception(Throwable t) {
        printErrorln("dataset watch service error", null, t, verbose());
      }
    };
  }

  private DatasetWatchServiceInfoHandler newDatasetWatchServiceInfoHandler() {
    return new DatasetWatchServiceInfoHandler() {
      @Override
      public void pollTime(long t) {
        if (!verbose()) return;
        println(format("Status (poll time: %d ms)", TimeUnit.NANOSECONDS.toMillis(t)));
      }
    };
  }

  private DatasetWatchServiceListener newDatasetWatchServiceListener() {
    return new DatasetWatchServiceListener() {
      @Override
      public void indexStart(Path p) {
        if (!verbose()) return;
        println(format("Indexing %s", p));
      }

      @Override
      public void indexEnd(DirectoryIndexResult r) {
        if (!verbose()) return;
        println(format("Indexing %s %s", r.succeeded() ? "OK" : "FAILED", r.directory()));
      }
    };
  }

  private DatasetProcessService newDatasetProcessService(AapConfiguration c) {
    DatasetProcessService result =
        new DatasetProcessService(c, sys, newDatasetProcessServiceExceptionHandler());
    result.addListener(newDatasetProcessServiceListener());
    return result;
  }

  private DatasetProcessServiceListener newDatasetProcessServiceListener() {
    return new DatasetProcessServiceListener() {
      @Override
      public void processStart(Dataset d, int numDatasetsPending) {
        println(format("Processing (%d pending) %s", numDatasetsPending, d.directory()));
      }

      @Override
      public void processEnd(DatasetProcessResult r) {
        println(
            format("Processing %s %s", r.succeeded() ? "OK" : "FAILED", r.dataset().directory()));
      }

      @Override
      public void postProcessHookStart(Dataset d, ProcessRunner r) {
      }

      @Override
      public void postProcessHookEnd(ProcessRunnerResult r) {
        if (r.hasExitStatus() && r.exitStatus() == 0 && !r.hasException()) return;
        StringBuilder b = new StringBuilder(256);
        b.append("Post-process hook failed (");
        Integer status = r.exitStatus();
        if (status != null) b.append("exit status: ").append(status).append("; ");
        String stderr = r.stderrString();
        if (stderr != null) b.append("error message: \"").append(stderr.trim()).append("\"; ");
        Throwable exception = r.exception();
        if (exception != null) b.append("exception: \"").append(exception).append("\"; ");
        b.delete(b.length() - "; ".length(), b.length());
        b.append(')');
        println(b.toString());
      }
    };
  }

  private DatasetProcessServiceExceptionHandler newDatasetProcessServiceExceptionHandler() {
    return new DatasetProcessServiceExceptionHandler() {
      @Override
      public void exception(Throwable t) {
        printErrorln("dataset process service error", null, t, verbose());
      }
    };
  }

  private Runnable newProcessServiceEnqueuer(final DatasetWatchService s1,
      final DatasetProcessService s2) {
    return new Runnable() {
      @Override
      public void run() {
        for (;;) {
          try {
            DatasetWatchResult r = s1.take();
            if (r.isInvalidTreeRoot()) {
              printErrorln(format("%s: invalid; removing from watch set", r.invalidTreeRoot()));
            } else {
              s2.put(r.dataset());
            }
          } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
            return;
          } catch (ClosedServiceException e) {
            return;
          } catch (ThreadDeath td) {
            throw td;
          } catch (Throwable t) {
            printErrorln("process service enqueuer error", null, t, verbose());
          }
        }
      }
    };
  }

  private void writeResults(List<DatasetProcessResult> results, Path cwd, List<Path> roots) {
    writeResults(results, cwd);
    for (Path each : roots) {
      Iterable<DatasetProcessResult> i = findResults(results, each);
      if (i.iterator().hasNext()) writeResults(i, each);
    }
  }

  private void writeResults(Iterable<DatasetProcessResult> results, Path dest) {
    Path tmpOutput = dest.resolve(toPath("aap-result.html.tmp"));
    Path output = dest.resolve(toPath("aap-result.html"));
    try {
      try {
        (new DatasetProcessResultsHtmlPrinter(results, tmpOutput)).print();
      } catch (IOException e) {
        String msg = uncapitalize(e.getMessage());
        if (msg == null || msg.trim().isEmpty()) {
          msg = "I/O error";
        }
        printErrorln(format("%s: failed to write result", tmpOutput), msg);
        return;
      }
      moveFile(tmpOutput, output);
    } finally {
      Utilities.deleteFileSilently(tmpOutput);
    }
  }

  private void moveFile(Path source, Path target) {
    try {
      Utilities.moveFile(source, target);
    } catch (IOException e) {
      String msg = uncapitalize(e.getMessage());
      if (msg == null || msg.trim().isEmpty()) {
        msg = "I/O error";
      }
      printErrorln(format("failed to move \"%s\" to \"%s\"", source, target), msg);
    }
  }

  private static void removeResult(List<DatasetProcessResult> results, Dataset d) {
    for (Iterator<DatasetProcessResult> i = results.iterator(); i.hasNext();) {
      if (i.next().dataset().equals(d)) {
        i.remove();
        return;
      }
    }
  }

  private static Iterable<DatasetProcessResult> findResults(
      final List<DatasetProcessResult> results, final Path root) {
    return new Iterable<DatasetProcessResult>() {
      @Override
      public Iterator<DatasetProcessResult> iterator() {
        return new Iterator<DatasetProcessResult>() {
          private Iterator<DatasetProcessResult> src = results.iterator();
          private DatasetProcessResult nextElement = null;
          private boolean atEnd = false;

          @Override
          public boolean hasNext() {
            stepOnce();
            return !atEnd;
          }

          @Override
          public DatasetProcessResult next() {
            if (!hasNext()) throw new NoSuchElementException();
            DatasetProcessResult result = nextElement;
            step();
            return result;
          }

          @Override
          public void remove() {
            throw new UnsupportedOperationException();
          }

          private void stepOnce() {
            if (!atEnd && nextElement == null) step();
          }

          private void step() {
            while (src.hasNext()) {
              DatasetProcessResult r = src.next();
              if (!r.dataset().directory().startsWith(root)) continue;
              nextElement = r;
              return;
            }
            atEnd = true;
          }
        };
      }
    };
  }

  private static String format(String format, Object... args) {
    return String.format(Locale.US, format, args);
  }

  private static void flushSilently(Flushable f) {
    if (f == null) return;
    try {
      f.flush();
    } catch (InterruptedIOException e) {
      Thread.currentThread().interrupt();
    } catch (ThreadDeath td) {
      throw td;
    } catch (Throwable t) {
      /* IGNORED */
    }
  }

  public static void main(String[] args) {
    runAndExit(new Aap(args), true);
  }
}
