/*
 * Copyright (c) 2014 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.util.Collections;
import java.util.List;
import java.util.Objects;

/**
 * Command line arguments for the {@link Aap} program.
 */
class AapArguments {
  private final String confFile;
  private final Integer maxNumProcessJobs;
  private final boolean stackTrace;
  private final boolean verbose;
  private final boolean version;
  private final List<String> directories;

  /**
   * @throws IllegalArgumentException if maxNumProcessJobs is less than 1
   */
  public AapArguments(String confFile, Integer maxNumProcessJobs, boolean stackTrace,
      boolean verbose, boolean version, List<String> directories) {
    super();
    this.confFile = confFile;
    if (maxNumProcessJobs != null && maxNumProcessJobs.intValue() < 1) {
      throw new IllegalArgumentException("maxNumProcessJobs less than 1");
    }
    this.maxNumProcessJobs = maxNumProcessJobs;
    this.stackTrace = stackTrace;
    this.verbose = verbose;
    this.version = version;
    this.directories = Collections.unmodifiableList(directories);
  }

  public boolean hasConfFile() {
    return confFile != null;
  }

  public String confFile() {
    return confFile;
  }

  public boolean hasMaxNumProcessJobs() {
    return maxNumProcessJobs != null;
  }

  public Integer maxNumProcessJobs() {
    return maxNumProcessJobs;
  }

  public boolean hasStackTraceOption() {
    return stackTrace;
  }

  public boolean hasVerboseOption() {
    return verbose;
  }

  public boolean hasVersionOption() {
    return version;
  }

  public List<String> directories() {
    return directories;
  }

  public AapConfiguration toConfiguration() {
    return new AapConfiguration(maxNumProcessJobs);
  }

  @Override
  public boolean equals(Object o) {
    if (o == this) return true;
    if (!(o instanceof AapArguments)) return false;
    AapArguments a = (AapArguments)o;
    if (!Objects.equals(confFile, a.confFile)) return false;
    if (!Objects.equals(maxNumProcessJobs, a.maxNumProcessJobs)) return false;
    if (stackTrace != a.stackTrace) return false;
    if (verbose != a.verbose) return false;
    if (version != a.version) return false;
    return directories.equals(a.directories);
  }

  @Override
  public int hashCode() {
    int result = 17;
    result = 31 * result + ((confFile == null) ? 0 : confFile.hashCode());
    result = 31 * result + ((maxNumProcessJobs == null) ? 0 : maxNumProcessJobs.hashCode());
    result = 31 * result + (stackTrace ? 1 : 0);
    result = 31 * result + (verbose ? 1 : 0);
    result = 31 * result + (version ? 1 : 0);
    result = 31 * result + directories.hashCode();
    return result;
  }

  @Override
  public String toString() {
    StringBuilder result = new StringBuilder();
    if (confFile != null) {
      result.append("confFile=");
      result.append(confFile);
    }
    if (maxNumProcessJobs != null) {
      if (result.length() != 0) result.append(", ");
      result.append("maxNumProcessJobs=");
      result.append(maxNumProcessJobs);
    }
    if (stackTrace) {
      if (result.length() != 0) result.append(", ");
      result.append("stackTrace");
    }
    if (verbose) {
      if (result.length() != 0) result.append(", ");
      result.append("verbose");
    }
    if (version) {
      if (result.length() != 0) result.append(", ");
      result.append("version");
    }
    if (!directories.isEmpty()) {
      if (result.length() != 0) result.append(", ");
      result.append("directories=");
      result.append(directories);
    }
    return result.toString();
  }
}
