/*
 * Copyright (c) 2014, 2015 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.List;

import org.imca_cat.prgm.ProgramArgumentsParseException;

import joptsimple.NonOptionArgumentSpec;
import joptsimple.OptionException;
import joptsimple.OptionParser;
import joptsimple.OptionSet;

/**
 * Command line arguments parser for the {@link Aap} program.
 */
class AapArgumentsParser {
  private static final String CONF_FILE = "f";
  private static final String PROCESS_JOBS = "j";
  private static final String STACK_TRACE = "s";
  private static final String VERBOSE = "v";
  private static final String VERSION = "V";

  private static final OptionParser PARSER = newOptionParser();
  private static final NonOptionArgumentSpec<String> PARSER_NON_OPTIONS = PARSER.nonOptions();

  private final List<String> args;

  public AapArgumentsParser(List<String> args) {
    super();
    this.args = args;
  }

  public AapArguments parse() throws ProgramArgumentsParseException {
    try {
      OptionSet options = PARSER.parse(args.toArray(new String[0]));
      String confFile = parseConfFile(options);
      Integer maxProcessJobs = parseMaxProcessJobs(options);
      boolean stackTrace = parseStackTrace(options);
      boolean verbose = parseVerbose(options);
      boolean version = parseVersion(options);
      List<String> directories = parseDirectories(options);
      return new AapArguments(confFile, maxProcessJobs, stackTrace, verbose, version, directories);
    } catch (OptionException e) {
      if (e.getMessage() != null) {
        throw new ProgramArgumentsParseException(e.getMessage(), e);
      } else {
        throw new ProgramArgumentsParseException("Illegal argument", e);
      }
    }
  }

  public static AapArguments parse(List<String> args) throws ProgramArgumentsParseException {
    return (new AapArgumentsParser(args)).parse();
  }

  private static String parseConfFile(OptionSet options) {
    return (String)options.valueOf(CONF_FILE);
  }

  private static Integer parseMaxProcessJobs(OptionSet options)
      throws ProgramArgumentsParseException {
    Integer value = parseInteger((String)options.valueOf(PROCESS_JOBS), PROCESS_JOBS);
    if (value == null) return null;
    StringBuilder directive = new StringBuilder();
    directive.append(AapConfigurationParser.PROP_MAX_NUM_PROCESS_JOBS);
    directive.append(" = ");
    directive.append(value.toString());
    try {
      return AapConfigurationParser.parse(directive.toString()).maxNumProcessJobs();
    } catch (IOException e) {
      throw new ProgramArgumentsParseException(
          "\"-" + PROCESS_JOBS + "\": bad implicit configuration directive conversion", e);
    } catch (ConfigurationException e) {
      throw new ProgramArgumentsParseException(e.getMessage(), e);
    }
  }

  private static boolean parseStackTrace(OptionSet options) {
    return options.has(STACK_TRACE);
  }

  private static boolean parseVerbose(OptionSet options) {
    return options.has(VERBOSE);
  }

  private static boolean parseVersion(OptionSet options) {
    return options.has(VERSION);
  }

  private static List<String> parseDirectories(OptionSet options) {
    return PARSER_NON_OPTIONS.values(options);
  }

  private static Integer parseInteger(String s, String opt) throws ProgramArgumentsParseException {
    if (s == null) return null;
    try {
      return Integer.parseInt(s);
    } catch (NumberFormatException e) {
      throw new ProgramArgumentsParseException("\"-" + opt + "\": \"" + s + "\": not an integer");
    }
  }

  public static void printUsage(PrintWriter w, String programName) {
    w.write("usage: " + programName);
    w.write(" [-" + STACK_TRACE + VERBOSE + VERSION + "]");
    w.write(" [-" + CONF_FILE + " file]");
    w.write(" [-" + PROCESS_JOBS + " max_process_jobs]");
    w.write(" [directory ...]");
    w.println();
  }

  private static OptionParser newOptionParser() {
    OptionParser result = new OptionParser();
    result.accepts(CONF_FILE).withRequiredArg();
    result.accepts(PROCESS_JOBS).withRequiredArg();
    result.accepts(STACK_TRACE);
    result.accepts(VERBOSE);
    result.accepts(VERSION);
    return result;
  }
}
