/*
 * Copyright (c) 2014, 2015, 2016 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.regex.Pattern;

/**
 * Configuration for the {@link Aap} program.
 */
class AapConfiguration {
  private final AapConfiguration defaults;
  private final Integer maxNumProcessJobs;
  private final Integer maxNumIndexJobs;
  private final Integer maxNumRegisterJobs;
  private final Integer maxNumDirChangeJobs;
  private final Boolean fsPollEnabled;
  private final TimeDuration fsPollPeriod;
  private final Integer fsPollMaxNumJobs;
  private final Boolean fsPollMetadataEnabled;
  private final Pattern processingDirPat;
  private final Pattern diffCenterImagePat;
  private final Pattern imagePat;
  private final String processingDirName;
  private final TimeDuration datasetSettleTime;
  private final Integer datasetMinNumImages;
  private final Boolean perTreeResultsEnabled;
  private final TimeDuration maxProcessRunTime;
  private final TimeDuration processSupervisorPeriod;
  private final BigDecimal maxRefinedDetDistDelta;
  private final Integer numThreadsPerProcessJob;
  private final Pattern excludeDirPat;
  private final List<String> postProcessHook;

  public AapConfiguration() {
    this(null);
  }

  /**
   * @throws IllegalArgumentException if {@code maxNumProcessJobs} is less than
   *           1
   */
  public AapConfiguration(Integer maxNumProcessJobs) {
    this(maxNumProcessJobs, null, null, null, null, null, null, null, null, null, null, null, null,
        null, null, null, null, null, null, null, null);
  }

  /**
   * @throws IllegalArgumentException if {@code maxNumProcessJobs},
   *           {@code maxNumIndexJobs}, {@code maxNumRegisterJobs},
   *           {@code maxNumDirChangeJobs}, {@code fsPollMaxNumJobs},
   *           {@code datasetMinNumImages}, or {@code numThreadsPerProcessJob}
   *           is less than 1, if any one of the {@code TimeDuration} instances
   *           is less than 1 ns, if any one of the {@code Pattern} instances
   *           is empty, if {@code processingDirName} is empty, or if
   *           {@code maxRefinedDetDistDelta} is less than 0.001 mm
   */
  public AapConfiguration(Integer maxNumProcessJobs, Integer maxNumIndexJobs,
      Integer maxNumRegisterJobs, Integer maxNumDirChangeJobs, Boolean fsPollEnabled,
      TimeDuration fsPollPeriod, Integer fsPollMaxNumJobs, Boolean fsPollMetadataEnabled,
      Pattern processingDirPat, Pattern diffCenterImagePat, Pattern imagePat,
      String processingDirName, TimeDuration datasetSettleTime, Integer datasetMinNumImages,
      Boolean perTreeResultsEnabled, TimeDuration maxProcessRunTime,
      TimeDuration processSupervisorPeriod, BigDecimal maxRefinedDetDistDelta,
      Integer numThreadsPerProcessJob, Pattern excludeDirPat, List<String> postProcessHook) {
    this(newDefaults(), maxNumProcessJobs, maxNumIndexJobs, maxNumRegisterJobs, maxNumDirChangeJobs,
        fsPollEnabled, fsPollPeriod, fsPollMaxNumJobs, fsPollMetadataEnabled, processingDirPat,
        diffCenterImagePat, imagePat, processingDirName, datasetSettleTime, datasetMinNumImages,
        perTreeResultsEnabled, maxProcessRunTime, processSupervisorPeriod, maxRefinedDetDistDelta,
        numThreadsPerProcessJob, excludeDirPat, postProcessHook);
  }

  /**
   * @throws NullPointerException if defaults is {@code null} and any other
   *           argument is {@code null}
   * @throws IllegalArgumentException if {@code maxNumProcessJobs},
   *           {@code maxNumIndexJobs}, {@code maxNumRegisterJobs},
   *           {@code maxNumDirChangeJobs}, {@code fsPollMaxNumJobs},
   *           {@code datasetMinNumImages}, or {@code numThreadsPerProcessJob}
   *           is less than 1, if any one of the {@code TimeDuration} instances
   *           is less than 1 ns, if any one of the {@code Pattern} instances
   *           is empty, if {@code processingDirName} is empty, or if
   *           {@code maxRefinedDetDistDelta} is less than 0.001 mm
   */
  public AapConfiguration(AapConfiguration defaults, Integer maxNumProcessJobs,
      Integer maxNumIndexJobs, Integer maxNumRegisterJobs, Integer maxNumDirChangeJobs,
      Boolean fsPollEnabled, TimeDuration fsPollPeriod, Integer fsPollMaxNumJobs,
      Boolean fsPollMetadataEnabled, Pattern processingDirPat, Pattern diffCenterImagePat,
      Pattern imagePat, String processingDirName, TimeDuration datasetSettleTime,
      Integer datasetMinNumImages, Boolean perTreeResultsEnabled, TimeDuration maxProcessRunTime,
      TimeDuration processSupervisorPeriod, BigDecimal maxRefinedDetDistDelta,
      Integer numThreadsPerProcessJob, Pattern excludeDirPat, List<String> postProcessHook) {
    super();
    this.defaults = defaults;
    if (defaults == null) {
      this.maxNumProcessJobs = Objects.requireNonNull(maxNumProcessJobs);
      this.maxNumIndexJobs = Objects.requireNonNull(maxNumIndexJobs);
      this.maxNumRegisterJobs = Objects.requireNonNull(maxNumRegisterJobs);
      this.maxNumDirChangeJobs = Objects.requireNonNull(maxNumDirChangeJobs);
      this.fsPollEnabled = Objects.requireNonNull(fsPollEnabled);
      this.fsPollPeriod = Objects.requireNonNull(fsPollPeriod);
      this.fsPollMaxNumJobs = Objects.requireNonNull(fsPollMaxNumJobs);
      this.fsPollMetadataEnabled = Objects.requireNonNull(fsPollMetadataEnabled);
      this.processingDirPat = Objects.requireNonNull(processingDirPat);
      this.diffCenterImagePat = Objects.requireNonNull(diffCenterImagePat);
      this.imagePat = Objects.requireNonNull(imagePat);
      this.processingDirName = Objects.requireNonNull(processingDirName);
      this.datasetSettleTime = Objects.requireNonNull(datasetSettleTime);
      this.datasetMinNumImages = Objects.requireNonNull(datasetMinNumImages);
      this.perTreeResultsEnabled = Objects.requireNonNull(perTreeResultsEnabled);
      this.maxProcessRunTime = Objects.requireNonNull(maxProcessRunTime);
      this.processSupervisorPeriod = Objects.requireNonNull(processSupervisorPeriod);
      this.maxRefinedDetDistDelta = Objects.requireNonNull(maxRefinedDetDistDelta);
      this.numThreadsPerProcessJob = Objects.requireNonNull(numThreadsPerProcessJob);
      this.excludeDirPat = Objects.requireNonNull(excludeDirPat);
      this.postProcessHook =
          Collections.unmodifiableList(new ArrayList<>(Objects.requireNonNull(postProcessHook)));
    } else {
      this.maxNumProcessJobs = maxNumProcessJobs;
      this.maxNumIndexJobs = maxNumIndexJobs;
      this.maxNumRegisterJobs = maxNumRegisterJobs;
      this.maxNumDirChangeJobs = maxNumDirChangeJobs;
      this.fsPollEnabled = fsPollEnabled;
      this.fsPollPeriod = fsPollPeriod;
      this.fsPollMaxNumJobs = fsPollMaxNumJobs;
      this.fsPollMetadataEnabled = fsPollMetadataEnabled;
      this.processingDirPat = processingDirPat;
      this.diffCenterImagePat = diffCenterImagePat;
      this.imagePat = imagePat;
      this.processingDirName = processingDirName;
      this.datasetSettleTime = datasetSettleTime;
      this.datasetMinNumImages = datasetMinNumImages;
      this.perTreeResultsEnabled = perTreeResultsEnabled;
      this.maxProcessRunTime = maxProcessRunTime;
      this.processSupervisorPeriod = processSupervisorPeriod;
      this.maxRefinedDetDistDelta = maxRefinedDetDistDelta;
      this.numThreadsPerProcessJob = numThreadsPerProcessJob;
      this.excludeDirPat = excludeDirPat;
      this.postProcessHook = (postProcessHook == null) ? null
          : Collections.unmodifiableList(new ArrayList<>(postProcessHook));
    }

    /*
     * Define temporary variables for values to be used below in argument
     * validation that represent the current value (either the specified value
     * or the value obtained from the defaults).
     */
    int xMaxNumProcessJobs;
    if (maxNumProcessJobs != null) {
      xMaxNumProcessJobs = maxNumProcessJobs.intValue();
    } else {
      xMaxNumProcessJobs = defaults.maxNumProcessJobs();
    }
    int xMaxNumIndexJobs;
    if (maxNumIndexJobs != null) {
      xMaxNumIndexJobs = maxNumIndexJobs.intValue();
    } else {
      xMaxNumIndexJobs = defaults.maxNumIndexJobs();
    }
    int xMaxNumRegisterJobs;
    if (maxNumRegisterJobs != null) {
      xMaxNumRegisterJobs = maxNumRegisterJobs.intValue();
    } else {
      xMaxNumRegisterJobs = defaults.maxNumRegisterJobs();
    }
    int xMaxNumDirChangeJobs;
    if (maxNumDirChangeJobs != null) {
      xMaxNumDirChangeJobs = maxNumDirChangeJobs.intValue();
    } else {
      xMaxNumDirChangeJobs = defaults.maxNumDirChangeJobs();
    }
    TimeDuration xFsPollPeriod;
    if (fsPollPeriod != null) {
      xFsPollPeriod = fsPollPeriod;
    } else {
      xFsPollPeriod = defaults.fileSystemPollPeriod();
    }
    int xFsPollMaxNumJobs;
    if (fsPollMaxNumJobs != null) {
      xFsPollMaxNumJobs = fsPollMaxNumJobs.intValue();
    } else {
      xFsPollMaxNumJobs = defaults.fileSystemPollMaxNumJobs();
    }
    Pattern xProcessingDirPat;
    if (processingDirPat != null) {
      xProcessingDirPat = processingDirPat;
    } else {
      xProcessingDirPat = defaults.processingDirPat();
    }
    Pattern xDiffCenterImagePat;
    if (diffCenterImagePat != null) {
      xDiffCenterImagePat = diffCenterImagePat;
    } else {
      xDiffCenterImagePat = defaults.diffCenterImagePat();
    }
    Pattern xImagePat;
    if (imagePat != null) {
      xImagePat = imagePat;
    } else {
      xImagePat = defaults.imagePat();
    }
    String xProcessingDirName;
    if (processingDirName != null) {
      xProcessingDirName = processingDirName;
    } else {
      xProcessingDirName = defaults.processingDirName();
    }
    TimeDuration xDatasetSettleTime;
    if (datasetSettleTime != null) {
      xDatasetSettleTime = datasetSettleTime;
    } else {
      xDatasetSettleTime = defaults.datasetSettleTime();
    }
    int xDatasetMinNumImages;
    if (datasetMinNumImages != null) {
      xDatasetMinNumImages = datasetMinNumImages.intValue();
    } else {
      xDatasetMinNumImages = defaults.datasetMinNumImages();
    }
    TimeDuration xMaxProcessRunTime;
    if (maxProcessRunTime != null) {
      xMaxProcessRunTime = maxProcessRunTime;
    } else {
      xMaxProcessRunTime = defaults.maxProcessRunTime();
    }
    TimeDuration xProcessSupervisorPeriod;
    if (processSupervisorPeriod != null) {
      xProcessSupervisorPeriod = processSupervisorPeriod;
    } else {
      xProcessSupervisorPeriod = defaults.processSupervisorPeriod();
    }
    BigDecimal xMaxRefinedDetDistDelta;
    if (maxRefinedDetDistDelta != null) {
      xMaxRefinedDetDistDelta = maxRefinedDetDistDelta;
    } else {
      xMaxRefinedDetDistDelta = defaults.maxRefinedDetDistDelta();
    }
    int xNumThreadsPerProcessJob;
    if (numThreadsPerProcessJob != null) {
      xNumThreadsPerProcessJob = numThreadsPerProcessJob.intValue();
    } else {
      xNumThreadsPerProcessJob = defaults.numThreadsPerProcessJob();
    }

    /* Validate arguments. */
    if (xMaxNumProcessJobs < 1) {
      throw new IllegalArgumentException(AapConfigurationParser.PROP_MAX_NUM_PROCESS_JOBS + ": \"" +
          xMaxNumProcessJobs + "\"': less than 1");
    }
    if (xMaxNumIndexJobs < 1) {
      throw new IllegalArgumentException(AapConfigurationParser.PROP_MAX_NUM_INDEX_JOBS + ": \"" +
          xMaxNumIndexJobs + "\": less than 1");
    }
    if (xMaxNumRegisterJobs < 1) {
      throw new IllegalArgumentException(AapConfigurationParser.PROP_MAX_NUM_REGISTER_JOBS +
          ": \"" + xMaxNumRegisterJobs + "\": less than 1");
    }
    if (xMaxNumDirChangeJobs < 1) {
      throw new IllegalArgumentException(AapConfigurationParser.PROP_MAX_NUM_DIR_CHANGE_JOBS +
          ": \"" + xMaxNumDirChangeJobs + "\": less than 1");
    }
    TimeDuration xMinFsPollPeriod = TimeDuration.nanoseconds(1L);
    if (xFsPollPeriod.compareTo(xMinFsPollPeriod) < 0) {
      throw new IllegalArgumentException(AapConfigurationParser.PROP_FILE_SYSTEM_POLL_PERIOD +
          ": \"" + toString(xFsPollPeriod) + "\": less than \"" + toString(xMinFsPollPeriod));
    }
    if (xFsPollMaxNumJobs < 1) {
      throw new IllegalArgumentException(AapConfigurationParser.PROP_FILE_SYSTEM_POLL_MAX_NUM_JOBS +
          ": \"" + xFsPollMaxNumJobs + "\": less than 1");
    }
    if (xProcessingDirPat.pattern().isEmpty()) {
      throw new IllegalArgumentException(
          AapConfigurationParser.PROP_PROCESSING_DIR_PAT + ": empty");
    }
    if (xDiffCenterImagePat.pattern().isEmpty()) {
      throw new IllegalArgumentException(
          AapConfigurationParser.PROP_DIFF_CENTER_IMAGE_PAT + ": empty");
    }
    if (xImagePat.pattern().isEmpty()) {
      throw new IllegalArgumentException(AapConfigurationParser.PROP_IMAGE_PAT + ": empty");
    }
    if (xProcessingDirName.isEmpty()) {
      throw new IllegalArgumentException(
          AapConfigurationParser.PROP_PROCESSING_DIR_NAME + ": empty");
    }
    TimeDuration xMinDatasetSettleTime = TimeDuration.nanoseconds(1L);
    if (xDatasetSettleTime.compareTo(xMinDatasetSettleTime) < 0) {
      throw new IllegalArgumentException(AapConfigurationParser.PROP_DATASET_SETTLE_TIME + ": \"" +
          toString(xDatasetSettleTime) + "\": less than \"" + toString(xMinDatasetSettleTime));
    }
    if (xDatasetMinNumImages < 1) {
      throw new IllegalArgumentException(AapConfigurationParser.PROP_DATASET_MIN_NUM_IMAGES +
          ": \"" + xDatasetMinNumImages + "\": less than 1");
    }
    TimeDuration xMinMaxProcessRunTime = TimeDuration.nanoseconds(1L);
    if (xMaxProcessRunTime.compareTo(xMinMaxProcessRunTime) < 0) {
      throw new IllegalArgumentException(AapConfigurationParser.PROP_MAX_PROCESS_RUN_TIME + ": \"" +
          toString(xMaxProcessRunTime) + "\": less than \"" + toString(xMinMaxProcessRunTime));
    }
    TimeDuration xMinProcessSupervisorPeriod = TimeDuration.nanoseconds(1L);
    if (xProcessSupervisorPeriod.compareTo(xMinProcessSupervisorPeriod) < 0) {
      throw new IllegalArgumentException(AapConfigurationParser.PROP_PROCESS_SUPERVISOR_PERIOD +
          ": \"" + toString(xProcessSupervisorPeriod) + "\": less than \"" +
          toString(xMinProcessSupervisorPeriod));
    }
    BigDecimal xMinMaxRefinedDetDistDelta = new BigDecimal("0.001");
    if (xMaxRefinedDetDistDelta.compareTo(xMinMaxRefinedDetDistDelta) < 0) {
      throw new IllegalArgumentException(AapConfigurationParser.PROP_MAX_REFINED_DET_DIST_DELTA +
          ": \"" + xMaxRefinedDetDistDelta + "\": less than \"" + xMinMaxRefinedDetDistDelta);
    }
    if (xNumThreadsPerProcessJob < 1) {
      throw new IllegalArgumentException(AapConfigurationParser.PROP_NUM_THREADS_PER_PROCESS_JOB +
          ": \"" + xNumThreadsPerProcessJob + "\"': less than 1");
    }
  }

  public int maxNumProcessJobs() {
    return (maxNumProcessJobs != null) ? maxNumProcessJobs : defaults.maxNumProcessJobs();
  }

  public int maxNumIndexJobs() {
    return (maxNumIndexJobs != null) ? maxNumIndexJobs : defaults.maxNumIndexJobs();
  }

  public int maxNumRegisterJobs() {
    return (maxNumRegisterJobs != null) ? maxNumRegisterJobs : defaults.maxNumRegisterJobs();
  }

  public int maxNumDirChangeJobs() {
    return (maxNumDirChangeJobs != null) ? maxNumDirChangeJobs : defaults.maxNumDirChangeJobs();
  }

  public boolean fileSystemPollEnabled() {
    return (fsPollEnabled != null) ? fsPollEnabled : defaults.fileSystemPollEnabled();
  }

  public TimeDuration fileSystemPollPeriod() {
    return (fsPollPeriod != null) ? fsPollPeriod : defaults.fileSystemPollPeriod();
  }

  public int fileSystemPollMaxNumJobs() {
    return (fsPollMaxNumJobs != null) ? fsPollMaxNumJobs : defaults.fileSystemPollMaxNumJobs();
  }

  public boolean fileSystemPollMetadataEnabled() {
    return (fsPollMetadataEnabled != null) ? fsPollMetadataEnabled
        : defaults.fileSystemPollMetadataEnabled();
  }

  public Pattern processingDirPat() {
    return (processingDirPat != null) ? processingDirPat : defaults.processingDirPat();
  }

  public Pattern diffCenterImagePat() {
    return (diffCenterImagePat != null) ? diffCenterImagePat : defaults.diffCenterImagePat();
  }

  public Pattern imagePat() {
    return (imagePat != null) ? imagePat : defaults.imagePat();
  }

  public String processingDirName() {
    return (processingDirName != null) ? processingDirName : defaults.processingDirName();
  }

  public TimeDuration datasetSettleTime() {
    return (datasetSettleTime != null) ? datasetSettleTime : defaults.datasetSettleTime();
  }

  public int datasetMinNumImages() {
    return (datasetMinNumImages != null) ? datasetMinNumImages : defaults.datasetMinNumImages();
  }

  public boolean perTreeResultsEnabled() {
    return (perTreeResultsEnabled != null) ? perTreeResultsEnabled
        : defaults.perTreeResultsEnabled();
  }

  public TimeDuration maxProcessRunTime() {
    return (maxProcessRunTime != null) ? maxProcessRunTime : defaults.maxProcessRunTime();
  }

  public TimeDuration processSupervisorPeriod() {
    return (processSupervisorPeriod != null) ? processSupervisorPeriod
        : defaults.processSupervisorPeriod();
  }

  public BigDecimal maxRefinedDetDistDelta() {
    return (maxRefinedDetDistDelta != null) ? maxRefinedDetDistDelta
        : defaults.maxRefinedDetDistDelta();
  }

  public int numThreadsPerProcessJob() {
    return (numThreadsPerProcessJob != null) ? numThreadsPerProcessJob
        : defaults.numThreadsPerProcessJob();
  }

  public Pattern excludeDirPat() {
    return (excludeDirPat != null) ? excludeDirPat : defaults.excludeDirPat();
  }

  public List<String> postProcessHook() {
    return (postProcessHook != null) ? postProcessHook : defaults.postProcessHook();
  }

  public AapConfiguration defaults(AapConfiguration c) {
    return new AapConfiguration(c, maxNumProcessJobs, maxNumIndexJobs, maxNumRegisterJobs,
        maxNumDirChangeJobs, fsPollEnabled, fsPollPeriod, fsPollMaxNumJobs, fsPollMetadataEnabled,
        processingDirPat, diffCenterImagePat, imagePat, processingDirName, datasetSettleTime,
        datasetMinNumImages, perTreeResultsEnabled, maxProcessRunTime, processSupervisorPeriod,
        maxRefinedDetDistDelta, numThreadsPerProcessJob, excludeDirPat, postProcessHook);
  }

  @Override
  public boolean equals(Object o) {
    if (o == this) return true;
    if (!(o instanceof AapConfiguration)) return false;
    AapConfiguration c = (AapConfiguration)o;
    if (!Objects.equals(maxNumProcessJobs, c.maxNumProcessJobs)) return false;
    if (!Objects.equals(maxNumIndexJobs, c.maxNumIndexJobs)) return false;
    if (!Objects.equals(maxNumRegisterJobs, c.maxNumRegisterJobs)) return false;
    if (!Objects.equals(maxNumDirChangeJobs, c.maxNumDirChangeJobs)) return false;
    if (!Objects.equals(fsPollEnabled, c.fsPollEnabled)) return false;
    if (!Objects.equals(fsPollPeriod, c.fsPollPeriod)) return false;
    if (!Objects.equals(fsPollMaxNumJobs, c.fsPollMaxNumJobs)) return false;
    if (!Objects.equals(fsPollMetadataEnabled, c.fsPollMetadataEnabled)) return false;
    if (!Objects.equals(processingDirPat, c.processingDirPat)) return false;
    if (!Objects.equals(diffCenterImagePat, c.diffCenterImagePat)) return false;
    if (!Objects.equals(imagePat, c.imagePat)) return false;
    if (!Objects.equals(processingDirName, c.processingDirName)) return false;
    if (!Objects.equals(datasetSettleTime, c.datasetSettleTime)) return false;
    if (!Objects.equals(datasetMinNumImages, c.datasetMinNumImages)) return false;
    if (!Objects.equals(perTreeResultsEnabled, c.perTreeResultsEnabled)) return false;
    if (!Objects.equals(maxProcessRunTime, c.maxProcessRunTime)) return false;
    if (!Objects.equals(processSupervisorPeriod, c.processSupervisorPeriod)) return false;
    if (!Objects.equals(maxRefinedDetDistDelta, c.maxRefinedDetDistDelta)) return false;
    if (!Objects.equals(numThreadsPerProcessJob, c.numThreadsPerProcessJob)) return false;
    if (!Objects.equals(excludeDirPat, c.excludeDirPat)) return false;
    if (!Objects.equals(postProcessHook, c.postProcessHook)) return false;
    return Objects.equals(defaults, c.defaults);
  }

  @Override
  public int hashCode() {
    int result = 17;
    result = 31 * result + ((defaults == null) ? 0 : defaults.hashCode());
    result = 31 * result + ((maxNumProcessJobs == null) ? 0 : maxNumProcessJobs.hashCode());
    result = 31 * result + ((maxNumIndexJobs == null) ? 0 : maxNumIndexJobs.hashCode());
    result = 31 * result + ((maxNumRegisterJobs == null) ? 0 : maxNumRegisterJobs.hashCode());
    result = 31 * result + ((maxNumDirChangeJobs == null) ? 0 : maxNumDirChangeJobs.hashCode());
    result = 31 * result + ((fsPollEnabled == null) ? 0 : fsPollEnabled.hashCode());
    result = 31 * result + ((fsPollPeriod == null) ? 0 : fsPollPeriod.hashCode());
    result = 31 * result + ((fsPollMaxNumJobs == null) ? 0 : fsPollMaxNumJobs.hashCode());
    result = 31 * result + ((fsPollMetadataEnabled == null) ? 0 : fsPollMetadataEnabled.hashCode());
    result = 31 * result + ((processingDirPat == null) ? 0 : processingDirPat.hashCode());
    result = 31 * result + ((diffCenterImagePat == null) ? 0 : diffCenterImagePat.hashCode());
    result = 31 * result + ((imagePat == null) ? 0 : imagePat.hashCode());
    result = 31 * result + ((processingDirName == null) ? 0 : processingDirName.hashCode());
    result = 31 * result + ((datasetSettleTime == null) ? 0 : datasetSettleTime.hashCode());
    result = 31 * result + ((datasetMinNumImages == null) ? 0 : datasetMinNumImages.hashCode());
    result = 31 * result + ((perTreeResultsEnabled == null) ? 0 : perTreeResultsEnabled.hashCode());
    result = 31 * result + ((maxProcessRunTime == null) ? 0 : maxProcessRunTime.hashCode());
    result =
        31 * result + ((processSupervisorPeriod == null) ? 0 : processSupervisorPeriod.hashCode());
    result =
        31 * result + ((maxRefinedDetDistDelta == null) ? 0 : maxRefinedDetDistDelta.hashCode());
    result =
        31 * result + ((numThreadsPerProcessJob == null) ? 0 : numThreadsPerProcessJob.hashCode());
    result = 31 * result + ((excludeDirPat == null) ? 0 : excludeDirPat.hashCode());
    result = 31 * result + ((postProcessHook == null) ? 0 : postProcessHook.hashCode());
    return result;
  }

  @Override
  public String toString() {
    StringBuilder result = new StringBuilder();
    if (maxNumProcessJobs != null) {
      result.append("maxNumProcessJobs=");
      result.append(maxNumProcessJobs);
    }
    if (maxNumIndexJobs != null) {
      if (result.length() > 0) result.append(", ");
      result.append("maxNumIndexJobs=");
      result.append(maxNumIndexJobs);
    }
    if (maxNumRegisterJobs != null) {
      if (result.length() > 0) result.append(", ");
      result.append("maxNumRegisterJobs=");
      result.append(maxNumRegisterJobs);
    }
    if (maxNumDirChangeJobs != null) {
      if (result.length() > 0) result.append(", ");
      result.append("maxNumDirChangeJobs=");
      result.append(maxNumDirChangeJobs);
    }
    if (fsPollEnabled != null) {
      if (result.length() > 0) result.append(", ");
      result.append("fsPollEnabled=");
      result.append(fsPollEnabled);
    }
    if (fsPollPeriod != null) {
      if (result.length() > 0) result.append(", ");
      result.append("fsPollPeriod=");
      result.append(fsPollPeriod);
    }
    if (fsPollMaxNumJobs != null) {
      if (result.length() > 0) result.append(", ");
      result.append("fsPollMaxNumJobs=");
      result.append(fsPollMaxNumJobs);
    }
    if (fsPollMetadataEnabled != null) {
      if (result.length() > 0) result.append(", ");
      result.append("fsPollMetadataEnabled=");
      result.append(fsPollMetadataEnabled);
    }
    if (processingDirPat != null) {
      if (result.length() > 0) result.append(", ");
      result.append("processingDirPat=");
      result.append(processingDirPat);
    }
    if (diffCenterImagePat != null) {
      if (result.length() > 0) result.append(", ");
      result.append("diffCenterImagePat=");
      result.append(diffCenterImagePat);
    }
    if (imagePat != null) {
      if (result.length() > 0) result.append(", ");
      result.append("imagePat=");
      result.append(imagePat);
    }
    if (processingDirName != null) {
      if (result.length() > 0) result.append(", ");
      result.append("processingDirName=");
      result.append(processingDirName);
    }
    if (datasetSettleTime != null) {
      if (result.length() > 0) result.append(", ");
      result.append("datasetSettleTime=");
      result.append(datasetSettleTime);
    }
    if (datasetMinNumImages != null) {
      if (result.length() > 0) result.append(", ");
      result.append("datasetMinNumImages=");
      result.append(datasetMinNumImages);
    }
    if (perTreeResultsEnabled != null) {
      if (result.length() > 0) result.append(", ");
      result.append("perTreeResultsEnabled=");
      result.append(perTreeResultsEnabled);
    }
    if (maxProcessRunTime != null) {
      if (result.length() > 0) result.append(", ");
      result.append("maxProcessRunTime=");
      result.append(maxProcessRunTime);
    }
    if (processSupervisorPeriod != null) {
      if (result.length() > 0) result.append(", ");
      result.append("processSupervisorPeriod=");
      result.append(processSupervisorPeriod);
    }
    if (maxRefinedDetDistDelta != null) {
      if (result.length() > 0) result.append(", ");
      result.append("maxRefinedDetDistDelta=");
      result.append(maxRefinedDetDistDelta);
    }
    if (numThreadsPerProcessJob != null) {
      if (result.length() > 0) result.append(", ");
      result.append("numThreadsPerProcessJob=");
      result.append(numThreadsPerProcessJob);
    }
    if (excludeDirPat != null) {
      if (result.length() > 0) result.append(", ");
      result.append("excludeDirPat=");
      result.append(excludeDirPat);
    }
    if (postProcessHook != null) {
      if (result.length() > 0) result.append(", ");
      result.append("postProcessHook=");
      result.append(postProcessHook);
    }
    if (defaults != null) {
      if (result.length() > 0) result.append(", ");
      result.append("defaults={");
      result.append(defaults);
      result.append("}");
    }
    return result.toString();
  }

  private static AapConfiguration newDefaults() {
    return new AapConfiguration(null, 1, 1, 1, 1, Boolean.FALSE, TimeDuration.seconds(10L), 1,
        Boolean.TRUE, Pattern.compile("(?i)processing"),
        Pattern.compile("(?i)cell_z(x|y)_(fly_)?[0-9_]+\\.(cbf|img)"),
        Pattern.compile("(?i).+\\.(cbf|img)"), "processing", TimeDuration.seconds(30L), 20,
        Boolean.FALSE, TimeDuration.minutes(45), TimeDuration.seconds(60), new BigDecimal("50"), 1,
        Pattern.compile(""), Collections.<String>emptyList());
  }

  private static String toString(TimeDuration d) {
    return AapConfigurationWriter.toString(d);
  }
}
