/*
 * Copyright (c) 2014, 2015, 2016 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.math.BigDecimal;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.Properties;
import java.util.concurrent.TimeUnit;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import org.imca_cat.shellwords.Shellwords;

/**
 * Configuration parser for the {@link Aap} program.
 */
class AapConfigurationParser {
  static final String PROP_MAX_NUM_PROCESS_JOBS = "MaxNumProcessJobs";
  static final String PROP_MAX_NUM_INDEX_JOBS = "MaxNumIndexJobs";
  static final String PROP_MAX_NUM_REGISTER_JOBS = "MaxNumRegisterJobs";
  static final String PROP_MAX_NUM_DIR_CHANGE_JOBS = "MaxNumDirChangeJobs";
  static final String PROP_FILE_SYSTEM_POLL_ENABLED = "FileSystemPoll";
  static final String PROP_FILE_SYSTEM_POLL_PERIOD = "FileSystemPollPeriod";
  static final String PROP_FILE_SYSTEM_POLL_MAX_NUM_JOBS = "FileSystemPollMaxNumJobs";
  static final String PROP_FILE_SYSTEM_POLL_METADATA_ENABLED = "FileSystemPollMetadata";
  static final String PROP_PROCESSING_DIR_PAT = "ProcessingDirectoryPattern";
  static final String PROP_DIFF_CENTER_IMAGE_PAT = "DiffractionCenterImagePattern";
  static final String PROP_IMAGE_PAT = "ImagePattern";
  static final String PROP_PROCESSING_DIR_NAME = "ProcessingDirectoryName";
  static final String PROP_DATASET_SETTLE_TIME = "DatasetSettleTime";
  static final String PROP_DATASET_MIN_NUM_IMAGES = "DatasetMinNumImages";
  static final String PROP_PER_TREE_RESULTS_ENABLED = "PerTreeResults";
  static final String PROP_MAX_PROCESS_RUN_TIME = "MaxProcessRunTime";
  static final String PROP_PROCESS_SUPERVISOR_PERIOD = "ProcessSupervisorPeriod";
  static final String PROP_MAX_REFINED_DET_DIST_DELTA = "MaxRefinedDetDistDelta";
  static final String PROP_NUM_THREADS_PER_PROCESS_JOB = "NumThreadsPerProcessJob";
  static final String PROP_EXCLUDE_DIR_PAT = "ExcludeDirectoryPattern";
  static final String PROP_POST_PROCESS_HOOK = "PostProcessHook";

  static final List<String> ALL_PROPS = Collections.unmodifiableList(
      Arrays.asList(PROP_MAX_NUM_PROCESS_JOBS, PROP_MAX_NUM_INDEX_JOBS, PROP_MAX_NUM_REGISTER_JOBS,
          PROP_MAX_NUM_DIR_CHANGE_JOBS, PROP_FILE_SYSTEM_POLL_ENABLED, PROP_FILE_SYSTEM_POLL_PERIOD,
          PROP_FILE_SYSTEM_POLL_MAX_NUM_JOBS, PROP_FILE_SYSTEM_POLL_METADATA_ENABLED,
          PROP_PROCESSING_DIR_PAT, PROP_DIFF_CENTER_IMAGE_PAT, PROP_IMAGE_PAT,
          PROP_PROCESSING_DIR_NAME, PROP_DATASET_SETTLE_TIME, PROP_DATASET_MIN_NUM_IMAGES,
          PROP_PER_TREE_RESULTS_ENABLED, PROP_MAX_PROCESS_RUN_TIME, PROP_PROCESS_SUPERVISOR_PERIOD,
          PROP_MAX_REFINED_DET_DIST_DELTA, PROP_NUM_THREADS_PER_PROCESS_JOB, PROP_EXCLUDE_DIR_PAT,
          PROP_POST_PROCESS_HOOK));

  private static final Pattern TIME_DURATION = Pattern.compile("^\\s*(?<a>-?\\d+)\\s*(?<u>\\S*)$");

  private final Reader in;

  public AapConfigurationParser(Reader in) {
    super();
    this.in = in;
  }

  public AapConfiguration parse() throws IOException, ConfigurationException {
    return parse(in);
  }

  public static AapConfiguration parse(Path confFile) throws IOException, ConfigurationException {
    try (BufferedReader confReader = newBufferedReader(confFile)) {
      return (new AapConfigurationParser(confReader)).parse();
    }
  }

  public static AapConfiguration parse(String confFileData)
      throws IOException, ConfigurationException {
    try (BufferedReader confReader = newBufferedReader(confFileData)) {
      return (new AapConfigurationParser(confReader)).parse();
    }
  }

  private static AapConfiguration parse(Reader r) throws IOException, ConfigurationException {
    Properties props = new Properties();
    props.load(r);
    return parse(props);
  }

  private static AapConfiguration parse(Properties p) throws ConfigurationException {
    Integer maxNumProcessJobs = parseMaxNumProcessJobs(p);
    Integer maxNumIndexJobs = parseMaxNumIndexJobs(p);
    Integer maxNumRegisterJobs = parseMaxNumRegisterJobs(p);
    Integer maxNumDirChangeJobs = parseMaxNumDirChangeJobs(p);
    Boolean fileSystemPollEnabled = parseFileSystemPollEnabled(p);
    TimeDuration fileSystemPollPeriod = parseFileSystemPollPeriod(p);
    Integer fileSystemPollMaxNumJobs = parseFileSystemPollMaxNumJobs(p);
    Boolean fileSystemPollMetadataEnabled = parseFileSystemPollMetadataEnabled(p);
    Pattern processingDirPat = parseProcessingDirPat(p);
    Pattern diffCenterImagePat = parseDiffCenterImagePat(p);
    Pattern imagePat = parseImagePat(p);
    String processingDirName = parseProcessingDirName(p);
    TimeDuration datasetSettleTime = parseDatasetSettleTime(p);
    Integer datasetMinNumImages = parseDatasetMinNumImages(p);
    Boolean perTreeResultsEnabled = parsePerTreeResultsEnabled(p);
    TimeDuration maxProcessRunTime = parseMaxProcessRunTime(p);
    TimeDuration processSupervisorPeriod = parseProcessSupervisorPeriod(p);
    BigDecimal maxRefinedDetDistDelta = parseMaxRefinedDetDistDelta(p);
    Integer numThreadsPerProcessJob = parseNumThreadsPerProcessJob(p);
    Pattern excludeDirPat = parseExcludeDirPat(p);
    List<String> postProcessHook = parsePostProcessHook(p);
    try {
      return new AapConfiguration(maxNumProcessJobs, maxNumIndexJobs, maxNumRegisterJobs,
          maxNumDirChangeJobs, fileSystemPollEnabled, fileSystemPollPeriod,
          fileSystemPollMaxNumJobs, fileSystemPollMetadataEnabled, processingDirPat,
          diffCenterImagePat, imagePat, processingDirName, datasetSettleTime, datasetMinNumImages,
          perTreeResultsEnabled, maxProcessRunTime, processSupervisorPeriod, maxRefinedDetDistDelta,
          numThreadsPerProcessJob, excludeDirPat, postProcessHook);
    } catch (IllegalArgumentException e) {
      throw new ConfigurationException(e.getMessage());
    }
  }

  private static Integer parseMaxNumProcessJobs(Properties p) throws ConfigurationException {
    String k = PROP_MAX_NUM_PROCESS_JOBS;
    return parseInteger(p.getProperty(k), k);
  }

  private static Integer parseMaxNumIndexJobs(Properties p) throws ConfigurationException {
    String k = PROP_MAX_NUM_INDEX_JOBS;
    return parseInteger(p.getProperty(k), k);
  }

  private static Integer parseMaxNumRegisterJobs(Properties p) throws ConfigurationException {
    String k = PROP_MAX_NUM_REGISTER_JOBS;
    return parseInteger(p.getProperty(k), k);
  }

  private static Integer parseMaxNumDirChangeJobs(Properties p) throws ConfigurationException {
    String k = PROP_MAX_NUM_DIR_CHANGE_JOBS;
    return parseInteger(p.getProperty(k), k);
  }

  private static Boolean parseFileSystemPollEnabled(Properties p) throws ConfigurationException {
    String k = PROP_FILE_SYSTEM_POLL_ENABLED;
    return parseBoolean(p.getProperty(k), k);
  }

  private static TimeDuration parseFileSystemPollPeriod(Properties p)
      throws ConfigurationException {
    String k = PROP_FILE_SYSTEM_POLL_PERIOD;
    return parseTimeDuration(p.getProperty(k), k, TimeUnit.SECONDS);
  }

  private static Integer parseFileSystemPollMaxNumJobs(Properties p) throws ConfigurationException {
    String k = PROP_FILE_SYSTEM_POLL_MAX_NUM_JOBS;
    return parseInteger(p.getProperty(k), k);
  }

  private static Boolean parseFileSystemPollMetadataEnabled(Properties p)
      throws ConfigurationException {
    String k = PROP_FILE_SYSTEM_POLL_METADATA_ENABLED;
    return parseBoolean(p.getProperty(k), k);
  }

  private static Pattern parseProcessingDirPat(Properties p) throws ConfigurationException {
    String k = PROP_PROCESSING_DIR_PAT;
    return parsePattern(p.getProperty(k), k);
  }

  private static Pattern parseDiffCenterImagePat(Properties p) throws ConfigurationException {
    String k = PROP_DIFF_CENTER_IMAGE_PAT;
    return parsePattern(p.getProperty(k), k);
  }

  private static Pattern parseImagePat(Properties p) throws ConfigurationException {
    String k = PROP_IMAGE_PAT;
    return parsePattern(p.getProperty(k), k);
  }

  private static String parseProcessingDirName(Properties p) throws ConfigurationException {
    String k = PROP_PROCESSING_DIR_NAME;
    return parseString(p.getProperty(k), k);
  }

  private static TimeDuration parseDatasetSettleTime(Properties p) throws ConfigurationException {
    String k = PROP_DATASET_SETTLE_TIME;
    return parseTimeDuration(p.getProperty(k), k, TimeUnit.SECONDS);
  }

  private static Integer parseDatasetMinNumImages(Properties p) throws ConfigurationException {
    String k = PROP_DATASET_MIN_NUM_IMAGES;
    return parseInteger(p.getProperty(k), k);
  }

  private static Boolean parsePerTreeResultsEnabled(Properties p) throws ConfigurationException {
    String k = PROP_PER_TREE_RESULTS_ENABLED;
    return parseBoolean(p.getProperty(k), k);
  }

  private static TimeDuration parseMaxProcessRunTime(Properties p) throws ConfigurationException {
    String k = PROP_MAX_PROCESS_RUN_TIME;
    return parseTimeDuration(p.getProperty(k), k, TimeUnit.MINUTES);
  }

  private static TimeDuration parseProcessSupervisorPeriod(Properties p)
      throws ConfigurationException {
    String k = PROP_PROCESS_SUPERVISOR_PERIOD;
    return parseTimeDuration(p.getProperty(k), k, TimeUnit.SECONDS);
  }

  private static BigDecimal parseMaxRefinedDetDistDelta(Properties p)
      throws ConfigurationException {
    String k = PROP_MAX_REFINED_DET_DIST_DELTA;
    return parseBigDecimal(p.getProperty(k), k);
  }

  private static Integer parseNumThreadsPerProcessJob(Properties p) throws ConfigurationException {
    String k = PROP_NUM_THREADS_PER_PROCESS_JOB;
    return parseInteger(p.getProperty(k), k);
  }

  private static Pattern parseExcludeDirPat(Properties p) throws ConfigurationException {
    String k = PROP_EXCLUDE_DIR_PAT;
    return parsePattern(p.getProperty(k), k);
  }

  private static List<String> parsePostProcessHook(Properties p) throws ConfigurationException {
    String k = PROP_POST_PROCESS_HOOK;
    return parseHook(p.getProperty(k), k);
  }

  private static Integer parseInteger(String s, String key) throws ConfigurationException {
    if (s == null) return null;
    try {
      return Integer.parseInt(s);
    } catch (NumberFormatException e) {
      throw new ConfigurationException("\"" + key + "\": \"" + s + "\": not an integer");
    }
  }

  private static BigDecimal parseBigDecimal(String s, String key) throws ConfigurationException {
    if (s == null) return null;
    try {
      return new BigDecimal(s);
    } catch (NumberFormatException e) {
      throw new ConfigurationException("\"" + key + "\": \"" + s + "\": not a decimal number");
    }
  }

  private static Boolean parseBoolean(String s, String key) throws ConfigurationException {
    if (s == null) return null;
    if (s.equalsIgnoreCase("yes")) {
      return Boolean.TRUE;
    } else if (s.equalsIgnoreCase("no")) {
      return Boolean.FALSE;
    } else {
      throw new ConfigurationException("\"" + key + "\": \"" + s + "\": invalid value");
    }
  }

  private static TimeDuration parseTimeDuration(String s, String key, TimeUnit defaultUnit)
      throws ConfigurationException {
    if (s == null) return null;
    Matcher m = TIME_DURATION.matcher(s);
    if (!m.matches()) {
      throw new ConfigurationException("\"" + key + "\": \"" + s + "\": not a time duration");
    }
    String amountString = m.group("a");
    if (amountString == null) {
      throw new ConfigurationException("\"" + key + "\": \"" + s + "\": not a time duration");
    }
    long amount;
    try {
      amount = Long.parseLong(amountString);
    } catch (NumberFormatException e) {
      throw new ConfigurationException("\"" + key + "\": \"" + amountString + "\": not an integer");
    }
    String unitString = m.group("u");
    if (unitString != null && unitString.isEmpty()) {
      unitString = null;
    }
    TimeUnit unit;
    try {
      unit = parseTimeUnit(unitString);
    } catch (IllegalArgumentException e) {
      throw new ConfigurationException(
          "\"" + key + "\": \"" + unitString + "\": illegal time unit");
    }
    if (unit == null) unit = defaultUnit;
    return new TimeDuration(amount, unit);
  }

  private static TimeUnit parseTimeUnit(String s) {
    if (s == null) return null;
    switch (s.trim().toLowerCase(Locale.ENGLISH)) {
      case "d":
        return TimeUnit.DAYS;
      case "h":
        return TimeUnit.HOURS;
      case "us":
        return TimeUnit.MICROSECONDS;
      case "\u00B5s":
        return TimeUnit.MICROSECONDS;
      case "\u03BCs":
        return TimeUnit.MICROSECONDS;
      case "ms":
        return TimeUnit.MILLISECONDS;
      case "m":
        return TimeUnit.MINUTES;
      case "ns":
        return TimeUnit.NANOSECONDS;
      case "s":
        return TimeUnit.SECONDS;
    }
    throw new IllegalArgumentException("\"" + s + "\": illegal time unit");
  }

  private static Pattern parsePattern(String s, String key) throws ConfigurationException {
    if (s == null) return null;
    try {
      return Pattern.compile(s);
    } catch (PatternSyntaxException e) {
      throw new ConfigurationException("\"" + key + "\": \"" + s + "\": invalid regular expression",
          e);
    }
  }

  private static List<String> parseHook(String s, String key) throws ConfigurationException {
    if (s == null) return null;
    try {
      return Shellwords.split(s);
    } catch (IllegalArgumentException e) {
      throw new ConfigurationException("\"" + key + "\": \"" + s + "\": invalid shell line", e);
    }
  }

  private static String parseString(String s, String key) throws ConfigurationException {
    return s;
  }

  private static BufferedReader newBufferedReader(Path p) throws IOException {
    return (p == null) ? null : Files.newBufferedReader(p, StandardCharsets.UTF_8);
  }

  private static BufferedReader newBufferedReader(String s) throws IOException {
    return (s == null) ? null : new BufferedReader(new StringReader(s));
  }
}
