/*
 * Copyright (c) 2014, 2015, 2016 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.Properties;
import java.util.concurrent.TimeUnit;
import java.util.regex.Pattern;

import org.imca_cat.shellwords.Shellwords;

/**
 * Configuration writer for the {@link Aap} program capable of writing an
 * {@link AapConfiguration} to a stream in the text form suitable for parsing
 * by {@link AapConfigurationParser}.
 */
class AapConfigurationWriter {
  private final AapConfiguration conf;
  private final BufferedWriter out;

  public AapConfigurationWriter(AapConfiguration conf, BufferedWriter out) {
    super();
    this.conf = conf;
    this.out = out;
  }

  public void write() throws IOException {
    Properties p = newProperties(conf);
    StringWriter w = null;
    StringReader r = null;
    BufferedReader br = null;
    try {
      w = new StringWriter();
      p.store(w, null);
      r = new StringReader(w.toString());
      br = new BufferedReader(r);
      List<String> lines = new ArrayList<>();
      for (String each = br.readLine(); each != null; each = br.readLine()) {
        if (each.startsWith("#")) continue;
        lines.add(each);
      }
      Collections.sort(lines);
      for (String each : lines) {
        out.append(each);
        out.newLine();
      }
    } finally {
      Utilities.closeSilently(br);
      Utilities.closeSilently(r);
      Utilities.closeSilently(w);
    }
  }

  private static final Properties newProperties(AapConfiguration c) {
    Properties p = new Properties();
    put(p, AapConfigurationParser.PROP_MAX_NUM_PROCESS_JOBS, c.maxNumProcessJobs());
    put(p, AapConfigurationParser.PROP_MAX_NUM_INDEX_JOBS, c.maxNumIndexJobs());
    put(p, AapConfigurationParser.PROP_MAX_NUM_REGISTER_JOBS, c.maxNumRegisterJobs());
    put(p, AapConfigurationParser.PROP_MAX_NUM_DIR_CHANGE_JOBS, c.maxNumDirChangeJobs());
    put(p, AapConfigurationParser.PROP_FILE_SYSTEM_POLL_ENABLED, c.fileSystemPollEnabled());
    put(p, AapConfigurationParser.PROP_FILE_SYSTEM_POLL_PERIOD, c.fileSystemPollPeriod());
    put(p, AapConfigurationParser.PROP_FILE_SYSTEM_POLL_MAX_NUM_JOBS, c.fileSystemPollMaxNumJobs());
    put(p, AapConfigurationParser.PROP_FILE_SYSTEM_POLL_METADATA_ENABLED,
        c.fileSystemPollMetadataEnabled());
    put(p, AapConfigurationParser.PROP_PROCESSING_DIR_PAT, c.processingDirPat());
    put(p, AapConfigurationParser.PROP_DIFF_CENTER_IMAGE_PAT, c.diffCenterImagePat());
    put(p, AapConfigurationParser.PROP_IMAGE_PAT, c.imagePat());
    put(p, AapConfigurationParser.PROP_PROCESSING_DIR_NAME, c.processingDirName());
    put(p, AapConfigurationParser.PROP_DATASET_SETTLE_TIME, c.datasetSettleTime());
    put(p, AapConfigurationParser.PROP_DATASET_MIN_NUM_IMAGES, c.datasetMinNumImages());
    put(p, AapConfigurationParser.PROP_PER_TREE_RESULTS_ENABLED, c.perTreeResultsEnabled());
    put(p, AapConfigurationParser.PROP_MAX_PROCESS_RUN_TIME, c.maxProcessRunTime());
    put(p, AapConfigurationParser.PROP_PROCESS_SUPERVISOR_PERIOD, c.processSupervisorPeriod());
    put(p, AapConfigurationParser.PROP_MAX_REFINED_DET_DIST_DELTA, c.maxRefinedDetDistDelta());
    put(p, AapConfigurationParser.PROP_NUM_THREADS_PER_PROCESS_JOB, c.numThreadsPerProcessJob());
    put(p, AapConfigurationParser.PROP_EXCLUDE_DIR_PAT, c.excludeDirPat());
    put(p, AapConfigurationParser.PROP_POST_PROCESS_HOOK, Shellwords.join(c.postProcessHook()));
    if (p.keySet().size() != AapConfigurationParser.ALL_PROPS.size()) {
      throw new AssertionError(String.format(Locale.US, "%s out of sync with %s",
          AapConfigurationWriter.class.getName(), AapConfigurationParser.class.getName()));
    }
    return p;
  }

  private static final void put(Properties p, String key, int i) {
    p.put(key, Integer.toString(i));
  }

  private static final void put(Properties p, String key, BigDecimal d) {
    p.put(key, toString(d));
  }

  private static final void put(Properties p, String key, boolean b) {
    p.put(key, b ? "yes" : "no");
  }

  private static final void put(Properties p, String key, TimeDuration d) {
    p.put(key, toString(d));
  }

  private static final void put(Properties p, String key, Pattern pat) {
    p.put(key, toString(pat));
  }

  private static final void put(Properties p, String key, String s) {
    p.put(key, s);
  }

  static String toString(TimeDuration d) {
    if (d == null) return null;
    return d.amount() + " " + toString(d.unit());
  }

  private static String toString(TimeUnit u) {
    if (u == null) return null;
    switch (u) {
      case DAYS:
        return "d";
      case HOURS:
        return "h";
      case MICROSECONDS:
        return "\u03BCs";
      case MILLISECONDS:
        return "ms";
      case MINUTES:
        return "m";
      case NANOSECONDS:
        return "ns";
      case SECONDS:
        return "s";
    }
    throw new IllegalArgumentException("\"" + u + "\": illegal TimeUnit");
  }

  private static String toString(BigDecimal d) {
    return (d == null) ? null : d.toString();
  }

  private static String toString(Pattern p) {
    return (p == null) ? null : p.pattern();
  }
}
