/*
 * Copyright (c) 2014, 2015 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.nio.channels.ClosedByInterruptException;
import java.nio.charset.StandardCharsets;
import java.nio.file.DirectoryIteratorException;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.LinkOption;
import java.nio.file.Path;
import java.nio.file.attribute.FileAttribute;
import java.nio.file.attribute.FileTime;
import java.nio.file.attribute.PosixFilePermission;
import java.nio.file.attribute.PosixFilePermissions;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.TimeUnit;

import org.imca_cat.aaproc.AaprocDbProtos.AaprocDbDataset;
import org.imca_cat.aaproc.AaprocDbProtos.AaprocDbDatasetProcessFailure;
import org.imca_cat.aaproc.AaprocDbProtos.AaprocDbDatasetProcessResult;
import org.imca_cat.procrun.DiscardingInputStreamConsumer;
import org.imca_cat.procrun.FixedSizeBufferInputStreamConsumer;
import org.imca_cat.procrun.InputStreamConsumerFactory;
import org.imca_cat.procrun.ProcessRunner;
import org.imca_cat.procrun.ProcessRunnerResult;

import com.google.protobuf.TextFormat;
import com.google.protobuf.Timestamp;

/**
 * Processes a {@link Dataset} using
 * <a href="http://www.globalphasing.com/autoproc/">autoPROC</a>.
 *
 * Tries to reuse existing processing results if they exist and if file
 * modification times indicate they are current. If this reuse attempt fails, a
 * from-scratch process attempt will be made.
 */
class AutoProcDatasetProcessor {
  private static final FileAttribute<Set<PosixFilePermission>> POSIX_DEFFILEMODE =
      PosixFilePermissions.asFileAttribute(PosixFilePermissions.fromString("rw-rw-rw-"));

  private final Dataset dataset;
  private final AapConfiguration conf;
  private final String helperExecutablePath;
  private final Path processingDir;
  private ProcessRunnerFactory processRunnerFactory;

  public AutoProcDatasetProcessor(Dataset dataset, AapConfiguration conf,
      String helperExecutablePath) {
    super();
    this.dataset = dataset;
    this.conf = conf;
    this.helperExecutablePath = helperExecutablePath;
    processingDir = dataset.directory().resolve(conf.processingDirName());
    processRunnerFactory = new DefaultProcessRunnerFactory();
  }

  public DatasetProcessResult process() throws InterruptedException {
    try {
      FileTime imagesLastModTime = imagesLastModifiedTime();
      FileTime resultLastModTime = resultLastModifiedTime();
      AaprocDbDatasetProcessResult dbResult = aaprocDbResult();
      if (dbResult != null && dbResult.hasFailure() && (imagesLastModTime == null ||
          !dbResult.hasTime() || toMillis(dbResult.getTime()) >= imagesLastModTime.toMillis())) {
        return newFailedDatasetProcessResult(dbResult.getFailure().getMessage());
      }
      if (resultLastModTime != null &&
          (imagesLastModTime == null || resultLastModTime.compareTo(imagesLastModTime) >= 0)) {
        return parseResult();
      }
      return basicProcess();
    } catch (InterruptedException e) {
      Thread.currentThread().interrupt();
      throw e;
    } catch (InterruptedIOException | ClosedByInterruptException e) {
      InterruptedException ex = new InterruptedException();
      ex.initCause(e);
      Thread.currentThread().interrupt();
      throw ex;
    } catch (ThreadDeath td) {
      throw td;
    } catch (Throwable t) {
      return new DatasetProcessResult(dataset, t);
    }
  }

  public ProcessRunnerFactory processRunnerFactory() {
    return processRunnerFactory;
  }

  public void processRunnerFactory(ProcessRunnerFactory f) {
    processRunnerFactory = f;
  }

  private DatasetProcessResult basicProcess() throws InterruptedException, IOException {
    deleteProcessingDir();
    createProcessingDir();
    String logFileName = newLogFileName();
    Path log = processingDir.resolve(logFileName);
    Path logTmp = processingDir.resolve(logFileName + ".tmp");
    ExecutorService service = null;
    Pair<ProcessRunnerResult, String> result = null;
    try {
      service = Executors.newCachedThreadPool();
      List<PairCallable<ProcessRunnerResult, String>> tasks = new ArrayList<>();
      tasks.add(newProcessRunnerPairCallable(logTmp));
      tasks.add(newProcessSupervisorPairCallable());
      result = service.invokeAny(tasks);
      service.shutdownNow();
      service.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);
      service = null;
      if (Files.exists(logTmp, LinkOption.NOFOLLOW_LINKS)) {
        Utilities.moveFile(logTmp, log);
      }
      logTmp = null;
    } catch (ExecutionException | RejectedExecutionException e) {
      return newFailedDatasetProcessResult("Failed to run process and supervisor", e);
    } finally {
      Utilities.shutdownNowSilently(service);
      Utilities.awaitTerminationSilently(service);
      Utilities.deleteFileSilently(logTmp);
    }
    ProcessRunnerResult prResult = result.first();
    String supervisorMessage = result.second();
    if (supervisorMessage != null) {
      writeFailureToAaprocDb(supervisorMessage);
      return newFailedDatasetProcessResult(supervisorMessage);
    }
    DatasetProcessResult failure = toDatasetProcessResultIfFailed(prResult);
    if (failure != null) return failure;
    return parseResult();
  }

  private void deleteProcessingDir() throws IOException {
    JavaUtilities.deleteRecursively(processingDir);
  }

  private void createProcessingDir() throws IOException {
    Files.createDirectory(processingDir);
  }

  private DatasetProcessResult toDatasetProcessResultIfFailed(ProcessRunnerResult r) {
    if (r.hasException()) {
      return new DatasetProcessResult(dataset, r.exception());
    }
    if (!r.hasExitStatus() || r.exitStatus().intValue() != 0) {
      return new DatasetProcessResult(dataset, new NonzeroExitStatusException(helperExecutablePath,
          r.exitStatus(), r.stderrString(StandardCharsets.UTF_8)));
    }
    return null;
  }

  private DatasetProcessResult parseResult() {
    return (new AutoProcResultParser(dataset, processingDir)).parse();
  }

  private FileTime resultLastModifiedTime() throws IOException {
    return AutoProcResultParser.resultLastModifiedTime(processingDir);
  }

  private FileTime imagesLastModifiedTime() throws IOException {
    FileTime result = null;
    try (DirectoryStream<Path> s = Files.newDirectoryStream(dataset.directory())) {
      for (Path each : s) {
        Path eachNamePath = each.getFileName();
        String eachName = (eachNamePath == null) ? "" : eachNamePath.toString();
        if (conf.diffCenterImagePat().matcher(eachName).matches()) continue;
        if (!conf.imagePat().matcher(eachName).matches()) continue;
        if (!Files.isRegularFile(each)) continue;
        FileTime eachTime = Files.getLastModifiedTime(each);
        if (result == null || eachTime.compareTo(result) > 0) result = eachTime;
      }
    } catch (DirectoryIteratorException e) {
      throw e.getCause();
    }
    return result;
  }

  private AaprocDbDatasetProcessResult aaprocDbResult() throws IOException {
    AaprocDbDataset db = readAaprocDb(dataset.directory().resolve(AaprocSystem.DB_FILE_NAME));
    return (db == null) ? null : db.getProcessResult();
  }

  private NioProcessBuilder newProcessBuilder(Path log) {
    NioProcessBuilder b =
        new DefaultNioProcessBuilder(helperExecutablePath, "-I", "..", "-B", "-nthreads",
            Integer.toString(conf.numThreadsPerProcessJob()));
    b.directory(processingDir);
    b.redirectOutput(log);
    return b;
  }

  private ProcessRunner newProcessRunner(Path log) {
    /*
     * Use a DiscardingInputStreamConsumer for the standard output stream
     * because the standard output stream is redirected by the ProcessBuilder
     * to the specified log.
     */
    return processRunnerFactory.newRunner(newProcessBuilder(log),
        DiscardingInputStreamConsumer.newFactory(1),
        FixedSizeBufferInputStreamConsumer.newFactory(512));
  }

  private PairCallable<ProcessRunnerResult, String> newProcessRunnerPairCallable(Path log) {
    return new PairCallable<ProcessRunnerResult, String>(newProcessRunner(log), null);
  }

  private AutoProcSupervisor newProcessSupervisor() {
    return new AutoProcSupervisor(dataset.directory(), processingDir,
        conf.maxProcessRunTime().amount(), conf.maxProcessRunTime().unit(),
        conf.processSupervisorPeriod().amount(), conf.processSupervisorPeriod().unit(),
        conf.maxRefinedDetDistDelta());
  }

  private PairCallable<ProcessRunnerResult, String> newProcessSupervisorPairCallable() {
    return new PairCallable<ProcessRunnerResult, String>(null, newProcessSupervisor());
  }

  private DatasetProcessResult newFailedDatasetProcessResult(String message) {
    return new DatasetProcessResult(dataset, new DatasetProcessFailedException(message));
  }

  private DatasetProcessResult newFailedDatasetProcessResult(String message, Throwable cause) {
    return new DatasetProcessResult(dataset, new DatasetProcessFailedException(message, cause));
  }

  private void writeFailureToAaprocDb(String message) throws IOException {
    Path dbFile = dataset.directory().resolve(AaprocSystem.DB_FILE_NAME);
    Path dbFileTmp = null;
    AaprocDbDataset db = readAaprocDb(dbFile);
    if (db == null) db = AaprocDbDataset.getDefaultInstance();
    AaprocDbDataset.Builder b = db.toBuilder();
    b.setProcessResult(newAaprocDbDatasetProcessResult(message));
    db = b.build();
    try {
      dbFileTmp = Files.createTempFile(dataset.directory(), AaprocSystem.DB_FILE_NAME + ".", ".tmp",
          POSIX_DEFFILEMODE);
      writeAaprocDb(dbFileTmp, db);
      Utilities.moveFile(dbFileTmp, dbFile);
      dbFileTmp = null;
    } finally {
      Utilities.deleteFileSilently(dbFileTmp);
    }
  }

  private static String newLogFileName() {
    DateFormat df = new SimpleDateFormat("yyyyMMdd_HHmmss", Locale.US);
    return String.format(Locale.US, "process.%s.log", df.format(new Date()));
  }

  private static AaprocDbDatasetProcessResult newAaprocDbDatasetProcessResult(
      String failureMessage) {
    AaprocDbDatasetProcessResult.Builder b = AaprocDbDatasetProcessResult.newBuilder();
    b.setTime(newTimestamp());
    b.setFailure(newAaprocDbDatasetProcessFailure(failureMessage));
    return b.build();
  }

  private static AaprocDbDatasetProcessFailure newAaprocDbDatasetProcessFailure(String message) {
    AaprocDbDatasetProcessFailure.Builder b = AaprocDbDatasetProcessFailure.newBuilder();
    b.setMessage(message);
    return b.build();
  }

  private static Timestamp newTimestamp() {
    return newTimestamp(System.currentTimeMillis());
  }

  private static Timestamp newTimestamp(long millisSinceEpoch) {
    long seconds = millisSinceEpoch / 1000L;
    long millis = millisSinceEpoch % 1000L;
    int nanos = (int)TimeUnit.MILLISECONDS.toNanos(millis);
    Timestamp.Builder b = Timestamp.newBuilder();
    b.setSeconds(seconds);
    b.setNanos(nanos);
    return b.build();
  }

  private static long toMillis(Timestamp t) {
    return TimeUnit.SECONDS.toMillis(t.getSeconds()) + TimeUnit.NANOSECONDS.toMillis(t.getNanos());
  }

  private static AaprocDbDataset readAaprocDb(Path p) throws IOException {
    AaprocDbDataset.Builder b = AaprocDbDataset.newBuilder();
    try (BufferedReader r = Files.newBufferedReader(p, StandardCharsets.UTF_8)) {
      TextFormat.merge(r, b);
    } catch (InterruptedIOException | ClosedByInterruptException e) {
      throw e;
    } catch (IOException e) {
      if (Files.notExists(p)) {
        /*
         * The db file either never existed or it was deleted (potentially at
         * the instant in time just before trying to read it), so ignore the
         * I/O exception and return null indicating there is no db file.
         */
        return null;
      } else {
        throw e;
      }
    }
    return b.build();
  }

  private static void writeAaprocDb(Path p, AaprocDbDataset d) throws IOException {
    try (BufferedWriter w = Files.newBufferedWriter(p, StandardCharsets.UTF_8)) {
      TextFormat.printer().print(d, w);
    }
  }

  private static class DefaultProcessRunnerFactory implements ProcessRunnerFactory {
    @Override
    public ProcessRunner newRunner(NioProcessBuilder builder,
        InputStreamConsumerFactory stdoutConsumerFactory,
        InputStreamConsumerFactory stderrConsumerFactory) {
      return new ProcessRunner(builder.toProcessBuilder(), stdoutConsumerFactory,
          stderrConsumerFactory);
    }
  }
}
