/*
 * Copyright (c) 2014, 2015 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.io.IOException;
import java.nio.file.DirectoryIteratorException;
import java.nio.file.DirectoryStream;
import java.nio.file.FileSystem;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.PathMatcher;
import java.nio.file.attribute.FileTime;

/**
 * Parses the result of an
 * <a href="http://www.globalphasing.com/autoproc/">autoPROC</a> processing run
 * on a dataset.
 */
class AutoProcResultParser {
  private static final String AUTO_PROC_XML_NAME = "autoPROC.xml";
  private static final String PROCESS_LOG_NAME_SYN_PAT = "regex:process\\.\\d{8}_\\d{6}\\.log";

  private final Dataset dataset;
  private final Path processingDir;

  public AutoProcResultParser(Dataset dataset, Path processingDir) {
    super();
    this.dataset = dataset;
    this.processingDir = processingDir;
  }

  public DatasetProcessResult parse() {
    Path autoProcXml = findAutoProcXml(processingDir);
    if (autoProcXml == null) {
      return new DatasetProcessResult(dataset,
          new DatasetProcessFailedException("File \"" + AUTO_PROC_XML_NAME + "\" not found"));
    }
    return AutoProcXmlParser.parse(dataset, autoProcXml);
  }

  public static FileTime resultLastModifiedTime(Path processingDir) throws IOException {
    Path processLog = findProcessLog(processingDir);
    return (processLog == null) ? null : Files.getLastModifiedTime(processLog);
  }

  private static Path findAutoProcXml(Path processingDir) {
    Path result = processingDir.resolve(AUTO_PROC_XML_NAME);
    return Files.isRegularFile(result) ? result : null;
  }

  private static Path findProcessLog(Path processingDir) {
    try (DirectoryStream<Path> s = Files.newDirectoryStream(processingDir,
        newProcessLogFilter(processingDir.getFileSystem()))) {
      Path result = null;
      for (Path each : s) {
        if (result == null || each.getFileName().compareTo(result.getFileName()) > 0) {
          result = each;
        }
      }
      if (result == null) return null;
      return Files.isRegularFile(result) ? result : null;
    } catch (DirectoryIteratorException e) {
      return null;
    } catch (IOException e) {
      return null;
    }
  }

  private static DirectoryStream.Filter<Path> newProcessLogFilter(FileSystem fs) {
    final PathMatcher matcher = fs.getPathMatcher(PROCESS_LOG_NAME_SYN_PAT);
    return new DirectoryStream.Filter<Path>() {
      @Override
      public boolean accept(Path entry) throws IOException {
        return matcher.matches(entry.getFileName());
      }
    };
  }
}
