/*
 * Copyright (c) 2015 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.io.IOException;
import java.math.BigDecimal;
import java.nio.file.DirectoryIteratorException;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.InvalidPathException;
import java.nio.file.LinkOption;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.TimeUnit;

/**
 * Monitors an <a href="http://www.globalphasing.com/autoproc/">autoPROC</a>
 * process, the files it creates, and the dataset it is processing, and, if
 * necessary, indicates that the process should be terminated. The criteria
 * that trigger a termination indication for the process are:
 * <ul>
 * <li>It exceeds the maximum running time.
 * <li>The integration refined detector distance exceeds the maximum delta.
 * <li>It is unable to carry out integration refinement at all.
 * </ul>
 */
class AutoProcSupervisor implements Callable<String> {
  private static final String INTEGRATE_LOG_NAME = "integ.log";

  @SuppressWarnings("unused")
  private final Path datasetDir;
  private final Path processingDir;
  private final long maxRunTime;
  private final TimeUnit maxRunTimeUnit;
  private final long period;
  private final TimeUnit periodUnit;
  private final BigDecimal maxDetDistDelta;

  /**
   * Constructs a new instance with the specified dataset directory, processing
   * directory, maximum running time, period, and maximum integration refined
   * detector distance delta.
   *
   * @param datasetDir the directory of the dataset the autoPROC process is
   *          processing
   * @param processingDir the processing directory of the autoPROC process
   * @param maxRunTime the maximum running time of the autoPROC process
   * @param maxRunTimeUnit the unit of {@code maxRunTime}
   * @param period the period between successive supervisory checks
   * @param periodUnit the unit of {@code period}
   * @param maxDetDistDelta the maximum delta, in millimeters, of the
   *          integration refined detector distance
   */
  public AutoProcSupervisor(Path datasetDir, Path processingDir, long maxRunTime,
      TimeUnit maxRunTimeUnit, long period, TimeUnit periodUnit, BigDecimal maxDetDistDelta) {
    super();
    this.datasetDir = datasetDir;
    this.processingDir = processingDir;
    this.maxRunTime = maxRunTime;
    this.maxRunTimeUnit = maxRunTimeUnit;
    this.period = period;
    this.periodUnit = periodUnit;
    this.maxDetDistDelta = maxDetDistDelta;
  }

  /**
   * Supervises the process. Returns a message indicating why the process
   * should be terminated if it should be terminated.
   *
   * @return a message indicating why the process should be terminated or
   *         {@code null} if the process does not need to be terminated
   */
  @Override
  public String call() throws InterruptedException {
    ExecutorService s = null;
    try {
      s = Executors.newCachedThreadPool();
      List<Callable<String>> tasks = new ArrayList<>();
      tasks.add(newMaxRunTimeTask(maxRunTime, maxRunTimeUnit));
      tasks.add(newRefinementTask(processingDir, period, periodUnit, maxDetDistDelta));
      return s.invokeAny(tasks);
    } catch (ExecutionException | RejectedExecutionException e) {
      throw new RuntimeException("Failed to run tasks", e);
    } finally {
      Utilities.shutdownNowSilently(s);
      Utilities.awaitTerminationSilently(s);
    }
  }

  private static Callable<String> newMaxRunTimeTask(final long max, final TimeUnit maxUnit) {
    return new Callable<String>() {
      @Override
      public String call() throws InterruptedException {
        maxUnit.sleep(max);
        return "Maximum running time exceeded";
      }
    };
  }

  private static Callable<String> newRefinementTask(final Path processingDir, final long period,
      final TimeUnit periodUnit, final BigDecimal maxDetDistDelta) {
    return new Callable<String>() {
      @Override
      public String call() throws InterruptedException {
        Path log;
        /* Wait for the integrate log to exist. */
        for (;;) {
          log = findIntegrateLog(processingDir);
          if (log != null) break;
          periodUnit.sleep(period);
        }
        /*
         * Watch the log for a refinement failure or runaway refined detector
         * distance.
         */
        for (;;) {
          XdsIntegrateResult r = XdsIntegrateLogParser.parse(log);
          if (!r.hasException()) {
            if (r.failedToRefine()) {
              return "Integration refinement failed";
            }
            BigDecimal detDistDelta = r.refinedDetectorDistanceDelta();
            if (detDistDelta != null && detDistDelta.compareTo(maxDetDistDelta) > 0) {
              return "Maximum integration refined detector distance delta exceeded";
            }
          }
          periodUnit.sleep(period);
        }
      }
    };
  }

  private static Path findIntegrateLog(Path processingDir) {
    try (DirectoryStream<Path> s = Files.newDirectoryStream(processingDir, newDirectoryFilter())) {
      for (Path each : s) {
        Path result;
        try {
          result = each.resolve(INTEGRATE_LOG_NAME);
        } catch (InvalidPathException e) {
          continue;
        }
        if (Files.exists(result)) return result;
      }
      return null;
    } catch (DirectoryIteratorException e) {
      return null;
    } catch (IOException e) {
      return null;
    }
  }

  private static DirectoryStream.Filter<Path> newDirectoryFilter() {
    return new DirectoryStream.Filter<Path>() {
      @Override
      public boolean accept(Path entry) throws IOException {
        return Files.isDirectory(entry, LinkOption.NOFOLLOW_LINKS);
      }
    };
  }
}
