/*
 * Copyright (c) 2017 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.Reader;
import java.math.BigDecimal;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.EndElement;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;

/**
 * Parses an XML autoPROC result (e.g., the contents of {@code autoPROC.xml}).
 */
class AutoProcXmlParser {
  private static final String CONTAINER = "AutoProcContainer";
  private static final String AUTO_PROC = "AutoProc";
  private static final String SCALING_CONTAINER = "AutoProcScalingContainer";
  private static final String SCALING_STATS = "AutoProcScalingStatistics";

  private static final String SCALING_STATS_TYPE = "scalingStatisticsType";
  private static final String RES_LIMIT_LOW = "resolutionLimitLow";
  private static final String RES_LIMIT_HIGH = "resolutionLimitHigh";
  private static final String R_MERGE = "rMerge";
  private static final String I_OVER_SIGMA = "meanIOverSigI";
  private static final String COMPLETENESS = "completeness";
  private static final String MULTIPLICITY = "multiplicity";
  private static final String UNIT_CELL_A = "refinedCell_a";
  private static final String UNIT_CELL_B = "refinedCell_b";
  private static final String UNIT_CELL_C = "refinedCell_c";
  private static final String UNIT_CELL_ALPHA = "refinedCell_alpha";
  private static final String UNIT_CELL_BETA = "refinedCell_beta";
  private static final String UNIT_CELL_GAMMA = "refinedCell_gamma";
  private static final String SPACE_GROUP = "spaceGroup";

  /* @formatter:off */
  private static final List<String> SCALING_STATS_TYPE_PATH = Collections.unmodifiableList(
      Arrays.asList(
          CONTAINER,
          SCALING_CONTAINER,
          SCALING_STATS,
          SCALING_STATS_TYPE));
  private static final List<String> RES_LIMIT_LOW_PATH = Collections.unmodifiableList(
      Arrays.asList(
          CONTAINER,
          SCALING_CONTAINER,
          SCALING_STATS,
          RES_LIMIT_LOW));
  private static final List<String> RES_LIMIT_HIGH_PATH = Collections.unmodifiableList(
      Arrays.asList(
          CONTAINER,
          SCALING_CONTAINER,
          SCALING_STATS,
          RES_LIMIT_HIGH));
  private static final List<String> R_MERGE_PATH = Collections.unmodifiableList(
      Arrays.asList(
          CONTAINER,
          SCALING_CONTAINER,
          SCALING_STATS,
          R_MERGE));
  private static final List<String> I_OVER_SIGMA_PATH = Collections.unmodifiableList(
      Arrays.asList(
          CONTAINER,
          SCALING_CONTAINER,
          SCALING_STATS,
          I_OVER_SIGMA));
  private static final List<String> COMPLETENESS_PATH = Collections.unmodifiableList(
      Arrays.asList(
          CONTAINER,
          SCALING_CONTAINER,
          SCALING_STATS,
          COMPLETENESS));
  private static final List<String> MULTIPLICITY_PATH = Collections.unmodifiableList(
      Arrays.asList(
          CONTAINER,
          SCALING_CONTAINER,
          SCALING_STATS,
          MULTIPLICITY));
  private static final List<String> UNIT_CELL_A_PATH = Collections.unmodifiableList(
      Arrays.asList(
          CONTAINER,
          AUTO_PROC,
          UNIT_CELL_A));
  private static final List<String> UNIT_CELL_B_PATH = Collections.unmodifiableList(
      Arrays.asList(
          CONTAINER,
          AUTO_PROC,
          UNIT_CELL_B));
  private static final List<String> UNIT_CELL_C_PATH = Collections.unmodifiableList(
      Arrays.asList(
          CONTAINER,
          AUTO_PROC,
          UNIT_CELL_C));
  private static final List<String> UNIT_CELL_ALPHA_PATH = Collections.unmodifiableList(
      Arrays.asList(
          CONTAINER,
          AUTO_PROC,
          UNIT_CELL_ALPHA));
  private static final List<String> UNIT_CELL_BETA_PATH = Collections.unmodifiableList(
      Arrays.asList(
          CONTAINER,
          AUTO_PROC,
          UNIT_CELL_BETA));
  private static final List<String> UNIT_CELL_GAMMA_PATH = Collections.unmodifiableList(
      Arrays.asList(
          CONTAINER,
          AUTO_PROC,
          UNIT_CELL_GAMMA));
  private static final List<String> SPACE_GROUP_PATH = Collections.unmodifiableList(
      Arrays.asList(
          CONTAINER,
          AUTO_PROC,
          SPACE_GROUP));
  /* @formatter:on */

  private static final int MAX_PATH_DEPTH;
  static {
    int max = 0;
    for (List<String> each : Arrays.asList(SCALING_STATS_TYPE_PATH, RES_LIMIT_LOW_PATH,
        RES_LIMIT_HIGH_PATH, R_MERGE_PATH, I_OVER_SIGMA_PATH, COMPLETENESS_PATH, MULTIPLICITY_PATH,
        UNIT_CELL_A_PATH, UNIT_CELL_B_PATH, UNIT_CELL_C_PATH, UNIT_CELL_ALPHA_PATH,
        UNIT_CELL_BETA_PATH, UNIT_CELL_GAMMA_PATH, SPACE_GROUP_PATH)) {
      if (each.size() > max) max = each.size();
    }
    MAX_PATH_DEPTH = max;
  };

  private static final String SCALING_STATS_TYPE_OVERALL = "overall";

  private final Dataset dataset;
  private final Reader in;

  /**
   * @throws NullPointerException if any arguments are {@code null}
   */
  public AutoProcXmlParser(Dataset dataset, Reader in) {
    super();
    this.dataset = Objects.requireNonNull(dataset);
    this.in = Objects.requireNonNull(in);
  }

  public DatasetProcessResult parse() {
    try {
      LinkedList<String> path = new LinkedList<String>();
      int pathDepth = 0;
      String scalingStatsType = null;
      boolean reachedEndOfAutoProcElement = false;
      boolean reachedEndOfOverallAutoProcScalingStatsElement = false;
      StringValuedDatasetProcessResultBuilder result = new StringValuedDatasetProcessResultBuilder(
          dataset);
      XMLInputFactory factory = XMLInputFactory.newFactory();
      factory.setProperty(XMLInputFactory.SUPPORT_DTD, false);
      factory.setProperty(XMLInputFactory.IS_SUPPORTING_EXTERNAL_ENTITIES, false);
      XMLEventReader r = factory.createXMLEventReader(in);
      while (r.hasNext()) {
        XMLEvent event = r.nextEvent();
        if (event.isStartElement()) {
          pathDepth++;
          if (pathDepth > MAX_PATH_DEPTH) continue;
          StartElement start = event.asStartElement();
          String name = start.getName().getLocalPart();
          path.addLast(name);
          switch (name) {
            case SCALING_STATS_TYPE:
              if (!SCALING_STATS_TYPE_PATH.equals(path)) continue;
              scalingStatsType = r.nextEvent().asCharacters().getData();
              continue;
            case RES_LIMIT_LOW:
              if (!RES_LIMIT_LOW_PATH.equals(path)) continue;
              if (!SCALING_STATS_TYPE_OVERALL.equals(scalingStatsType)) continue;
              result.lowResolution(r.nextEvent().asCharacters().getData());
              continue;
            case RES_LIMIT_HIGH:
              if (!RES_LIMIT_HIGH_PATH.equals(path)) continue;
              if (!SCALING_STATS_TYPE_OVERALL.equals(scalingStatsType)) continue;
              result.highResolution(r.nextEvent().asCharacters().getData());
              continue;
            case R_MERGE:
              if (!R_MERGE_PATH.equals(path)) continue;
              if (!SCALING_STATS_TYPE_OVERALL.equals(scalingStatsType)) continue;
              result.rMerge(r.nextEvent().asCharacters().getData());
              continue;
            case I_OVER_SIGMA:
              if (!I_OVER_SIGMA_PATH.equals(path)) continue;
              if (!SCALING_STATS_TYPE_OVERALL.equals(scalingStatsType)) continue;
              result.iOverSigma(r.nextEvent().asCharacters().getData());
              continue;
            case COMPLETENESS:
              if (!COMPLETENESS_PATH.equals(path)) continue;
              if (!SCALING_STATS_TYPE_OVERALL.equals(scalingStatsType)) continue;
              result.completeness(r.nextEvent().asCharacters().getData());
              continue;
            case MULTIPLICITY:
              if (!MULTIPLICITY_PATH.equals(path)) continue;
              if (!SCALING_STATS_TYPE_OVERALL.equals(scalingStatsType)) continue;
              result.multiplicity(r.nextEvent().asCharacters().getData());
              continue;
            case UNIT_CELL_A:
              if (!UNIT_CELL_A_PATH.equals(path)) continue;
              result.unitCell("a", r.nextEvent().asCharacters().getData());
              continue;
            case UNIT_CELL_B:
              if (!UNIT_CELL_B_PATH.equals(path)) continue;
              result.unitCell("b", r.nextEvent().asCharacters().getData());
              continue;
            case UNIT_CELL_C:
              if (!UNIT_CELL_C_PATH.equals(path)) continue;
              result.unitCell("c", r.nextEvent().asCharacters().getData());
              continue;
            case UNIT_CELL_ALPHA:
              if (!UNIT_CELL_ALPHA_PATH.equals(path)) continue;
              result.unitCell("alpha", r.nextEvent().asCharacters().getData());
              continue;
            case UNIT_CELL_BETA:
              if (!UNIT_CELL_BETA_PATH.equals(path)) continue;
              result.unitCell("beta", r.nextEvent().asCharacters().getData());
              continue;
            case UNIT_CELL_GAMMA:
              if (!UNIT_CELL_GAMMA_PATH.equals(path)) continue;
              result.unitCell("gamma", r.nextEvent().asCharacters().getData());
              continue;
            case SPACE_GROUP:
              if (!SPACE_GROUP_PATH.equals(path)) continue;
              result.spaceGroup(r.nextEvent().asCharacters().getData());
              continue;
            default:
              continue;
          }
        } else if (event.isEndElement()) {
          pathDepth--;
          if (pathDepth >= MAX_PATH_DEPTH) continue;
          EndElement end = event.asEndElement();
          String name = end.getName().getLocalPart();
          switch (name) {
            case AUTO_PROC:
              reachedEndOfAutoProcElement = true;
              break;
            case SCALING_STATS:
              if (SCALING_STATS_TYPE_OVERALL.equals(scalingStatsType)) {
                reachedEndOfOverallAutoProcScalingStatsElement = true;
              }
              break;
          }
          if (reachedEndOfAutoProcElement && reachedEndOfOverallAutoProcScalingStatsElement) {
            return result.build();
          }
          path.removeLast();
        }
      }
      String msg;
      if (result.isEmpty()) {
        msg = "No result values found";
      } else if (!reachedEndOfAutoProcElement) {
        msg = "Missing " + AUTO_PROC + " XML element end tag";
      } else if (!reachedEndOfOverallAutoProcScalingStatsElement) {
        msg = "Missing overall " + SCALING_STATS + " XML element end tag";
      } else {
        msg = "Invalid autoPROC XML file format";
      }
      return new DatasetProcessResult(dataset, new DataFormatException(msg));
    } catch (XMLStreamException e) {
      return new DatasetProcessResult(dataset, new DataFormatException(e));
    }
  }

  public static DatasetProcessResult parse(Dataset dataset, Path xml) {
    try (BufferedReader reader = Files.newBufferedReader(xml, StandardCharsets.UTF_8)) {
      return (new AutoProcXmlParser(dataset, reader)).parse();
    } catch (IOException e) {
      if (e instanceof InterruptedIOException) {
        Thread.currentThread().interrupt();
      }
      return new DatasetProcessResult(dataset, e);
    }
  }

  private static class StringValuedDatasetProcessResultBuilder {
    private static final List<String> UNIT_CELL_KEYS = Collections
        .unmodifiableList(Arrays.asList("a", "b", "c", "alpha", "beta", "gamma"));

    private final Dataset dataset;
    private String lowResolution;
    private String highResolution;
    private String rMerge;
    private String iOverSigma;
    private String completeness;
    private String multiplicity;
    private Map<String, String> unitCell = new HashMap<>(6);
    private String spaceGroup;

    public StringValuedDatasetProcessResultBuilder(Dataset dataset) {
      super();
      this.dataset = dataset;
    }

    @SuppressWarnings("unused")
    public String lowResolution() {
      return lowResolution;
    }

    public StringValuedDatasetProcessResultBuilder lowResolution(String lowResolution) {
      this.lowResolution = lowResolution;
      return this;
    }

    @SuppressWarnings("unused")
    public String highResolution() {
      return highResolution;
    }

    public StringValuedDatasetProcessResultBuilder highResolution(String highResolution) {
      this.highResolution = highResolution;
      return this;
    }

    @SuppressWarnings("unused")
    public String rMerge() {
      return rMerge;
    }

    public StringValuedDatasetProcessResultBuilder rMerge(String rMerge) {
      this.rMerge = rMerge;
      return this;
    }

    @SuppressWarnings("unused")
    public String iOverSigma() {
      return iOverSigma;
    }

    public StringValuedDatasetProcessResultBuilder iOverSigma(String iOverSigma) {
      this.iOverSigma = iOverSigma;
      return this;
    }

    @SuppressWarnings("unused")
    public String completeness() {
      return completeness;
    }

    public StringValuedDatasetProcessResultBuilder completeness(String completeness) {
      this.completeness = completeness;
      return this;
    }

    @SuppressWarnings("unused")
    public String multiplicity() {
      return multiplicity;
    }

    public StringValuedDatasetProcessResultBuilder multiplicity(String multiplicity) {
      this.multiplicity = multiplicity;
      return this;
    }

    /**
     * @param key one of "a", "b", "c", "alpha", "beta", or "gamma"
     *
     * @throws IllegalArgumentException if {@code key} is not one of allowed
     *           keys
     */
    @SuppressWarnings("unused")
    public String unitCell(String key) {
      if (!UNIT_CELL_KEYS.contains(key)) {
        throw new IllegalArgumentException("key not one of: a, b, c, alpha, beta, gamma");
      }
      return unitCell.get(key);
    }

    /**
     * @param key one of "a", "b", "c", "alpha", "beta", or "gamma"
     *
     * @throws IllegalArgumentException if {@code key} is not one of allowed
     *           keys
     */
    public StringValuedDatasetProcessResultBuilder unitCell(String key, String value) {
      if (!UNIT_CELL_KEYS.contains(key)) {
        throw new IllegalArgumentException("key not one of: a, b, c, alpha, beta, gamma");
      }
      unitCell.put(key, value);
      return this;
    }

    @SuppressWarnings("unused")
    public String spaceGroup() {
      return spaceGroup;
    }

    public StringValuedDatasetProcessResultBuilder spaceGroup(String spaceGroup) {
      this.spaceGroup = spaceGroup;
      return this;
    }

    public boolean isEmpty() {
      return lowResolution == null && highResolution == null && rMerge == null &&
          iOverSigma == null && completeness == null && multiplicity == null &&
          unitCell.isEmpty() && spaceGroup == null;
    }

    public DatasetProcessResult build() {
      DatasetProcessResultBuilder b = new DatasetProcessResultBuilder();
      try {
        b.dataset(dataset);
        if (lowResolution != null) {
          try {
            b.lowResolution(new BigDecimal(lowResolution));
          } catch (NumberFormatException e) {
            throw new DataFormatException("Invalid low resolution limit \"" + lowResolution + "\"",
                e);
          }
        }
        if (highResolution != null) {
          try {
            b.highResolution(new BigDecimal(highResolution));
          } catch (NumberFormatException e) {
            throw new DataFormatException(
                "Invalid high resolution limit \"" + highResolution + "\"", e);
          }
        }
        if (rMerge != null) {
          try {
            b.rMerge(new BigDecimal(rMerge));
          } catch (NumberFormatException e) {
            throw new DataFormatException("Invalid Rmerge \"" + rMerge + "\"", e);
          }
        }
        if (iOverSigma != null) {
          try {
            b.iOverSigma(new BigDecimal(iOverSigma));
          } catch (NumberFormatException e) {
            throw new DataFormatException("Invalid mean I/sigma \"" + iOverSigma + "\"", e);
          }
        }
        if (completeness != null) {
          try {
            b.completeness(new BigDecimal(completeness));
          } catch (NumberFormatException e) {
            throw new DataFormatException("Invalid completeness \"" + completeness + "\"", e);
          }
        }
        if (multiplicity != null) {
          try {
            b.multiplicity(new BigDecimal(multiplicity));
          } catch (NumberFormatException e) {
            throw new DataFormatException("Invalid multiplicity \"" + multiplicity + "\"", e);
          }
        }
        if (!unitCell.isEmpty()) {
          BigDecimal[] values = new BigDecimal[UNIT_CELL_KEYS.size()];
          for (int i = 0; i < values.length; i++) {
            String valueString = unitCell.get(UNIT_CELL_KEYS.get(i));
            if (valueString == null) {
              throw new DataFormatException(
                  "Missing unit cell " + UNIT_CELL_KEYS.get(i) + " value");
            }
            try {
              values[i] = new BigDecimal(valueString);
            } catch (NumberFormatException e) {
              throw new DataFormatException(
                  "Invalid unit cell " + UNIT_CELL_KEYS.get(i) + " \"" + valueString + "\"", e);
            }
          }
          b.unitCell(new UnitCell(values));
        }
        if (spaceGroup != null) {
          b.spaceGroup(spaceGroup);
        }
      } catch (DataFormatException e) {
        return new DatasetProcessResult(dataset, e);
      }
      return b.toDatasetProcessResult();
    }
  }
}
