/*
 * Copyright (c) 2014, 2015 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.io.BufferedWriter;
import java.io.IOException;
import java.math.BigDecimal;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Locale;

/**
 * Prints a list of {@link DatasetProcessResult} instances as an HTML document.
 */
class DatasetProcessResultsHtmlPrinter {
  private final Iterable<DatasetProcessResult> results;
  private final Path destination;
  private BufferedWriter out;

  /**
   * Constructs an instance with the specified results and destination.
   *
   * @param results results to print
   * @param destination where to print the results
   */
  public DatasetProcessResultsHtmlPrinter(Iterable<DatasetProcessResult> results,
      Path destination) {
    super();
    this.results = results;
    this.destination = destination;
    out = null;
  }

  public void print() throws IOException {
    try (BufferedWriter w = Files.newBufferedWriter(destination, StandardCharsets.UTF_8)) {
      out = w;
      printHeader();
      printResults();
      printFooter();
    } finally {
      out = null;
    }
  }

  private void printHeader() throws IOException {
    println("<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01//EN\"");
    println("  \"http://www.w3.org/TR/html4/strict.dtd\">");
    println("<html lang=\"en-us\">");
    println("<head>");
    println("<title>Aap Result</title>");
    println("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">");
    println("<style type=\"text/css\">");
    println("body {");
    println("  font-family: monospace;");
    println("}");
    println();
    println("table {");
    println("  border-collapse: collapse;");
    println("}");
    println();
    println("table td {");
    println("  border-style: solid;");
    println("  border-color: black;");
    println("  border-width: 1px;");
    println("}");
    println("</style>");
    println("</head>");
    println("<body>");
    println();
  }

  private void printResults() throws IOException {
    printResultsHeader();
    printResultsRows();
    printResultsFooter();
  }

  private void printResultsHeader() throws IOException {
    println("<table>");
    println("<tr>");
    println("  <th>Path</th>");
    println("  <th>Status</th>");
    println("  <th>SG</th>");
    println("  <th>Unit Cell</th>");
    println("  <th>Res</th>");
    println("  <th>Rmerge</th>");
    println("  <th>Comp.</th>");
    println("  <th>I/SigI</th>");
    println("  <th>Mult.</th>");
    println("</tr>");
  }

  private void printResultsRows() throws IOException {
    for (DatasetProcessResult each : results) {
      printResultRow(each);
    }
  }

  private void printResultRow(DatasetProcessResult r) throws IOException {
    println("<tr>");
    printf("  <td>%s</td>%n", escape(r.dataset().directory().toString()));
    if (r.succeeded()) {
      println("  <td>OK</td>");
    } else {
      print("  <td ");
      if (r.hasException()) {
        printf("title=\"%s\" ", escape(toString(r.exception())));
      }
      println("style=\"color: red\">FAILED</td>");
    }
    printf("  <td>%s</td>%n", escape(toString(r.spaceGroup())));
    printf("  <td>%s</td>%n", escape(toString(r.unitCell())));
    if (r.lowResolution() == null && r.highResolution() == null) {
      println("  <td></td>");
    } else {
      printf("  <td>%s&ndash;%s</td>%n", escape(toString(r.lowResolution())),
          escape(toString(r.highResolution())));
    }
    printf("  <td>%s</td>%n", escape(toString(r.rMerge())));
    printf("  <td>%s</td>%n", escape(toString(r.completeness())));
    printf("  <td>%s</td>%n", escape(toString(r.iOverSigma())));
    printf("  <td>%s</td>%n", escape(toString(r.multiplicity())));
    println("</tr>");
  }

  private void printResultsFooter() throws IOException {
    println("</table>");
  }

  private void printFooter() throws IOException {
    println();
    println("</body>");
    println("</html>");
  }

  private void print(String s) throws IOException {
    out.write(s);
  }

  private void println() throws IOException {
    out.newLine();
  }

  private void println(String s) throws IOException {
    out.write(s);
    out.newLine();
  }

  private void printf(String format, Object... args) throws IOException {
    out.write(format(format, args));
  }

  private static String escape(String value) {
    return JavaUtilities.escapeHtml(value);
  }

  private static String toString(String s) {
    return (s == null) ? "" : s;
  }

  private static String toString(Throwable t) {
    if (t == null) return "";
    String m = t.getMessage();
    if (m != null) return m;
    return t.toString();
  }

  private static String toString(UnitCell c) {
    if (c == null) return "";
    List<BigDecimal> lengths = c.lengths();
    List<BigDecimal> angles = c.angles();
    return format("%s %s %s %s %s %s", lengths.get(0), lengths.get(1), lengths.get(2),
        angles.get(0), angles.get(1), angles.get(2));
  }

  private static String toString(BigDecimal d) {
    return (d == null) ? "" : d.toString();
  }

  private static String format(String format, Object... args) {
    return String.format(Locale.US, format, args);
  }
}
