/*
 * Copyright (c) 2014 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.io.IOException;
import java.io.InterruptedIOException;
import java.nio.channels.ClosedByInterruptException;
import java.nio.file.DirectoryIteratorException;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.LinkOption;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.attribute.FileTime;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Callable;
import java.util.concurrent.CompletionService;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.imca_cat.procrun.GrowingBufferInputStreamConsumer;
import org.imca_cat.procrun.InputStreamConsumerFactory;
import org.imca_cat.procrun.ProcessRunner;
import org.imca_cat.procrun.ProcessRunnerResult;

/**
 * Processes datasets.
 */
class DatasetProcessService implements AutoCloseable {
  private static final Pattern HOOK_VAR = Pattern.compile("(?<var>(?>\\$[a-zA-Z_][a-zA-Z0-9_]*))");

  private final AapConfiguration conf;
  private final AaprocSystem sys;
  private final DatasetProcessServiceExceptionHandler exceptionHandler;
  private final BlockingQueue<DatasetItem> datasetQueue;
  private final AtomicInteger datasetQueueUnprocessedCount;
  private final BlockingQueue<DatasetProcessResult> processed;
  private final List<DatasetProcessServiceListener> listeners;
  private final ExecutorService processorEnqueuer;
  private final ExecutorService processorExecutor;
  private final CompletionService<DatasetProcessResult> processor;
  private final ExecutorService processorDequeuer;
  private final ExecutorService hookRunnerExecutor;
  private final CompletionService<ProcessRunnerResult> hookRunner;
  private final ExecutorService hookRunnerDequeuer;
  private final Object closeLock;
  private final DatasetProcessResult closedSentinel;
  private volatile boolean closed;
  private boolean started;
  private DatasetProcessorFactory datasetProcessorFactory;
  private ProcessRunnerFactory processRunnerFactory;

  public DatasetProcessService(AapConfiguration conf, AaprocSystem sys,
      DatasetProcessServiceExceptionHandler exceptionHandler) {
    super();
    this.conf = conf;
    this.sys = sys;
    this.exceptionHandler = exceptionHandler;
    datasetQueue = new LinkedBlockingQueue<>();
    datasetQueueUnprocessedCount = new AtomicInteger();
    processed = new LinkedBlockingQueue<>();
    listeners = new CopyOnWriteArrayList<>();
    processorEnqueuer = Executors.newSingleThreadExecutor();
    processorExecutor = Executors.newFixedThreadPool(conf.maxNumProcessJobs());
    processor = new ExecutorCompletionService<>(processorExecutor);
    processorDequeuer = Executors.newSingleThreadExecutor();
    hookRunnerExecutor = Executors.newCachedThreadPool();
    hookRunner = new ExecutorCompletionService<>(hookRunnerExecutor);
    hookRunnerDequeuer = Executors.newSingleThreadExecutor();
    closeLock = new Object();
    closedSentinel = new DatasetProcessResult(new Dataset(Paths.get("")), null);
    closed = false;
    started = false;
    datasetProcessorFactory = new DefaultDatasetProcessorFactory();
    processRunnerFactory = new DefaultProcessRunnerFactory();
  }

  /**
   * @throws ClosedServiceException if this service is closed
   * @throws IllegalStateException if this service is already started
   */
  public void start() {
    synchronized (closeLock) {
      if (closed) throw new ClosedServiceException();
      if (started) throw new IllegalStateException("Already started");
      started = true;
      processorEnqueuer.submit(newProcessDatasetsRunnable(this));
      processorDequeuer.submit(newEnqueueProcessedResultsRunnable(this));
      hookRunnerDequeuer.submit(newHandlePostProcessHookResultsRunnable(this));
    }
  }

  /**
   * @throws ClosedServiceException if this service is closed
   */
  public void put(Dataset d) throws InterruptedException {
    if (closed) throw new ClosedServiceException();
    addToDatasetQueue(DatasetItem.unprocessed(d));
  }

  /**
   * @throws ClosedServiceException if this service is closed
   */
  public DatasetProcessResult take() throws InterruptedException {
    if (closed) throw new ClosedServiceException();
    DatasetProcessResult result = processed.take();
    if (result == closedSentinel) throw new ClosedServiceException();
    return result;
  }

  @Override
  public void close() {
    synchronized (closeLock) {
      if (closed) return;
      closed = true;
      Throwable t = null;
      boolean interrupted = Thread.interrupted();
      for (ExecutorService each : Arrays.asList(processorDequeuer, processorEnqueuer,
          processorExecutor, hookRunnerDequeuer, hookRunnerExecutor)) {
        try {
          each.shutdownNow();
          for (;;) {
            try {
              each.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);
              break;
            } catch (InterruptedException e) {
              interrupted = true;
            }
          }
        } catch (ThreadDeath td) {
          t = td;
        } catch (Throwable th) {
          if (t == null) {
            t = th;
          } else {
            t.addSuppressed(th);
          }
        }
      }
      try {
        processed.clear();
        processed.add(closedSentinel);
      } catch (ThreadDeath td) {
        t = td;
      } catch (Throwable th) {
        if (t == null) {
          t = th;
        } else {
          t.addSuppressed(th);
        }
      }
      try {
        listeners.clear();
      } catch (ThreadDeath td) {
        t = td;
      } catch (Throwable th) {
        if (t == null) {
          t = th;
        } else {
          t.addSuppressed(th);
        }
      }
      if (interrupted) Thread.currentThread().interrupt();
      if (t != null) throwUncheckedException(t);
    }
  }

  public void addListener(DatasetProcessServiceListener l) {
    listeners.add(l);
  }

  public void removeListener(DatasetProcessServiceListener l) {
    listeners.remove(l);
  }

  public DatasetProcessorFactory datasetProcessorFactory() {
    return datasetProcessorFactory;
  }

  public void datasetProcessorFactory(DatasetProcessorFactory f) {
    datasetProcessorFactory = f;
  }

  public ProcessRunnerFactory processRunnerFactory() {
    return processRunnerFactory;
  }

  public void processRunnerFactory(ProcessRunnerFactory f) {
    processRunnerFactory = f;
  }

  private void notifyProcessStart(Dataset d, int pendingCount) {
    for (DatasetProcessServiceListener each : listeners) {
      each.processStart(d, pendingCount);
    }
  }

  private void notifyProcessEnd(DatasetProcessResult r) {
    for (DatasetProcessServiceListener each : listeners) {
      each.processEnd(r);
    }
  }

  private void notifyPostProcessHookStart(Dataset d, ProcessRunner r) {
    for (DatasetProcessServiceListener each : listeners) {
      each.postProcessHookStart(d, r);
    }
  }

  private void notifyPostProcessHookEnd(ProcessRunnerResult r) {
    for (DatasetProcessServiceListener each : listeners) {
      each.postProcessHookEnd(r);
    }
  }

  private boolean addToDatasetQueue(DatasetItem i) {
    boolean added = datasetQueue.add(i);
    if (added && i.unprocessed()) {
      datasetQueueUnprocessedCount.incrementAndGet();
    }
    return added;
  }

  private DatasetItem takeFromDatasetQueue() throws InterruptedException {
    DatasetItem result = datasetQueue.take();
    if (result.unprocessed()) {
      datasetQueueUnprocessedCount.decrementAndGet();
    }
    return result;
  }

  private void runPostProcessHook(Dataset d) {
    ProcessRunner hook = newPostProcessHookRunner(d);
    if (hook == null) return;
    notifyPostProcessHookStart(d, hook);
    hookRunner.submit(hook);
  }

  private ProcessRunner newPostProcessHookRunner(Dataset d) {
    NioProcessBuilder builder = newPostProcessHookBuilder(d);
    if (builder == null) return null;
    return processRunnerFactory.newRunner(builder, GrowingBufferInputStreamConsumer.newFactory(),
        GrowingBufferInputStreamConsumer.newFactory());
  }

  private NioProcessBuilder newPostProcessHookBuilder(Dataset d) {
    if (conf.postProcessHook().isEmpty()) return null;
    Path datasetFile = null;
    try {
      datasetFile = findDatasetFile(d.directory());
    } catch (ClosedByInterruptException | InterruptedIOException e) {
      Thread.currentThread().interrupt();
    } catch (IOException e) {
      exceptionHandler.exception(e);
    }
    if (datasetFile == null) return null;
    String datasetVarValueMatcherQuoted = Matcher.quoteReplacement(datasetFile.toString());
    List<String> commandWithVars = conf.postProcessHook();
    List<String> command = new ArrayList<>(commandWithVars.size());
    StringBuffer expansion = new StringBuffer(128);
    for (String each : commandWithVars) {
      Matcher m = HOOK_VAR.matcher(each);
      expansion.setLength(0);
      while (m.find()) {
        switch (m.group("var")) {
          case "$dataset":
            m.appendReplacement(expansion, datasetVarValueMatcherQuoted);
            break;
          default:
            m.appendReplacement(expansion, Matcher.quoteReplacement(m.group("var")));
            break;
        }
      }
      m.appendTail(expansion);
      command.add(expansion.toString());
    }
    return new DefaultNioProcessBuilder(command);
  }

  private Path findDatasetFile(Path dir) throws IOException {
    Map<String, List<String>> datasets = findDatasets(dir);
    if (datasets.isEmpty()) return null;
    for (List<String> each : datasets.values())
      Collections.sort(each);
    if (datasets.size() == 1) return dir.resolve(datasets.values().iterator().next().get(0));
    List<Path> datasetFirstFiles = new ArrayList<>(datasets.size());
    for (List<String> each : datasets.values())
      datasetFirstFiles.add(dir.resolve(each.get(0)));
    Collections.sort(datasetFirstFiles);
    Path result = datasetFirstFiles.get(0);
    FileTime resultTime = Files.getLastModifiedTime(result, LinkOption.NOFOLLOW_LINKS);
    for (Path each : datasetFirstFiles.subList(1, datasetFirstFiles.size())) {
      FileTime eachTime = Files.getLastModifiedTime(each, LinkOption.NOFOLLOW_LINKS);
      if (eachTime.compareTo(resultTime) <= 0) continue;
      result = each;
      resultTime = eachTime;
    }
    return result;
  }

  private Map<String, List<String>> findDatasets(Path dir) throws IOException {
    try (DirectoryStream<Path> s = Files.newDirectoryStream(dir)) {
      Map<String, List<String>> result = new HashMap<>();
      for (Path each : s) {
        Path eachNamePath = each.getFileName();
        if (eachNamePath == null) continue;
        String eachName = eachNamePath.toString();
        if (!isDatasetImage(eachName)) continue;
        if (!Files.isRegularFile(each, LinkOption.NOFOLLOW_LINKS)) continue;
        String eachBaseName = Utilities.datasetFileBaseName(eachName);
        List<String> d = result.get(eachBaseName);
        if (d == null) {
          d = new ArrayList<>();
          result.put(eachBaseName, d);
        }
        d.add(eachName);
      }
      int minNumImages = conf.datasetMinNumImages();
      for (Iterator<List<String>> i = result.values().iterator(); i.hasNext();) {
        if (i.next().size() < minNumImages) i.remove();
      }
      return result;
    } catch (DirectoryIteratorException e) {
      throw e.getCause();
    }
  }

  private boolean isDatasetImage(String filename) {
    return !conf.diffCenterImagePat().matcher(filename).matches() &&
        conf.imagePat().matcher(filename).matches();
  }

  /*
   * Note: runs in a separate thread.
   */
  private void processDatasets() throws InterruptedException {
    Set<Dataset> processing = new HashSet<>();
    Set<Dataset> needsReprocessing = new HashSet<>();
    for (;;) {
      try {
        DatasetItem i = takeFromDatasetQueue();
        Dataset d = i.dataset();
        if (i.unprocessed()) {
          if (processing.contains(d)) {
            needsReprocessing.add(d);
          } else {
            processing.add(d);
            processor.submit(newProcessCallable(this, d, datasetQueueUnprocessedCount.get()));
          }
        } else {
          boolean needsToBeReprocessed = needsReprocessing.remove(d);
          if (needsToBeReprocessed) {
            processor.submit(newProcessCallable(this, d, datasetQueueUnprocessedCount.get()));
          } else {
            processing.remove(d);
          }
        }
      } catch (InterruptedException e) {
        Thread.currentThread().interrupt();
        throw e;
      } catch (ThreadDeath td) {
        throw td;
      } catch (Throwable t) {
        exceptionHandler.exception(t);
      }
    }
  }

  /*
   * Note: runs in a separate thread.
   */
  private DatasetProcessResult process(Dataset d, int pendingCount) throws InterruptedException {
    try {
      notifyProcessStart(d, pendingCount);
      DatasetProcessor p = datasetProcessorFactory.newProcessor(d, conf, sys);
      DatasetProcessResult r = p.process();
      notifyProcessEnd(r);
      runPostProcessHook(d);
      return r;
    } catch (InterruptedException e) {
      Thread.currentThread().interrupt();
      throw e;
    } catch (ThreadDeath td) {
      throw td;
    } catch (Throwable t) {
      return new DatasetProcessResult(d, t);
    } finally {
      addToDatasetQueue(DatasetItem.processed(d));
    }
  }

  /*
   * Note: runs in a separate thread.
   */
  private void enqueueProcessedResults() throws InterruptedException {
    for (;;) {
      try {
        processed.add(processor.take().get());
      } catch (InterruptedException e) {
        Thread.currentThread().interrupt();
        throw e;
      } catch (ThreadDeath td) {
        throw td;
      } catch (Throwable t) {
        exceptionHandler.exception(t);
      }
    }
  }

  /*
   * Note: runs in a separate thread.
   */
  private void handlePostProcessHookResults() throws InterruptedException {
    for (;;) {
      try {
        notifyPostProcessHookEnd(hookRunner.take().get());
      } catch (InterruptedException e) {
        Thread.currentThread().interrupt();
        throw e;
      } catch (ThreadDeath td) {
        throw td;
      } catch (Throwable t) {
        exceptionHandler.exception(t);
      }
    }
  }

  private static Runnable newProcessDatasetsRunnable(final DatasetProcessService s) {
    return new Runnable() {
      @Override
      public void run() {
        try {
          s.processDatasets();
        } catch (InterruptedException e) {
          Thread.currentThread().interrupt();
        }
      }
    };
  }

  private static Callable<DatasetProcessResult> newProcessCallable(final DatasetProcessService s,
      final Dataset d, final int pendingCount) {
    return new Callable<DatasetProcessResult>() {
      @Override
      public DatasetProcessResult call() throws InterruptedException {
        return s.process(d, pendingCount);
      }
    };
  }

  private static Runnable newEnqueueProcessedResultsRunnable(final DatasetProcessService s) {
    return new Runnable() {
      @Override
      public void run() {
        try {
          s.enqueueProcessedResults();
        } catch (InterruptedException e) {
          Thread.currentThread().interrupt();
        }
      }
    };
  }

  private static Runnable newHandlePostProcessHookResultsRunnable(final DatasetProcessService s) {
    return new Runnable() {
      @Override
      public void run() {
        try {
          s.handlePostProcessHookResults();
        } catch (InterruptedException e) {
          Thread.currentThread().interrupt();
        }
      }
    };
  }

  private static void throwUncheckedException(Throwable t) {
    if (t instanceof RuntimeException) throw (RuntimeException)t;
    if (t instanceof Error) throw (Error)t;
    throw new RuntimeException(t);
  }

  private static class DatasetItem {
    private final Dataset dataset;
    private final boolean processed;

    public DatasetItem(Dataset d, boolean processed) {
      super();
      this.dataset = Objects.requireNonNull(d);
      this.processed = processed;
    }

    public Dataset dataset() {
      return dataset;
    }

    @SuppressWarnings("unused")
    public boolean processed() {
      return processed;
    }

    public boolean unprocessed() {
      return !processed;
    }

    @Override
    public boolean equals(Object o) {
      if (this == o) return true;
      if (!(o instanceof DatasetItem)) return false;
      DatasetItem i = (DatasetItem)o;
      if (processed != i.processed) return false;
      return dataset.equals(i.dataset);
    }

    @Override
    public int hashCode() {
      int result = 17;
      result = 31 * result + dataset.hashCode();
      result = 31 * result + (processed ? 1 : 0);
      return result;
    }

    @Override
    public String toString() {
      return dataset + (processed ? " (processed)" : " (unprocessed)");
    }

    public static DatasetItem unprocessed(Dataset d) {
      return new DatasetItem(d, false);
    }

    public static DatasetItem processed(Dataset d) {
      return new DatasetItem(d, true);
    }
  }

  private static class DefaultDatasetProcessorFactory implements DatasetProcessorFactory {
    @Override
    public DatasetProcessor newProcessor(Dataset dataset, AapConfiguration conf, AaprocSystem sys) {
      return new DatasetProcessor(dataset, conf, sys);
    }
  }

  private static class DefaultProcessRunnerFactory implements ProcessRunnerFactory {
    @Override
    public ProcessRunner newRunner(NioProcessBuilder builder,
        InputStreamConsumerFactory stdoutConsumerFactory,
        InputStreamConsumerFactory stderrConsumerFactory) {
      return new ProcessRunner(builder.toProcessBuilder(), stdoutConsumerFactory,
          stderrConsumerFactory);
    }
  }
}
