/*
 * Copyright (c) 2014, 2015, 2016 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.io.IOException;
import java.nio.file.ClosedWatchServiceException;
import java.nio.file.DirectoryIteratorException;
import java.nio.file.DirectoryStream;
import java.nio.file.FileSystem;
import java.nio.file.Files;
import java.nio.file.LinkOption;
import java.nio.file.NotDirectoryException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardWatchEventKinds;
import java.nio.file.WatchEvent;
import java.nio.file.WatchKey;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.DelayQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import org.imca_cat.pollingwatchservice.EventingWatchService;
import org.imca_cat.pollingwatchservice.PathWatchService;
import org.imca_cat.pollingwatchservice.PollingWatchService;
import org.imca_cat.pollingwatchservice.PollingWatchServiceExceptionHandler;
import org.imca_cat.pollingwatchservice.PollingWatchServiceInfoHandler;

/**
 * Watches for datasets.
 */
class DatasetWatchService implements AutoCloseable {
  private static final WatchEvent.Kind<?>[] WATCH_EVENT_KINDS = new WatchEvent.Kind<?>[] {
      StandardWatchEventKinds.ENTRY_CREATE,
      StandardWatchEventKinds.ENTRY_MODIFY,
      StandardWatchEventKinds.ENTRY_DELETE};

  private final AapConfiguration conf;
  private final DatasetWatchServiceExceptionHandler exceptionHandler;
  private final Set<Path> treeRoots;
  private final BlockingQueue<Path> toBeIndexed;
  private final Object registryLock;
  private final BlockingQueue<Path> toBeRegistered;
  private final BlockingQueue<Path> toBeUnregistered;
  private final BlockingQueue<ImageSet> toBeConsidered;
  private final ExecutorService directoryIndexerExecutor;
  private final ExecutorService directoryRegisterExecutor;
  private final ExecutorService directoryChangeHandlerExecutor;
  private final ExecutorService executor;
  private final DelayQueue<DelayedPathHolder> datasets;
  private final BlockingQueue<DatasetWatchResult> results;
  private final List<DatasetWatchServiceListener> listeners;
  private final PathWatchService fsWatcher;
  private final Object closeLock;
  private final DatasetWatchResult closedSentinel;
  private volatile boolean closed;
  private boolean started;

  /**
   * @throws IOException if will use a {@code WatchService} obtained via the
   *           specified {@code FileSystem} to watch for file system changes
   *           and an I/O error occurs while getting the new
   *           {@code WatchService} instance
   */
  public DatasetWatchService(AapConfiguration conf, FileSystem fs,
      DatasetWatchServiceExceptionHandler exceptionHandler,
      DatasetWatchServiceInfoHandler infoHandler) throws IOException {
    super();
    this.conf = conf;
    this.exceptionHandler = exceptionHandler;
    treeRoots = Collections.newSetFromMap(
        new ConcurrentHashMap<Path, Boolean>(16, 0.75f, conf.maxNumIndexJobs()));
    toBeIndexed = new LinkedBlockingQueue<>();
    registryLock = new Object();
    toBeRegistered = new LinkedBlockingQueue<>();
    toBeUnregistered = new LinkedBlockingQueue<>();
    toBeConsidered = new LinkedBlockingQueue<>();
    directoryIndexerExecutor = new ThreadPoolExecutor(0, conf.maxNumIndexJobs(), 15L,
        TimeUnit.MINUTES, new LinkedBlockingQueue<Runnable>());
    directoryRegisterExecutor = new ThreadPoolExecutor(0, conf.maxNumRegisterJobs(), 15L,
        TimeUnit.MINUTES, new LinkedBlockingQueue<Runnable>());
    directoryChangeHandlerExecutor = new ThreadPoolExecutor(0, conf.maxNumDirChangeJobs(), 15L,
        TimeUnit.MINUTES, new LinkedBlockingQueue<Runnable>());
    executor = Executors.newCachedThreadPool();
    datasets = new DelayQueue<>();
    results = new LinkedBlockingQueue<>();
    listeners = new CopyOnWriteArrayList<>();
    if (conf.fileSystemPollEnabled()) {
      PollingWatchService pfsWatcher = new PollingWatchService(conf.fileSystemPollMaxNumJobs(),
          conf.fileSystemPollPeriod().amount(), conf.fileSystemPollPeriod().unit(),
          newPollingWatchServiceExceptionHandler(exceptionHandler),
          newPollingWatchServiceInfoHandler(infoHandler));
      pfsWatcher.useFileMetadata(conf.fileSystemPollMetadataEnabled());
      fsWatcher = pfsWatcher;
    } else {
      fsWatcher = new EventingWatchService(fs.newWatchService());
    }
    closeLock = new Object();
    closedSentinel = new DatasetWatchResult(new Dataset(Paths.get("")));
    closed = false;
    started = false;
  }

  /**
   * @throws ClosedServiceException if this service is closed
   * @throws IllegalStateException if this service is already started
   */
  public void start() {
    synchronized (closeLock) {
      if (closed) throw new ClosedServiceException();
      if (started) throw new IllegalStateException("Already started");
      started = true;
      executor.execute(newDirectoriesIndexer(this));
      executor.execute(newDirectoriesChangeHandler(this));
      executor.execute(newImageSetsConsiderer(this));
      executor.execute(newSettledDatasetResultEnqueuer(this));
      executor.execute(newDirectoriesRegisterer(this));
      executor.execute(newDirectoriesUnregisterer(this));
      fsWatcher.start();
    }
  }

  /**
   * Registers directory tree rooted at {@code dir} to be watched.
   *
   * @throws ClosedServiceException if this service is closed
   * @throws NotDirectoryException if {@code dir} is not a directory
   */
  public void register(Path dir) throws NotDirectoryException {
    if (closed) throw new ClosedServiceException();
    /*
     * This is obviously vulnerable to a TOCTOU attack, but it will be checked
     * later when it is actually indexed, and we do not support operating on
     * unsafe directories anyway.
     */
    if (!Files.isDirectory(dir)) {
      throw new NotDirectoryException(dir.toString());
    }
    addTreeRoot(dir);
    toBeIndexed.add(dir);
  }

  /**
   * Registers directory trees rooted at each element in {@code c} to be
   * watched.
   *
   * @throws ClosedServiceException if this service is closed
   * @throws NotDirectoryException if any element in {@code c} is not a
   *           directory
   */
  public void registerAll(Collection<Path> c) throws NotDirectoryException {
    if (closed) throw new ClosedServiceException();
    /*
     * This is obviously vulnerable to a TOCTOU attack, but they will be
     * checked later when they are actually indexed, and we do not support
     * operating on unsafe directories anyway.
     */
    for (Path each : c) {
      if (!Files.isDirectory(each)) {
        throw new NotDirectoryException(each.toString());
      }
    }
    addTreeRoots(c);
    toBeIndexed.addAll(c);
  }

  /**
   * @throws ClosedServiceException if this service is closed
   */
  public DatasetWatchResult take() throws InterruptedException {
    if (closed) throw new ClosedServiceException();
    DatasetWatchResult result = results.take();
    if (result == closedSentinel) throw new ClosedServiceException();
    return result;
  }

  @Override
  public void close() throws IOException {
    synchronized (closeLock) {
      if (closed) return;
      closed = true;
      Throwable t = null;
      boolean interrupted = Thread.interrupted();
      for (ExecutorService each : Arrays.asList(directoryChangeHandlerExecutor,
          directoryIndexerExecutor, executor)) {
        try {
          each.shutdownNow();
          for (;;) {
            try {
              each.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);
              break;
            } catch (InterruptedException e) {
              interrupted = true;
            }
          }
        } catch (ThreadDeath td) {
          t = td;
        } catch (Throwable th) {
          if (t == null) {
            t = th;
          } else {
            t.addSuppressed(th);
          }
        }
      }
      try {
        fsWatcher.close();
      } catch (ThreadDeath td) {
        t = td;
      } catch (Throwable th) {
        if (t == null) {
          t = th;
        } else {
          t.addSuppressed(th);
        }
      }
      try {
        results.clear();
        results.add(closedSentinel);
      } catch (ThreadDeath td) {
        t = td;
      } catch (Throwable th) {
        if (t == null) {
          t = th;
        } else {
          t.addSuppressed(th);
        }
      }
      for (Collection<?> each : Arrays.asList(treeRoots, toBeIndexed, toBeRegistered,
          toBeUnregistered, toBeConsidered, datasets, listeners)) {
        try {
          each.clear();
        } catch (ThreadDeath td) {
          t = td;
        } catch (Throwable th) {
          if (t == null) {
            t = th;
          } else {
            t.addSuppressed(th);
          }
        }
      }
      if (interrupted) Thread.currentThread().interrupt();
      if (t != null) throwUncheckedException(t);
    }
  }

  public void addListener(DatasetWatchServiceListener l) {
    listeners.add(l);
  }

  public void removeListener(DatasetWatchServiceListener l) {
    listeners.remove(l);
  }

  private void notifyIndexStart(Path p) {
    for (DatasetWatchServiceListener each : listeners) {
      each.indexStart(p);
    }
  }

  private void notifyIndexEnd(DirectoryIndexResult r) {
    for (DatasetWatchServiceListener each : listeners) {
      each.indexEnd(r);
    }
  }

  /*
   * Note: runs in a single separate thread.
   */
  private void indexDirectories() throws InterruptedException {
    for (;;) {
      try {
        directoryIndexerExecutor.execute(newDirectoryIndexer(this, toBeIndexed.take()));
      } catch (InterruptedException e) {
        Thread.currentThread().interrupt();
        throw e;
      } catch (ThreadDeath td) {
        throw td;
      } catch (Throwable t) {
        exceptionHandler.exception(t);
      }
    }
  }

  /*
   * Note: runs in multiple separate threads.
   */
  private void indexDirectory(Path dir) throws InterruptedException {
    /*
     * A secure directory [1] (i.e. one in which there is no risk of a file
     * system attack by a malicious user) is assumed.
     *
     * [1] https://www.securecoding.cert.org/confluence/x/ioHWAw
     */
    try {
      if (!Files.isDirectory(dir)) return;
      notifyIndexStart(dir);
      boolean succeeded = true;
      Map<String, Set<Path>> images = new HashMap<>();
      try (DirectoryStream<Path> s = Files.newDirectoryStream(dir)) {
        for (Path each : s) {
          Path eachNamePath = each.getFileName();
          if (eachNamePath == null) continue;
          String eachName = eachNamePath.toString();
          if (!isProcessingDirectory(eachName) && !isExcludedDirectory(eachName) &&
              Files.isDirectory(each, LinkOption.NOFOLLOW_LINKS)) {
            toBeIndexed.put(each);
          } else if (isDatasetImage(eachName) &&
              Files.isRegularFile(each, LinkOption.NOFOLLOW_LINKS)) {
            String eachBaseName = datasetImageBaseName(eachName);
            Set<Path> imagesWithSameBaseName = images.get(eachBaseName);
            if (imagesWithSameBaseName == null) {
              imagesWithSameBaseName = new HashSet<>();
              images.put(eachBaseName, imagesWithSameBaseName);
            }
            imagesWithSameBaseName.add(eachNamePath);
          }
        }
      } catch (DirectoryIteratorException e) {
        succeeded = false;
        exceptionHandler.exception(e.getCause());
      } catch (IOException e) {
        succeeded = false;
        exceptionHandler.exception(e);
      }
      toBeRegistered.put(dir);
      for (Set<Path> each : images.values()) {
        toBeConsidered.put(new ImageSet(dir, each));
      }
      notifyIndexEnd(new DirectoryIndexResult(succeeded, dir));
    } catch (InterruptedException e) {
      Thread.currentThread().interrupt();
      throw e;
    } catch (ThreadDeath td) {
      throw td;
    } catch (Throwable t) {
      exceptionHandler.exception(t);
    }
  }

  /*
   * Note: runs in a single separate thread.
   */
  private void handleDirectoryChanges() throws InterruptedException {
    for (;;) {
      try {
        directoryChangeHandlerExecutor.execute(newDirectoryChangeHandler(this, fsWatcher.take()));
      } catch (InterruptedException e) {
        Thread.currentThread().interrupt();
        throw e;
      } catch (ClosedWatchServiceException e) {
        return;
      } catch (ThreadDeath td) {
        throw td;
      } catch (Throwable t) {
        exceptionHandler.exception(t);
      }
    }
  }

  /*
   * Note: runs in multiple separate threads.
   */
  private void handleDirectoryChange(WatchKey k) throws InterruptedException {
    try {
      Path dir = (Path)k.watchable();
      Map<String, Set<Path>> images = new HashMap<>();
      boolean dirIndexNeeded = false;
      boolean valid;
      try {
        for (WatchEvent<?> each : k.pollEvents()) {
          WatchEvent.Kind<?> eachKind = each.kind();
          if (eachKind == StandardWatchEventKinds.OVERFLOW) {
            dirIndexNeeded = true;
            continue;
          }
          Path eachItem = (Path)each.context();
          if (eachItem == null) continue;
          Path eachPath = dir.resolve(eachItem);
          Path eachNamePath = eachPath.getFileName();
          if (eachNamePath == null) continue;
          String eachName = eachNamePath.toString();
          if (eachKind == StandardWatchEventKinds.ENTRY_CREATE) {
            if (!isProcessingDirectory(eachName) && !isExcludedDirectory(eachName) &&
                Files.isDirectory(eachPath, LinkOption.NOFOLLOW_LINKS)) {
              toBeIndexed.put(eachPath);
            } else if (isDatasetImage(eachName) &&
                Files.isRegularFile(eachPath, LinkOption.NOFOLLOW_LINKS)) {
              String eachBaseName = datasetImageBaseName(eachName);
              Set<Path> imagesWithSameBaseName = images.get(eachBaseName);
              if (imagesWithSameBaseName == null) {
                imagesWithSameBaseName = new HashSet<>();
                images.put(eachBaseName, imagesWithSameBaseName);
              }
              imagesWithSameBaseName.add(eachNamePath);
            }
          } else if (eachKind == StandardWatchEventKinds.ENTRY_MODIFY) {
            if (isDatasetImage(eachName) &&
                Files.isRegularFile(eachPath, LinkOption.NOFOLLOW_LINKS)) {
              String eachBaseName = datasetImageBaseName(eachName);
              Set<Path> imagesWithSameBaseName = images.get(eachBaseName);
              if (imagesWithSameBaseName == null) {
                imagesWithSameBaseName = new HashSet<>();
                images.put(eachBaseName, imagesWithSameBaseName);
              }
              imagesWithSameBaseName.add(eachNamePath);
            }
          }
        }
      } finally {
        valid = k.reset();
      }
      if (valid) {
        if (dirIndexNeeded) toBeIndexed.put(dir);
      } else {
        toBeUnregistered.put(dir);
      }
      for (Set<Path> each : images.values()) {
        toBeConsidered.put(new ImageSet(dir, each));
      }
    } catch (InterruptedException e) {
      Thread.currentThread().interrupt();
      throw e;
    } catch (ThreadDeath td) {
      throw td;
    } catch (Throwable t) {
      exceptionHandler.exception(t);
    }
  }

  /*
   * Note: runs in a single separate thread.
   */
  private void considerImageSets() throws InterruptedException {
    final Map<Path, Map<String, ImageSet>> imageSets = new HashMap<>();
    for (;;) {
      try {
        ImageSet newOrChanged = toBeConsidered.take();
        Path dir = newOrChanged.directory();
        Map<String, ImageSet> imageSetsByBaseName = imageSets.get(dir);
        if (imageSetsByBaseName == null) {
          imageSetsByBaseName = new HashMap<>();
          imageSets.put(dir, imageSetsByBaseName);
        }
        String imageBaseName = datasetImageBaseName(newOrChanged);
        ImageSet current = imageSetsByBaseName.get(imageBaseName);
        current = (current == null) ? newOrChanged : current.merge(newOrChanged);
        imageSetsByBaseName.put(imageBaseName, current);
        if (isDataset(current)) {
          DelayedPathHolder dd = newDelayedPathHolder(dir);
          datasets.remove(dd);
          datasets.add(dd);
        }
      } catch (InterruptedException e) {
        Thread.currentThread().interrupt();
        throw e;
      } catch (ThreadDeath td) {
        throw td;
      } catch (Throwable t) {
        exceptionHandler.exception(t);
      }
    }
  }

  /*
   * Note: runs in a single separate thread.
   */
  private void registerDirectories() throws InterruptedException {
    for (;;) {
      try {
        directoryRegisterExecutor.execute(newDirectoryRegisterer(this, toBeRegistered.take()));
      } catch (InterruptedException e) {
        Thread.currentThread().interrupt();
        throw e;
      } catch (ThreadDeath td) {
        throw td;
      } catch (Throwable t) {
        exceptionHandler.exception(t);
      }
    }
  }

  /*
   * Note: runs in multiple separate threads.
   */
  private void registerDirectory(Path dir) throws InterruptedException {
    /*
     * Do not synchronize on registryLock here because it is not needed and
     * because it allows multiple registry threads of execution to execute
     * concurrently. This is important for the case where a PollingWatchService
     * is being used since the PollingWatchService.register method call will
     * result in a directory entries read which may be time consuming.
     */
    try {
      fsWatcher.register(dir, WATCH_EVENT_KINDS);
    } catch (IOException e) {
      unregisterDirectory(dir);
    } catch (ThreadDeath td) {
      throw td;
    } catch (Throwable t) {
      exceptionHandler.exception(t);
    }
  }

  /*
   * Note: runs in a single separate thread.
   */
  private void unregisterDirectories() throws InterruptedException {
    for (;;) {
      try {
        unregisterDirectory(toBeUnregistered.take());
      } catch (InterruptedException e) {
        Thread.currentThread().interrupt();
        throw e;
      } catch (ThreadDeath td) {
        throw td;
      } catch (Throwable t) {
        exceptionHandler.exception(t);
      }
    }
  }

  private void unregisterDirectory(Path dir) throws InterruptedException {
    synchronized (registryLock) {
      if (removeTreeRoot(dir)) {
        results.put(new DatasetWatchResult(dir));
      }
    }
  }

  private boolean addTreeRoots(Collection<Path> dirs) {
    synchronized (registryLock) {
      return treeRoots.addAll(dirs);
    }
  }

  private boolean addTreeRoot(Path dir) {
    synchronized (registryLock) {
      return treeRoots.add(dir);
    }
  }

  private boolean removeTreeRoot(Path dir) {
    synchronized (registryLock) {
      return treeRoots.remove(dir);
    }
  }

  /*
   * Note: runs in a single separate thread.
   */
  private void enqueueSettledDatasetResults() throws InterruptedException {
    for (;;) {
      try {
        results.put(new DatasetWatchResult(new Dataset(datasets.take().path())));
      } catch (InterruptedException e) {
        Thread.currentThread().interrupt();
        throw e;
      } catch (ThreadDeath td) {
        throw td;
      } catch (Throwable t) {
        exceptionHandler.exception(t);
      }
    }
  }

  private boolean isDataset(ImageSet s) {
    return s.size() >= conf.datasetMinNumImages();
  }

  private boolean isExcludedDirectory(String name) {
    return conf.excludeDirPat().matcher(name).matches();
  }

  private boolean isProcessingDirectory(String name) {
    return conf.processingDirPat().matcher(name).matches();
  }

  private boolean isDatasetImage(String name) {
    return !conf.diffCenterImagePat().matcher(name).matches() &&
        conf.imagePat().matcher(name).matches();
  }

  private DelayedPathHolder newDelayedPathHolder(Path dir) {
    return new DelayedPathHolder(dir, conf.datasetSettleTime().amount(),
        conf.datasetSettleTime().unit());
  }

  private static PollingWatchServiceExceptionHandler newPollingWatchServiceExceptionHandler(
      final DatasetWatchServiceExceptionHandler handler) {
    return new PollingWatchServiceExceptionHandler() {
      @Override
      public void exception(Throwable t) {
        handler.exception(t);
      }
    };
  }

  private static PollingWatchServiceInfoHandler newPollingWatchServiceInfoHandler(
      final DatasetWatchServiceInfoHandler handler) {
    return new PollingWatchServiceInfoHandler() {
      @Override
      public void pollTime(long t) {
        handler.pollTime(t);
      }
    };
  }

  private static Runnable newDirectoriesIndexer(final DatasetWatchService s) {
    return new Runnable() {
      @Override
      public void run() {
        try {
          s.indexDirectories();
        } catch (InterruptedException e) {
          Thread.currentThread().interrupt();
        }
      }
    };
  }

  private static Runnable newDirectoryIndexer(final DatasetWatchService s, final Path d) {
    return new Runnable() {
      @Override
      public void run() {
        try {
          s.indexDirectory(d);
        } catch (InterruptedException e) {
          Thread.currentThread().interrupt();
        }
      }
    };
  }

  private static Runnable newDirectoriesChangeHandler(final DatasetWatchService s) {
    return new Runnable() {
      @Override
      public void run() {
        try {
          s.handleDirectoryChanges();
        } catch (InterruptedException e) {
          Thread.currentThread().interrupt();
        }
      }
    };
  }

  private static Runnable newDirectoryChangeHandler(final DatasetWatchService s, final WatchKey k) {
    return new Runnable() {
      @Override
      public void run() {
        try {
          s.handleDirectoryChange(k);
        } catch (InterruptedException e) {
          Thread.currentThread().interrupt();
        }
      }
    };
  }

  private static Runnable newImageSetsConsiderer(final DatasetWatchService s) {
    return new Runnable() {
      @Override
      public void run() {
        try {
          s.considerImageSets();
        } catch (InterruptedException e) {
          Thread.currentThread().interrupt();
        }
      }
    };
  }

  private static Runnable newSettledDatasetResultEnqueuer(final DatasetWatchService s) {
    return new Runnable() {
      @Override
      public void run() {
        try {
          s.enqueueSettledDatasetResults();
        } catch (InterruptedException e) {
          Thread.currentThread().interrupt();
        }
      }
    };
  }

  private static Runnable newDirectoriesRegisterer(final DatasetWatchService s) {
    return new Runnable() {
      @Override
      public void run() {
        try {
          s.registerDirectories();
        } catch (InterruptedException e) {
          Thread.currentThread().interrupt();
        }
      }
    };
  }

  private static Runnable newDirectoryRegisterer(final DatasetWatchService s, final Path d) {
    return new Runnable() {
      @Override
      public void run() {
        try {
          s.registerDirectory(d);
        } catch (InterruptedException e) {
          Thread.currentThread().interrupt();
        }
      }
    };
  }

  private static Runnable newDirectoriesUnregisterer(final DatasetWatchService s) {
    return new Runnable() {
      @Override
      public void run() {
        try {
          s.unregisterDirectories();
        } catch (InterruptedException e) {
          Thread.currentThread().interrupt();
        }
      }
    };
  }

  private static void throwUncheckedException(Throwable t) {
    if (t instanceof RuntimeException) throw (RuntimeException)t;
    if (t instanceof Error) throw (Error)t;
    throw new RuntimeException(t);
  }

  private static String datasetImageBaseName(ImageSet s) {
    if (s == null) return "";
    if (s.isEmpty()) return "";
    Path p = s.iterator().next().getFileName();
    if (p == null) return "";
    return datasetImageBaseName(p.toString());
  }

  private static String datasetImageBaseName(String imageFileName) {
    if (imageFileName == null) return "";
    return Utilities.datasetFileBaseName(imageFileName);
  }
}
