/*
 * Copyright (c) 2015 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.io.IOException;
import java.lang.ProcessBuilder.Redirect.Type;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Objects;

/**
 * An {@code NioProcessBuilder} that delegates to a {@code ProcessBuilder}. The
 * {@code environment()} method returns the environment of the
 * {@code ProcessBuilder}, and the {@code redirectErrorStream} methods delegate
 * to the {@code ProcessBuilder}. The rest of the methods, with the exception
 * of {@code start()} and {@code toProcessBuilder()}, do <em>not</em> delegate
 * to the {@code ProcessBuilder} when they are called; instead, their
 * delegation is delayed until the {@code start()} or
 * {@code toProcessBuilder()} methods are called.
 */
class DefaultNioProcessBuilder implements NioProcessBuilder {
  private final ProcessBuilder delegate = new ProcessBuilder();
  private List<String> command = null;
  private Path directory = null;
  private NioProcessBuilder.Redirect stdinRedirect = DefaultRedirect.PIPE;
  private NioProcessBuilder.Redirect stdoutRedirect = DefaultRedirect.PIPE;
  private NioProcessBuilder.Redirect stderrRedirect = DefaultRedirect.PIPE;

  public DefaultNioProcessBuilder(List<String> command) {
    super();
    this.command = command;
  }

  public DefaultNioProcessBuilder(String... command) {
    super();
    this.command = new ArrayList<>(command.length);
    for (String each : command) {
      this.command.add(each);
    }
  }

  @Override
  public List<String> command() {
    return command;
  }

  @Override
  public NioProcessBuilder command(List<String> command) {
    this.command = command;
    return this;
  }

  @Override
  public NioProcessBuilder command(String... command) {
    this.command = new ArrayList<>(command.length);
    for (String each : command) {
      this.command.add(each);
    }
    return this;
  }

  @Override
  public Path directory() {
    return directory;
  }

  @Override
  public NioProcessBuilder directory(Path directory) {
    this.directory = directory;
    return this;
  }

  @Override
  public Map<String, String> environment() {
    return delegate.environment();
  }

  @Override
  public NioProcessBuilder inheritIO() {
    stdinRedirect = DefaultRedirect.INHERIT;
    stdoutRedirect = DefaultRedirect.INHERIT;
    stderrRedirect = DefaultRedirect.INHERIT;
    return this;
  }

  @Override
  public NioProcessBuilder.Redirect redirectError() {
    return stderrRedirect;
  }

  @Override
  public NioProcessBuilder redirectError(Path file) {
    return redirectError(DefaultRedirect.to(file));
  }

  @Override
  public NioProcessBuilder redirectError(NioProcessBuilder.Redirect destination) {
    stderrRedirect = destination;
    return this;
  }

  @Override
  public boolean redirectErrorStream() {
    return delegate.redirectErrorStream();
  }

  @Override
  public NioProcessBuilder redirectErrorStream(boolean redirectErrorStream) {
    delegate.redirectErrorStream(redirectErrorStream);
    return this;
  }

  @Override
  public NioProcessBuilder.Redirect redirectInput() {
    return stdinRedirect;
  }

  @Override
  public NioProcessBuilder redirectInput(Path file) {
    return redirectInput(DefaultRedirect.from(file));
  }

  @Override
  public NioProcessBuilder redirectInput(NioProcessBuilder.Redirect source) {
    stdinRedirect = source;
    return this;
  }

  @Override
  public NioProcessBuilder.Redirect redirectOutput() {
    return stdoutRedirect;
  }

  @Override
  public NioProcessBuilder redirectOutput(Path file) {
    return redirectOutput(DefaultRedirect.to(file));
  }

  @Override
  public NioProcessBuilder redirectOutput(NioProcessBuilder.Redirect destination) {
    stdoutRedirect = destination;
    return this;
  }

  @Override
  public Process start() throws IOException {
    return toProcessBuilder().start();
  }

  @Override
  public ProcessBuilder toProcessBuilder() {
    delegate.command(command);
    if (directory != null) delegate.directory(directory.toFile());
    delegate.redirectInput(toProcessBuilderRedirect(stdinRedirect));
    delegate.redirectOutput(toProcessBuilderRedirect(stdoutRedirect));
    delegate.redirectError(toProcessBuilderRedirect(stderrRedirect));
    return delegate;
  }

  private static ProcessBuilder.Redirect toProcessBuilderRedirect(NioProcessBuilder.Redirect r) {
    return (r == null) ? null : r.toProcessBuilderRedirect();
  }

  /**
   * An {@code NioProcessBuilder.Redirect} like {@code ProcessBuilder.Redirect}
   * suitable for use with {@code DefaultNioProcessBuilder}.
   */
  static class DefaultRedirect implements NioProcessBuilder.Redirect {
    public static final DefaultRedirect INHERIT = new DefaultRedirect(Type.INHERIT);
    public static final DefaultRedirect PIPE = new DefaultRedirect(Type.PIPE);

    private final Type type;
    private final Path file;

    protected DefaultRedirect(Type type) {
      this(type, null);
    }

    protected DefaultRedirect(Type type, Path file) {
      super();
      this.type = Objects.requireNonNull(type);
      this.file = file;
    }

    @Override
    public Type type() {
      return type;
    }

    @Override
    public Path file() {
      return file;
    }

    @Override
    public boolean equals(Object o) {
      if (o == this) return true;
      if (!(o instanceof DefaultRedirect)) return false;
      DefaultRedirect r = (DefaultRedirect)o;
      if (!type.equals(r.type)) return false;
      return Objects.equals(file, r.file);
    }

    @Override
    public int hashCode() {
      int result = 17;
      result = 31 * result + type.hashCode();
      result = 31 * result + ((file == null) ? 0 : file.hashCode());
      return result;
    }

    @Override
    public ProcessBuilder.Redirect toProcessBuilderRedirect() {
      switch (type) {
        case APPEND:
          return ProcessBuilder.Redirect.appendTo(file.toFile());
        case INHERIT:
          return ProcessBuilder.Redirect.INHERIT;
        case PIPE:
          return ProcessBuilder.Redirect.PIPE;
        case READ:
          return ProcessBuilder.Redirect.from(file.toFile());
        case WRITE:
          return ProcessBuilder.Redirect.to(file.toFile());
        default:
          throw new AssertionError("Unknown type: " + type);
      }
    }

    public static DefaultRedirect appendTo(Path file) {
      return new DefaultRedirect(Type.APPEND, file);
    }

    public static DefaultRedirect from(Path file) {
      return new DefaultRedirect(Type.READ, file);
    }

    public static DefaultRedirect to(Path file) {
      return new DefaultRedirect(Type.WRITE, file);
    }
  }
}
