/*
 * Copyright (c) 2014 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.nio.file.Path;
import java.util.Objects;
import java.util.concurrent.Delayed;
import java.util.concurrent.TimeUnit;

/**
 * A {@code Delayed} implementation that holds a {@code Path}.
 *
 * Note: this class has a natural ordering that is inconsistent with equals.
 */
class DelayedPathHolder implements Delayed {
  private final Path path;
  private final long end;

  public DelayedPathHolder(Path path, long delay, TimeUnit delayUnit) {
    this(path, System.nanoTime() + delayUnit.toNanos(delay));
  }

  public DelayedPathHolder(Path path, long end) {
    super();
    this.path = Objects.requireNonNull(path);
    this.end = end;
  }

  public Path path() {
    return path;
  }

  /**
   * Indicates whether this object is equal to the specified object.
   * <p>
   * The delay is ignored when testing for equality; only the paths of this
   * object and the specified object are compared.
   *
   * @param o object to compare with this object for equality
   *
   * @return true if this object is equal to the specified object, ignoring the
   *         delay; false otherwise
   */
  @Override
  public boolean equals(Object o) {
    if (this == o) return true;
    if (!(o instanceof DelayedPathHolder)) return false;
    return path.equals(((DelayedPathHolder)o).path);
  }

  @Override
  public int hashCode() {
    int result = 17;
    result = 31 * result + path.hashCode();
    return result;
  }

  @Override
  public int compareTo(Delayed d) {
    if (d instanceof DelayedPathHolder) {
      return Long.compare(end, ((DelayedPathHolder)d).end);
    } else {
      return Long.compare(getDelay(TimeUnit.NANOSECONDS), d.getDelay(TimeUnit.NANOSECONDS));
    }
  }

  @Override
  public long getDelay(TimeUnit unit) {
    return unit.convert(end - System.nanoTime(), TimeUnit.NANOSECONDS);
  }

  @Override
  public String toString() {
    return "path=" + path.toString() + ", end=" + end;
  }
}
