/*
 * Copyright (c) 2014 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.nio.file.Path;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Objects;
import java.util.Set;

/**
 * An immutable set of images.
 */
class ImageSet implements Set<Path> {
  private final Path dir;
  private final Set<Path> images;

  /**
   * Constructs an instance with the specified directory and images.
   *
   * @param dir directory containing images
   * @param images images in {@code dir} relative to {@code dir}; the caller
   *          must not modify {@code images} once this call is made
   */
  public ImageSet(Path dir, Set<Path> images) {
    super();
    this.dir = Objects.requireNonNull(dir);
    this.images = Collections.unmodifiableSet(Objects.requireNonNull(images));
  }

  public Path directory() {
    return dir;
  }

  @Override
  public boolean equals(Object o) {
    if (o == this) return true;
    if (!(o instanceof ImageSet)) return false;
    ImageSet s = (ImageSet)o;
    if (!dir.equals(s.dir)) return false;
    return images.equals(s.images);
  }

  @Override
  public int hashCode() {
    int result = 17;
    result = 31 * result + dir.hashCode();
    result = 31 * result + images.hashCode();
    return result;
  }

  @Override
  public int size() {
    return images.size();
  }

  @Override
  public boolean isEmpty() {
    return images.isEmpty();
  }

  @Override
  public boolean contains(Object o) {
    return images.contains(o);
  }

  @Override
  public Iterator<Path> iterator() {
    return images.iterator();
  }

  @Override
  public Object[] toArray() {
    return images.toArray();
  }

  @Override
  public <T> T[] toArray(T[] a) {
    return images.toArray(a);
  }

  @Override
  public boolean add(Path e) {
    return images.add(e);
  }

  @Override
  public boolean remove(Object o) {
    return images.remove(o);
  }

  @Override
  public boolean containsAll(Collection<?> c) {
    return images.containsAll(c);
  }

  @Override
  public boolean addAll(Collection<? extends Path> c) {
    return images.addAll(c);
  }

  @Override
  public boolean retainAll(Collection<?> c) {
    return images.retainAll(c);
  }

  @Override
  public boolean removeAll(Collection<?> c) {
    return images.removeAll(c);
  }

  @Override
  public void clear() {
    images.clear();
  }

  @Override
  public String toString() {
    StringBuilder result = new StringBuilder();
    result.append("dir=");
    result.append(dir);
    result.append(", images=");
    if (images.size() > 1000) {
      result.append("(");
      result.append(images.size());
      result.append(" items)");
    } else {
      result.append(images);
    }
    return result.toString();
  }

  /**
   * Creates a new {@code ImageSet} from the images of this set and the images
   * of the specified sets.
   *
   * @param sets image sets to merge with this
   *
   * @throws IllegalArgumentException if the directories of the
   *           {@code ImageSet} instances are not equal
   */
  public ImageSet merge(ImageSet... sets) {
    Set<Path> mergedImages = new HashSet<>(images);
    for (ImageSet each : sets) {
      if (!dir.equals(each.directory())) {
        throw new IllegalArgumentException("Directories not equal");
      }
      mergedImages.addAll(each);
    }
    return new ImageSet(dir, mergedImages);
  }

  public Dataset toDataset() {
    return new Dataset(dir);
  }
}
