/*
 * Copyright (c) 2014, 2015 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.io.IOException;
import java.nio.file.FileVisitResult;
import java.nio.file.Files;
import java.nio.file.LinkOption;
import java.nio.file.NoSuchFileException;
import java.nio.file.Path;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.attribute.BasicFileAttributes;

/**
 * Static utility methods to augment the core Java library.
 */
class JavaUtilities {
  private JavaUtilities() {
    throw new AssertionError();
  }

  /**
   * Deletes the specified file or directory. If the path specifies a
   * directory, the entire tree rooted at the directory will be deleted.
   * Symbolic links are not followed.
   * <p>
   * Note: this method is only safe when operating on a secure directory (i.e.
   * one in which there is no risk of a file system attack by a malicious
   * user).
   * <p>
   * According to the CERT Secure Coding website, rule
   * <a href="https://www.securecoding.cert.org/confluence/x/ioHWAw">FIO00-J.
   * Do not operate on files in shared directories</a> (accessed on July 1,
   * 2014):
   * <q>A directory is secure with respect to a particular user if only the
   * user and the system administrator are allowed to create, move, or delete
   * files inside the directory. Furthermore, each parent directory must itself
   * be a secure directory up to and including the root directory. On most
   * systems, home or user directories are secure by default and only shared
   * directories are insecure.</q>
   */
  public static void deleteRecursively(Path p) throws IOException {
    if (!Files.exists(p, LinkOption.NOFOLLOW_LINKS)) return;
    Files.walkFileTree(p, new SimpleFileVisitor<Path>() {
      @Override
      public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
        try {
          Files.delete(file);
        } catch (NoSuchFileException e) {
          /* IGNORED */
        }
        return FileVisitResult.CONTINUE;
      }

      @Override
      public FileVisitResult postVisitDirectory(Path dir, IOException exc) throws IOException {
        if (exc != null) throw exc;
        try {
          Files.delete(dir);
        } catch (NoSuchFileException e) {
          /* IGNORED */
        }
        return FileVisitResult.CONTINUE;
      }
    });
  }

  /**
   * Joins specified strings separated by the specified separator.
   */
  public static String join(String sep, String first, String... more) {
    if (more.length == 0) return first;
    StringBuilder b = new StringBuilder();
    b.append(first);
    for (String each : more) {
      b.append(sep);
      b.append(each);
    }
    return b.toString();
  }

  /**
   * Escapes special HTML characters in the specified string thus making the
   * result suitable for inclusion in an HTML document.
   * <p>
   * This method escapes the following characters:
   *
   * <pre>    &amp;"'`&lt;&gt;</pre>
   * <p>
   * To safely include the escaped string in an HTML document, the following
   * conditions must be met:
   * <ol>
   * <li>The document must quote all attribute values where the escaped string
   * may be used.
   * <li>The escaped string must not be used inside a {@code <script>} element.
   * <li>The escaped string must not be used inside JavaScript code.
   * <li>The HTML document must specify a character encoding. This is done with
   * something like {@code
   * <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">} in
   * the {@code <head>} element.
   * <li>If the HTML document is being served by a web server, the web server
   * must report the same character encoding as that specified in the HTML
   * document.
   * </ol>
   * <p>
   * Note: this method is <em>not</em> for encoding a string in a URL. To do
   * that (i.e. encode), you should use a {@code URI} and convert it to a
   * {@code URL}. To decode, covert the {@code URL} to a {@code URI}.
   *
   * @see java.net.URI
   * @see java.net.URL
   * @see <a href="http://wonko.com/post/html-escaping">There's more to HTML
   *      escaping than &amp;, &lt;, &gt;, and "</a>
   * @see <a href=
   *      "https://www.owasp.org/index.php/XSS_(Cross_Site_Scripting)_Prevention_Cheat_Sheet">
   *      XSS (Cross Site Scripting) Prevention Cheat Sheet</a>
   */
  public static String escapeHtml(String s) {
    if (s == null) return s;
    return s.replace("&", "&amp;").replace("\"", "&quot;").replace("'", "&apos;").replace("`",
        "&#96;").replace("<", "&lt;").replace(">", "&gt;");
  }
}
