/*
 * Copyright (c) 2015 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.io.IOException;
import java.nio.file.Path;
import java.util.List;
import java.util.Map;

/**
 * A process builder like {@link ProcessBuilder} that uses NIO {@code Path}
 * objects instead of {@code File} objects.
 *
 * {@code NioProcessBuilder} is also more useful than {@code ProcessBuilder}
 * because it is an interface and hence easy to create implementations for it
 * with different behavior (e.g. mocking for testing). This is in contrast to
 * {@code ProcessBuilder} which is a class declared {@code final} and hence
 * provides no nice way to alter its behavior.
 */
interface NioProcessBuilder {
  List<String> command();

  NioProcessBuilder command(List<String> command);

  NioProcessBuilder command(String... command);

  Path directory();

  NioProcessBuilder directory(Path directory);

  Map<String, String> environment();

  NioProcessBuilder inheritIO();

  Redirect redirectError();

  NioProcessBuilder redirectError(Path file);

  NioProcessBuilder redirectError(Redirect destination);

  boolean redirectErrorStream();

  NioProcessBuilder redirectErrorStream(boolean redirectErrorStream);

  Redirect redirectInput();

  NioProcessBuilder redirectInput(Path file);

  NioProcessBuilder redirectInput(Redirect source);

  Redirect redirectOutput();

  NioProcessBuilder redirectOutput(Path file);

  NioProcessBuilder redirectOutput(Redirect destination);

  Process start() throws IOException;

  ProcessBuilder toProcessBuilder();

  /**
   * A redirect like {@link ProcessBuilder.Redirect} that uses NIO {@code Path}
   * objects instead of {@code File} objects.
   */
  interface Redirect {
    boolean equals(Object obj);

    Path file();

    int hashCode();

    ProcessBuilder.Redirect.Type type();

    ProcessBuilder.Redirect toProcessBuilderRedirect();
  }
}
