/*
 * Copyright (c) 2014 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.util.Objects;

/**
 * Thrown to indicate an OS process exited with a nonzero exit status.
 */
class NonzeroExitStatusException extends Exception {
  private static final long serialVersionUID = 1L;

  private final String processName;
  private final Integer exitStatus;
  private final String stderr;

  public NonzeroExitStatusException(String processName) {
    this(processName, null);
  }

  public NonzeroExitStatusException(String processName, Integer exitStatus) {
    this(processName, exitStatus, null);
  }

  public NonzeroExitStatusException(String processName, Integer exitStatus, String stderr) {
    super(newMessage(processName, exitStatus, stderr));
    this.processName = Objects.requireNonNull(processName);
    this.exitStatus = exitStatus;
    this.stderr = stderr;
  }

  public String processName() {
    return processName;
  }

  public Integer exitStatus() {
    return exitStatus;
  }

  public boolean hasExitStatus() {
    return exitStatus != null;
  }

  public String stderr() {
    return stderr;
  }

  public boolean hasStderr() {
    return stderr != null;
  }

  private static String newMessage(String processName, Integer exitStatus, String stderr) {
    StringBuilder b = new StringBuilder();
    b.append(processName);
    b.append(": ");
    b.append((exitStatus == null) ? "no exit status" : "nonzero exit status: " + exitStatus);
    if (stderr != null && !stderr.isEmpty()) {
      b.append(": standard error: ");
      b.append(toOneLine(stderr));
    }
    return b.toString();
  }

  private static String toOneLine(String s) {
    if (s == null) return s;
    if (s.isEmpty()) return s;
    final String r = "; ";
    String result = s.replace("\r\n", r).replace("\r", r).replace("\n", r);
    if (result.endsWith(r) && !s.endsWith(r)) {
      result = result.substring(0, result.length() - r.length());
    }
    return result;
  }
}
