/*
 * Copyright (c) 2014 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.util.Objects;
import java.util.concurrent.TimeUnit;

/**
 * An amount of time.
 * <p>
 * Note: this class has a natural ordering that is inconsistent with equals.
 */
class TimeDuration implements Comparable<TimeDuration> {
  private final long amount;
  private final TimeUnit unit;

  public TimeDuration(long amount, TimeUnit unit) {
    super();
    this.amount = amount;
    this.unit = Objects.requireNonNull(unit);
  }

  public long amount() {
    return amount;
  }

  public TimeUnit unit() {
    return unit;
  }

  public TimeDuration to(TimeUnit u) {
    return new TimeDuration(u.convert(amount, unit), u);
  }

  @Override
  public boolean equals(Object o) {
    if (o == this) return true;
    if (!(o instanceof TimeDuration)) return false;
    TimeDuration td = (TimeDuration)o;
    if (amount != td.amount) return false;
    return unit.equals(td.unit);
  }

  @Override
  public int hashCode() {
    int result = 17;
    result = 31 * result + (int)(amount ^ (amount >>> 32));
    result = 31 * result + unit.hashCode();
    return result;
  }

  @Override
  public int compareTo(TimeDuration td) {
    if (td == this) return 0;
    if (unit.equals(td.unit)) {
      return (amount == td.amount) ? 0 : ((amount > td.amount) ? 1 : -1);
    } else {
      return smallerThan(unit, td.unit) ? compareTo(td.to(unit)) : to(td.unit).compareTo(td);
    }
  }

  @Override
  public String toString() {
    return amount + " " + unit;
  }

  public static TimeDuration days(long amount) {
    return new TimeDuration(amount, TimeUnit.DAYS);
  }

  public static TimeDuration hours(long amount) {
    return new TimeDuration(amount, TimeUnit.HOURS);
  }

  public static TimeDuration microseconds(long amount) {
    return new TimeDuration(amount, TimeUnit.MICROSECONDS);
  }

  public static TimeDuration milliseconds(long amount) {
    return new TimeDuration(amount, TimeUnit.MILLISECONDS);
  }

  public static TimeDuration minutes(long amount) {
    return new TimeDuration(amount, TimeUnit.MINUTES);
  }

  public static TimeDuration nanoseconds(long amount) {
    return new TimeDuration(amount, TimeUnit.NANOSECONDS);
  }

  public static TimeDuration seconds(long amount) {
    return new TimeDuration(amount, TimeUnit.SECONDS);
  }

  private static boolean smallerThan(TimeUnit a, TimeUnit b) {
    return smallToLargeOrder(a) < smallToLargeOrder(b);
  }

  private static int smallToLargeOrder(TimeUnit u) {
    switch (u) {
      case NANOSECONDS:
        return 0;
      case MICROSECONDS:
        return 1;
      case MILLISECONDS:
        return 2;
      case SECONDS:
        return 3;
      case MINUTES:
        return 4;
      case HOURS:
        return 5;
      case DAYS:
        return 6;
    }
    throw new IllegalArgumentException("Unknown TimeUnit '" + u + "'");
  }
}
