/*
 * Copyright (c) 2014 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.math.BigDecimal;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Locale;

/**
 * A crystallography unit cell.
 */
class UnitCell {
  private final List<BigDecimal> lengths;
  private final List<BigDecimal> angles;

  public UnitCell(BigDecimal[] lengthsAndAngles) {
    this(lengthsAndAngles[0], lengthsAndAngles[1], lengthsAndAngles[2], lengthsAndAngles[3],
        lengthsAndAngles[4], lengthsAndAngles[5]);
  }

  public UnitCell(BigDecimal l1, BigDecimal l2, BigDecimal l3, BigDecimal a1, BigDecimal a2,
      BigDecimal a3) {
    this(Arrays.asList(l1, l2, l3), Arrays.asList(a1, a2, a3));
  }

  /**
   * Constructs an instance with the specified lengths and angles.
   *
   * @param lengths unit cell lengths; the caller must not modify the
   *          {@code lengths} list once this call is made
   * @param angles unit cell angles; the caller must not modify the
   *          {@code angles} list once this call is made
   */
  public UnitCell(List<BigDecimal> lengths, List<BigDecimal> angles) {
    super();
    if (lengths.size() != 3) {
      throw new IllegalArgumentException("lengths size not equal to 3");
    }
    if (angles.size() != 3) {
      throw new IllegalArgumentException("angles size not equal to 3");
    }
    this.lengths = Collections.unmodifiableList(lengths);
    this.angles = Collections.unmodifiableList(angles);
  }

  public List<BigDecimal> lengths() {
    return lengths;
  }

  public List<BigDecimal> angles() {
    return angles;
  }

  /**
   * Compares this unit cell to the specified object for equality.
   * <p>
   * Note: this compares the values <em>and</em> scales of the
   * {@code BigDecimal} lengths and angles. For example, {@code 2.0} is not
   * equal to {@code 2.00}. To ignore scale, use
   * {@link #equalsIgnoreScale(UnitCell)} instead.
   *
   * @see #equalsIgnoreScale(UnitCell)
   */
  @Override
  public boolean equals(Object o) {
    if (this == o) return true;
    if (!(o instanceof UnitCell)) return false;
    UnitCell u = (UnitCell)o;
    return lengths.equals(u.lengths) && angles.equals(u.angles);
  }

  /**
   * Compares this unit cell to the specified unit cell for equality, ignoring
   * the scale of the {@code BigDecimal} lengths and angles.
   */
  public boolean equalsIgnoreScale(UnitCell u) {
    if (this == u) return true;
    if (u == null) return false;
    final int lengthsSize = lengths.size();
    for (int i = 0; i < lengthsSize; i++) {
      if (lengths.get(i).compareTo(u.lengths.get(i)) != 0) return false;
    }
    final int anglesSize = angles.size();
    for (int i = 0; i < anglesSize; i++) {
      if (angles.get(i).compareTo(u.angles.get(i)) != 0) return false;
    }
    return true;
  }

  @Override
  public int hashCode() {
    int result = 17;
    result = 31 * result + lengths.hashCode();
    result = 31 * result + angles.hashCode();
    return result;
  }

  @Override
  public String toString() {
    return String.format(Locale.US, "(%s, %s)", lengths, angles);
  }
}