/*
 * Copyright (c) 2015 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.io.IOException;
import java.io.InterruptedIOException;
import java.nio.file.AtomicMoveNotSupportedException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardCopyOption;
import java.util.Objects;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.TimeUnit;

/**
 * Static utility methods.
 */
class Utilities {
  private Utilities() {
    throw new AssertionError();
  }

  /**
   * Closes the specified resource, trying not to throw an exception.
   *
   * All exceptions are caught and ignored except for
   * {@code InterruptedException} and {@code InterruptedIOException}, for which
   * the current thread is interrupted, and {@code ThreadDeath}, which is
   * rethrown.
   *
   * @param c the resource to close; if {@code null}, returns immediately
   */
  public static void closeSilently(AutoCloseable c) {
    if (c == null) return;
    boolean interrupted = false;
    try {
      for (;;) {
        try {
          c.close();
          break;
        } catch (InterruptedException e) {
          interrupted = true;
        }
      }
    } catch (InterruptedIOException e) {
      Thread.currentThread().interrupt();
    } catch (ThreadDeath td) {
      throw td;
    } catch (Throwable t) {
      /* IGNORED */
    } finally {
      if (interrupted) Thread.currentThread().interrupt();
    }
  }

  /**
   * Deletes the specified file, trying not to throw an exception.
   *
   * All exceptions are caught and ignored except for
   * {@code InterruptedIOException}, for which the current thread is
   * interrupted, and {@code ThreadDeath}, which is rethrown.
   *
   * @param f the file to delete
   */
  public static void deleteFileSilently(Path f) {
    if (f == null) return;
    try {
      Files.delete(f);
    } catch (InterruptedIOException e) {
      Thread.currentThread().interrupt();
    } catch (ThreadDeath td) {
      throw td;
    } catch (Throwable t) {
      /* IGNORED */
    }
  }

  /**
   * Moves the specified source file to the specified target file, atomically
   * if possible.
   *
   * A move as an atomic file system operation is attempted first. If that
   * fails with an {@code IOException}, the target file is deleted if it
   * exists, and the move as an atomic file system operation is attempted
   * again. If either move attempt fails with an indication that the move
   * cannot be performed as an atomic file system operation, then a move as a
   * nonatomic file system operation is attempted, replacing the target file if
   * it exists.
   *
   * @param source the file to move
   * @param target the destination of the file to move
   *
   * @throws IOException if an I/O error occurs
   * @throws SecurityException In the case of the default provider, and a
   *           security manager is installed, the
   *           {@link SecurityManager#checkWrite(String) checkWrite} method is
   *           invoked to check write access to both the source and target
   *           file.
   */
  public static void moveFile(Path source, Path target) throws IOException {
    for (int attempt = 0; attempt < 2; attempt++) {
      try {
        moveFileAtomically(source, target);
        return;
      } catch (AtomicMoveNotSupportedException e) {
        moveFileNonatomically(source, target);
        return;
      } catch (IOException e) {
        if (attempt == 0) {
          /*
           * An atomic implementation may fail if the target exists. Delete the
           * target, and try again.
           */
          Files.deleteIfExists(target);
        } else {
          throw e;
        }
      }
    }
  }

  private static void moveFileAtomically(Path source, Path target) throws IOException {
    Files.move(source, target, StandardCopyOption.ATOMIC_MOVE);
  }

  private static void moveFileNonatomically(Path source, Path target) throws IOException {
    Files.move(source, target, StandardCopyOption.REPLACE_EXISTING);
  }

  /**
   * Shuts down the specified service, trying not to throw an exception.
   *
   * All exceptions are caught and ignored except for {@code ThreadDeath},
   * which is rethrown.
   *
   * @param s the service to shut down; if {@code null}, returns immediately
   */
  public static void shutdownNowSilently(ExecutorService s) {
    if (s == null) return;
    try {
      s.shutdownNow();
    } catch (ThreadDeath td) {
      throw td;
    } catch (Throwable t) {
      /* IGNORED */
    }
  }

  /**
   * Waits for the specified service to terminate, trying not to throw an
   * exception.
   *
   * All exceptions are caught and ignored except for
   * {@code InterruptedException}, which is noted and causes the current thread
   * to be interrupted before returning, and {@code ThreadDeath}, which is
   * rethrown.
   *
   * @param s the service to shut down; if {@code null}, returns immediately
   */
  public static void awaitTerminationSilently(ExecutorService s) {
    if (s == null) return;
    boolean interrupted = false;
    try {
      for (;;) {
        try {
          s.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);
          break;
        } catch (InterruptedException e) {
          interrupted = true;
        }
      }
    } catch (ThreadDeath td) {
      throw td;
    } catch (Throwable t) {
      /* IGNORED */
    } finally {
      if (interrupted) Thread.currentThread().interrupt();
    }
  }

  /**
   * Returns the base name of a dataset filename. The base name is everything
   * in the filename up to and including the last underscore character. If
   * there is no underscore, but the filename ends with "<tt>master.h5</tt>",
   * the base name is everything up to, but not including, that. Otherwise, the
   * base name is everything in the filename up to, but not including, the
   * first digit in the last group of digits. The last group of digits must
   * either end the filename or be followed by an extension where an extension
   * is a dot followed by zero or more non-dot characters.
   *
   * @param filename the name of the dataset file or the name of an image file
   *          in a dataset
   *
   * @return the base name, or the {@code filename} parameter if the base name
   *         could not be determined
   *
   * @throws NullPointerException if the {@code filename} parameter is
   *           {@code null}
   */
  public static String datasetFileBaseName(String filename) {
    Objects.requireNonNull(filename, "filename must not be null");
    int underIdx = filename.lastIndexOf('_');
    if (underIdx != -1) return filename.substring(0, underIdx + 1);
    String mh5 = "master.h5";
    if (filename.endsWith(mh5)) return filename.substring(0, filename.length() - mh5.length());
    int dotIdx = filename.lastIndexOf('.');
    int effectiveEndIdx = (dotIdx == -1) ? filename.length() - 1 : dotIdx - 1;
    if (effectiveEndIdx < 0) return filename;
    int digitGroupIdx = -1;
    for (int i = effectiveEndIdx; i >= 0; i--) {
      char c = filename.charAt(i);
      if (c < '0' || c > '9') break;
      digitGroupIdx = i;
    }
    return (digitGroupIdx >= 0) ? filename.substring(0, digitGroupIdx) : filename;
  }
}
