/*
 * Copyright (c) 2015 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.math.BigDecimal;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Parses an <a href="http://xds.mpimf-heidelberg.mpg.de/">XDS</a> integrate
 * log.
 */
class XdsIntegrateLogParser {
  /* @formatter:off */
  private static final Pattern SECTION_INTEGRATE_START_RE = Pattern.compile(
      "^\\s*\\*\\*\\*\\*\\*\\s+INTEGRATE\\s+\\*\\*\\*\\*\\*.*", Pattern.CASE_INSENSITIVE);
  private static final Pattern SECTION_PROCESSING_OF_IMAGES_START_RE = Pattern.compile(
      "^\\s*PROCESSING OF IMAGES\\s+.*", Pattern.CASE_INSENSITIVE);
  private static final Pattern SECTION_REFINEMENT_START_RE = Pattern.compile(
      "(?:^\\s*\\d+\\s+OUT OF\\s+\\d+\\s+REFLECTIONS ACCEPTED FOR REFINEMENT.*)|" +
          "(?:^\\s*!!! WARNING !!! UNABLE TO CARRY OUT REFINEMENT.*)", Pattern.CASE_INSENSITIVE);
  private static final Pattern DET_DIST_RE = Pattern.compile(
      "^\\s*DETECTOR_DISTANCE\\s*=\\s*([0-9.]+).*", Pattern.CASE_INSENSITIVE);
  private static final Pattern UNABLE_TO_REFINE_RE = Pattern.compile(
      "(?:^\\s*!!! WARNING !!! UNABLE TO CARRY OUT REFINEMENT.*)", Pattern.CASE_INSENSITIVE);
  private static final Pattern REFINED_DET_DIST_RE = Pattern.compile(
      "^\\s*CRYSTAL TO DETECTOR DISTANCE\\s*\\(mm\\)\\s*([0-9.]+).*", Pattern.CASE_INSENSITIVE);
  /* @formatter:on */

  private final BufferedReader in;
  private Section section;
  private String sectionLineBuffer;

  public XdsIntegrateLogParser(BufferedReader in) {
    super();
    this.in = in;
    section = null;
    sectionLineBuffer = null;
  }

  public XdsIntegrateResult parse() {
    try {
      XdsIntegrateResultBuilder b = new XdsIntegrateResultBuilder();
      b.detectorDistance(parseDetectorDistance());
      RefinedParameters rp = parseLastRefinedParameters();
      if (rp != null) {
        b.failedToRefine(rp.unableToRefine());
        b.refinedDetectorDistance(rp.detectorDistance());
      }
      return b.toXdsIntegrateResult();
    } catch (IOException e) {
      if (e instanceof InterruptedIOException) {
        Thread.currentThread().interrupt();
      }
      return new XdsIntegrateResult(e);
    }
  }

  private BigDecimal parseDetectorDistance() throws IOException {
    for (String line = readLineInIntegrate(); line != null; line = readLineInIntegrate()) {
      Matcher m = DET_DIST_RE.matcher(line);
      if (m.matches()) return toBigDecimal(m.group(1));
    }
    return null;
  }

  private RefinedParameters parseLastRefinedParameters() throws IOException {
    RefinedParameters result = null;
    for (;;) {
      RefinedParameters newResult = parseRefinedParameters();
      if (newResult == null) return result;
      result = newResult;
    }
  }

  private RefinedParameters parseRefinedParameters() throws IOException {
    boolean unableToRefine = false;
    BigDecimal detDist = null;
    String line = readLineInRefinement();
    if (line == null) return null;
    do {
      Matcher m = UNABLE_TO_REFINE_RE.matcher(line);
      if (m.matches()) {
        unableToRefine = true;
      } else {
        m = REFINED_DET_DIST_RE.matcher(line);
        if (m.matches()) detDist = toBigDecimal(m.group(1));
      }
      line = readLineInRefinement();
    } while (line != null);
    return new RefinedParameters(unableToRefine, detDist);
  }

  private String readLineInIntegrate() throws IOException {
    return readLine(Section.INTEGRATE);
  }

  private String readLineInRefinement() throws IOException {
    return readLine(Section.REFINEMENT);
  }

  private String readLine(Section target) throws IOException {
    Section origSection = section;
    for (;;) {
      String line;
      if (sectionLineBuffer != null) {
        line = sectionLineBuffer;
        sectionLineBuffer = null;
      } else {
        line = readLine();
      }
      if (line == null) return null;
      if (in(target)) return line;
      if (origSection != null && origSection.equals(target)) {
        /*
         * Buffer the first line of the next section and use it in the next
         * call to this method. This allows a read of the next section to get
         * the first section line rather than the second.
         */
        sectionLineBuffer = line;
        return null;
      }
    }
  }

  private String readLine() throws IOException {
    String line = in.readLine();
    if (line == null) return null;
    Section s = matchSectionStart(line);
    if (s != null) entered(s);
    return line;
  }

  private boolean in(Section aSection) {
    return (section == null) ? false : section.equals(aSection);
  }

  private void entered(Section newSection) {
    section = newSection;
  }

  public static XdsIntegrateResult parse(Path log) {
    try (BufferedReader reader = Files.newBufferedReader(log, StandardCharsets.UTF_8)) {
      return (new XdsIntegrateLogParser(reader)).parse();
    } catch (IOException e) {
      if (e instanceof InterruptedIOException) {
        Thread.currentThread().interrupt();
      }
      return new XdsIntegrateResult(e);
    }
  }

  private static Section matchSectionStart(String line) {
    if (SECTION_INTEGRATE_START_RE.matcher(line).matches()) {
      return Section.INTEGRATE;
    }
    if (SECTION_PROCESSING_OF_IMAGES_START_RE.matcher(line).matches()) {
      return Section.PROCESSING_OF_IMAGES;
    }
    if (SECTION_REFINEMENT_START_RE.matcher(line).matches()) {
      return Section.REFINEMENT;
    }
    return null;
  }

  private static BigDecimal toBigDecimal(String value) {
    try {
      return new BigDecimal(value);
    } catch (NumberFormatException e) {
      return null;
    }
  }

  private static enum Section {
    INTEGRATE,
    PROCESSING_OF_IMAGES,
    REFINEMENT
  }

  private static class RefinedParameters {
    private final boolean unableToRefine;
    private final BigDecimal detectorDistance;

    public RefinedParameters(boolean unableToRefine, BigDecimal detectorDistance) {
      super();
      this.unableToRefine = unableToRefine;
      this.detectorDistance = detectorDistance;
    }

    public boolean unableToRefine() {
      return unableToRefine;
    }

    public BigDecimal detectorDistance() {
      return detectorDistance;
    }

    @Override
    public String toString() {
      StringBuilder b = new StringBuilder();
      b.append("unableToRefine=");
      b.append(unableToRefine);
      b.append(", detectorDistance=");
      b.append(detectorDistance);
      return b.toString();
    }
  }
}
