/*
 * Copyright (c) 2014, 2015 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.math.BigDecimal;
import java.util.Arrays;
import java.util.Collections;

import org.junit.Test;

/**
 * Tests {@link AapConfigurationParser}.
 */
public class AapConfigurationParserTest {
  public AapConfigurationParserTest() {
    super();
  }

  @Test
  public void testDatasetSettleTime_normal() throws IOException, ConfigurationException {
    String conf = "DatasetSettleTime = 5 s";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.seconds(5L), c.datasetSettleTime());
  }

  @Test
  public void testDiffCenterImagePat_normal() throws IOException, ConfigurationException {
    String conf = "DiffractionCenterImagePattern = (?i)cell_.*\\\\.tiff";
    AapConfiguration c = parse(conf);
    assertEquals("(?i)cell_.*\\.tiff", c.diffCenterImagePat().pattern());
  }

  @Test
  public void testFileSystemPollEnabled_yes() throws IOException, ConfigurationException {
    String conf = "FileSystemPoll = yes";
    AapConfiguration c = parse(conf);
    assertTrue(c.fileSystemPollEnabled());
  }

  @Test
  public void testFileSystemPollEnabled_no() throws IOException, ConfigurationException {
    String conf = "FileSystemPoll = no";
    AapConfiguration c = parse(conf);
    assertFalse(c.fileSystemPollEnabled());
  }

  @Test(expected = ConfigurationException.class)
  public void testFileSystemPollEnabled_invalid() throws IOException, ConfigurationException {
    String conf = "FileSystemPoll = false";
    parse(conf);
  }

  @Test
  public void testFileSystemPollPeriod_normal() throws IOException, ConfigurationException {
    String conf = "FileSystemPollPeriod = 5 s";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.seconds(5L), c.fileSystemPollPeriod());
  }

  @Test
  public void testFileSystemPollPeriod_default() throws IOException, ConfigurationException {
    String conf = "";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.seconds(10L), c.fileSystemPollPeriod());
  }

  @Test(expected = ConfigurationException.class)
  public void testFileSystemPollPeriod_zero() throws IOException, ConfigurationException {
    String conf = "FileSystemPollPeriod = 0";
    parse(conf);
  }

  @Test(expected = ConfigurationException.class)
  public void testFileSystemPollPeriod_negative() throws IOException, ConfigurationException {
    String conf = "FileSystemPollPeriod = -1";
    parse(conf);
  }

  @Test
  public void testFileSystemPollPeriod_defaultUnit() throws IOException, ConfigurationException {
    String conf = "FileSystemPollPeriod = 5";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.seconds(5L), c.fileSystemPollPeriod());
  }

  @Test
  public void testFileSystemPollPeriod_defaultUnitWithTrailingSpace()
      throws IOException, ConfigurationException {
    String conf = "FileSystemPollPeriod = 5 ";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.seconds(5L), c.fileSystemPollPeriod());
  }

  @Test
  public void testFileSystemPollPeriod_noSpaceBeforeUnit()
      throws IOException, ConfigurationException {
    String conf = "FileSystemPollPeriod = 5m";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.minutes(5L), c.fileSystemPollPeriod());
  }

  @Test
  public void testFileSystemPollPeriod_days() throws IOException, ConfigurationException {
    String conf = "FileSystemPollPeriod = 5 d";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.days(5L), c.fileSystemPollPeriod());
  }

  @Test
  public void testFileSystemPollPeriod_hours() throws IOException, ConfigurationException {
    String conf = "FileSystemPollPeriod = 5 h";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.hours(5L), c.fileSystemPollPeriod());
  }

  @Test
  public void testFileSystemPollPeriod_minutes() throws IOException, ConfigurationException {
    String conf = "FileSystemPollPeriod = 5 m";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.minutes(5L), c.fileSystemPollPeriod());
  }

  @Test
  public void testFileSystemPollPeriod_seconds() throws IOException, ConfigurationException {
    String conf = "FileSystemPollPeriod = 5 s";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.seconds(5L), c.fileSystemPollPeriod());
  }

  @Test
  public void testFileSystemPollPeriod_milliseconds() throws IOException, ConfigurationException {
    String conf = "FileSystemPollPeriod = 5 ms";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.milliseconds(5L), c.fileSystemPollPeriod());
  }

  @Test
  public void testFileSystemPollPeriod_microseconds() throws IOException, ConfigurationException {
    String conf = "FileSystemPollPeriod = 5 \u03BCs";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.microseconds(5L), c.fileSystemPollPeriod());
  }

  @Test
  public void testFileSystemPollPeriod_microsecondsAscii()
      throws IOException, ConfigurationException {
    String conf = "FileSystemPollPeriod = 5 us";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.microseconds(5L), c.fileSystemPollPeriod());
  }

  @Test
  public void testFileSystemPollPeriod_nanoseconds() throws IOException, ConfigurationException {
    String conf = "FileSystemPollPeriod = 5 ns";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.nanoseconds(5L), c.fileSystemPollPeriod());
  }

  @Test
  public void testFileSystemPollMaxNumJobs_normal() throws IOException, ConfigurationException {
    String conf = "FileSystemPollMaxNumJobs = 2";
    AapConfiguration c = parse(conf);
    assertEquals(2, c.fileSystemPollMaxNumJobs());
  }

  @Test
  public void testImagePat_normal() throws IOException, ConfigurationException {
    String conf = "ImagePattern = (?i).+\\\\.tiff";
    AapConfiguration c = parse(conf);
    assertEquals("(?i).+\\.tiff", c.imagePat().pattern());
  }

  @Test
  public void testMaxNumDirChangeJobs_normal() throws IOException, ConfigurationException {
    String conf = "MaxNumDirChangeJobs = 2";
    AapConfiguration c = parse(conf);
    assertEquals(2, c.maxNumDirChangeJobs());
  }

  @Test
  public void testMaxNumDirChangeJobs_default() throws IOException, ConfigurationException {
    String conf = "";
    AapConfiguration c = parse(conf);
    assertEquals(1, c.maxNumDirChangeJobs());
  }

  @Test(expected = ConfigurationException.class)
  public void testMaxNumDirChangeJobs_lessThanOne() throws IOException, ConfigurationException {
    String conf = "MaxNumDirChangeJobs = 0";
    parse(conf);
  }

  @Test
  public void testMaxNumIndexJobs_normal() throws IOException, ConfigurationException {
    String conf = "MaxNumIndexJobs = 2";
    AapConfiguration c = parse(conf);
    assertEquals(2, c.maxNumIndexJobs());
  }

  @Test
  public void testMaxNumIndexJobs_default() throws IOException, ConfigurationException {
    String conf = "";
    AapConfiguration c = parse(conf);
    assertEquals(1, c.maxNumIndexJobs());
  }

  @Test(expected = ConfigurationException.class)
  public void testMaxNumIndexJobs_lessThanOne() throws IOException, ConfigurationException {
    String conf = "MaxNumIndexJobs = 0";
    parse(conf);
  }

  @Test
  public void testMaxNumRegisterJobs_normal() throws IOException, ConfigurationException {
    String conf = "MaxNumRegisterJobs = 2";
    AapConfiguration c = parse(conf);
    assertEquals(2, c.maxNumRegisterJobs());
  }

  @Test
  public void testMaxNumRegisterJobs_default() throws IOException, ConfigurationException {
    String conf = "";
    AapConfiguration c = parse(conf);
    assertEquals(1, c.maxNumRegisterJobs());
  }

  @Test(expected = ConfigurationException.class)
  public void testMaxNumRegisterJobs_lessThanOne() throws IOException, ConfigurationException {
    String conf = "MaxNumRegisterJobs = 0";
    parse(conf);
  }

  @Test
  public void testMaxNumProcessJobs_normal() throws IOException, ConfigurationException {
    String conf = "MaxNumProcessJobs = 2";
    AapConfiguration c = parse(conf);
    assertEquals(2, c.maxNumProcessJobs());
  }

  @Test
  public void testMaxNumProcessJobs_default() throws IOException, ConfigurationException {
    String conf = "";
    AapConfiguration c = parse(conf);
    assertEquals(1, c.maxNumProcessJobs());
  }

  @Test(expected = ConfigurationException.class)
  public void testMaxNumProcessJobs_lessThanOne() throws IOException, ConfigurationException {
    String conf = "MaxNumProcessJobs = 0";
    parse(conf);
  }

  @Test
  public void testProcessingDirName_default() throws IOException, ConfigurationException {
    String conf = "ProcessingDirectoryName = xprocessing";
    AapConfiguration c = parse(conf);
    assertEquals("xprocessing", c.processingDirName());
  }

  @Test
  public void testProcessingDirPat_default() throws IOException, ConfigurationException {
    String conf = "ProcessingDirectoryPattern = (?i)xprocessing";
    AapConfiguration c = parse(conf);
    assertEquals("(?i)xprocessing", c.processingDirPat().pattern());
  }

  @Test
  public void testPerTreeResultsEnabled_yes() throws IOException, ConfigurationException {
    String conf = "PerTreeResults = yes";
    AapConfiguration c = parse(conf);
    assertTrue(c.perTreeResultsEnabled());
  }

  @Test
  public void testPerTreeResultsEnabled_no() throws IOException, ConfigurationException {
    String conf = "PerTreeResults = no";
    AapConfiguration c = parse(conf);
    assertFalse(c.perTreeResultsEnabled());
  }

  @Test(expected = ConfigurationException.class)
  public void testPerTreeResultsEnabled_invalid() throws IOException, ConfigurationException {
    String conf = "PerTreeResults = false";
    parse(conf);
  }

  @Test
  public void testMaxProcessRunTime_normal() throws IOException, ConfigurationException {
    String conf = "MaxProcessRunTime = 5 m";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.minutes(5L), c.maxProcessRunTime());
  }

  @Test
  public void testMaxProcessRunTime_default() throws IOException, ConfigurationException {
    String conf = "";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.minutes(45L), c.maxProcessRunTime());
  }

  @Test(expected = ConfigurationException.class)
  public void testMaxProcessRunTime_zero() throws IOException, ConfigurationException {
    String conf = "MaxProcessRunTime = 0";
    parse(conf);
  }

  @Test(expected = ConfigurationException.class)
  public void testMaxProcessRunTime_negative() throws IOException, ConfigurationException {
    String conf = "MaxProcessRunTime = -1";
    parse(conf);
  }

  @Test
  public void testMaxProcessRunTime_defaultUnit() throws IOException, ConfigurationException {
    String conf = "MaxProcessRunTime = 5";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.minutes(5L), c.maxProcessRunTime());
  }

  @Test
  public void testProcessSupervisorPeriod_normal() throws IOException, ConfigurationException {
    String conf = "ProcessSupervisorPeriod = 120 s";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.seconds(120L), c.processSupervisorPeriod());
  }

  @Test
  public void testProcessSupervisorPeriod_default() throws IOException, ConfigurationException {
    String conf = "";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.seconds(60L), c.processSupervisorPeriod());
  }

  @Test(expected = ConfigurationException.class)
  public void testProcessSupervisorPeriod_zero() throws IOException, ConfigurationException {
    String conf = "ProcessSupervisorPeriod = 0";
    parse(conf);
  }

  @Test(expected = ConfigurationException.class)
  public void testProcessSupervisorPeriod_negative() throws IOException, ConfigurationException {
    String conf = "ProcessSupervisorPeriod = -1";
    parse(conf);
  }

  @Test
  public void testProcessSupervisorPeriod_defaultUnit() throws IOException, ConfigurationException {
    String conf = "ProcessSupervisorPeriod = 30";
    AapConfiguration c = parse(conf);
    assertEquals(TimeDuration.seconds(30L), c.processSupervisorPeriod());
  }

  @Test
  public void testMaxRefinedDetDistDelta_normalInteger()
      throws IOException, ConfigurationException {
    String conf = "MaxRefinedDetDistDelta = 23";
    AapConfiguration c = parse(conf);
    assertEquals(new BigDecimal("23"), c.maxRefinedDetDistDelta());
  }

  @Test
  public void testMaxRefinedDetDistDelta_normalDecimal()
      throws IOException, ConfigurationException {
    String conf = "MaxRefinedDetDistDelta = 23.9";
    AapConfiguration c = parse(conf);
    assertEquals(new BigDecimal("23.9"), c.maxRefinedDetDistDelta());
  }

  @Test
  public void testMaxRefinedDetDistDelta_default() throws IOException, ConfigurationException {
    String conf = "";
    AapConfiguration c = parse(conf);
    assertEquals(new BigDecimal("50"), c.maxRefinedDetDistDelta());
  }

  @Test(expected = ConfigurationException.class)
  public void testMaxRefinedDetDistDelta_zero() throws IOException, ConfigurationException {
    String conf = "MaxRefinedDetDistDelta = 0";
    parse(conf);
  }

  @Test(expected = ConfigurationException.class)
  public void testMaxRefinedDetDistDelta_negative() throws IOException, ConfigurationException {
    String conf = "MaxRefinedDetDistDelta = -1";
    parse(conf);
  }

  @Test
  public void testNumThreadsPerProcessJob_normal() throws IOException, ConfigurationException {
    String conf = "NumThreadsPerProcessJob = 2";
    AapConfiguration c = parse(conf);
    assertEquals(2, c.numThreadsPerProcessJob());
  }

  @Test
  public void testNumThreadsPerProcessJob_default() throws IOException, ConfigurationException {
    String conf = "";
    AapConfiguration c = parse(conf);
    assertEquals(1, c.numThreadsPerProcessJob());
  }

  @Test(expected = ConfigurationException.class)
  public void testNumThreadsPerProcessJob_lessThanOne() throws IOException, ConfigurationException {
    String conf = "NumThreadsPerProcessJob = 0";
    parse(conf);
  }

  @Test
  public void testPostProcessHook_normal() throws IOException, ConfigurationException {
    String conf = "PostProcessHook = hook-prog $dataset";
    AapConfiguration c = parse(conf);
    assertEquals(Arrays.asList("hook-prog", "$dataset"), c.postProcessHook());
  }

  @Test
  public void testPostProcessHook_empty() throws IOException, ConfigurationException {
    String conf = "PostProcessHook = ";
    AapConfiguration c = parse(conf);
    assertEquals(Collections.<String>emptyList(), c.postProcessHook());
  }

  @Test
  public void testPostProcessHook_default() throws IOException, ConfigurationException {
    String conf = "";
    AapConfiguration c = parse(conf);
    assertEquals(Collections.<String>emptyList(), c.postProcessHook());
  }

  @Test(expected = ConfigurationException.class)
  public void testPostProcessHook_lessThanOne() throws IOException, ConfigurationException {
    String conf = "PostProcessHook = hook-prog \"unbalanced double-quote";
    parse(conf);
  }

  private static AapConfiguration parse(String conf) throws IOException, ConfigurationException {
    try (Reader r = new StringReader(conf)) {
      return (new AapConfigurationParser(r)).parse();
    }
  }
}
