/*
 * Copyright (c) 2015 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.FileSystem;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.attribute.FileTime;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

import org.imca_cat.aaproc.AaprocDbProtos.AaprocDbDataset;
import org.imca_cat.aaproc.AaprocDbProtos.AaprocDbDatasetProcessFailure;
import org.imca_cat.aaproc.AaprocDbProtos.AaprocDbDatasetProcessResult;
import org.imca_cat.procrun.InputStreamConsumerFactory;
import org.imca_cat.procrun.ProcessRunner;
import org.imca_cat.procrun.ProcessRunnerResult;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import com.github.marschall.memoryfilesystem.MemoryFileSystemBuilder;
import com.google.protobuf.TextFormat;

/**
 * Tests {@link AutoProcDatasetProcessor}.
 */
public class AutoProcDatasetProcessorTest {
  private FileSystem fs;

  public AutoProcDatasetProcessorTest() {
    super();
    fs = null;
  }

  @Before
  public void setUp() throws IOException {
    fs = MemoryFileSystemBuilder.newLinux().build("test");
  }

  @After
  public void tearDown() throws Throwable {
    Utilities.closeSilently(fs);
  }

  @Test
  public void testProcess_normal() throws InterruptedException, IOException {
    final AapConfiguration conf = new AapConfiguration();
    final Path datasetDir = fs.getPath("/tmp", "data", "blue52");
    Files.createDirectories(datasetDir);
    Path image1 = datasetDir.resolve("Image_0001.cbf");
    Files.createFile(image1);
    AutoProcDatasetProcessor p =
        new AutoProcDatasetProcessor(new Dataset(datasetDir), conf, "nonexistent");
    p.processRunnerFactory(newProcessRunnerFactory(new ProcessRunnerStub() {
      @Override
      public ProcessRunnerResult call() {
        try {
          Path autoProcXml = datasetDir.resolve(conf.processingDirName()).resolve("autoPROC.xml");
          try (BufferedWriter w = Files.newBufferedWriter(autoProcXml, StandardCharsets.UTF_8)) {
            w.write(newNormalAutoProcXml());
          }
          return new ProcessRunnerResult(0, null, null, null);
        } catch (IOException e) {
          throw new RuntimeException(e);
        }
      }
    }));
    DatasetProcessResult r = p.process();
    assertTrue(r.succeeded());
  }

  @Test
  public void testProcess_supervisorAbort() throws InterruptedException, IOException {
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.maxProcessRunTime(TimeDuration.milliseconds(1));
    final AapConfiguration conf = b.toAapConfiguration();
    final Path datasetDir = fs.getPath("/tmp", "data", "blue52");
    Files.createDirectories(datasetDir);
    Path image1 = datasetDir.resolve("Image_0001.cbf");
    Files.createFile(image1);
    AutoProcDatasetProcessor p =
        new AutoProcDatasetProcessor(new Dataset(datasetDir), conf, "nonexistent");
    p.processRunnerFactory(newProcessRunnerFactory(new ProcessRunnerStub() {
      @Override
      public ProcessRunnerResult call() {
        try {
          TimeUnit.DAYS.sleep(Long.MAX_VALUE);
          return new ProcessRunnerResult(0, null, null, null);
        } catch (InterruptedException e) {
          Thread.currentThread().interrupt();
          return new ProcessRunnerResult(null, e, null, null);
        }
      }
    }));
    DatasetProcessResult r = p.process();
    assertFalse(r.succeeded());
    Path db = datasetDir.resolve(AaprocSystem.DB_FILE_NAME);
    assertTrue(Files.exists(db));
    AaprocDbDataset.Builder dbBuilder = AaprocDbDataset.newBuilder();
    try (BufferedReader dbReader = Files.newBufferedReader(db, StandardCharsets.UTF_8)) {
      TextFormat.merge(dbReader, dbBuilder);
    }
    AaprocDbDataset dbDataset = dbBuilder.build();
    assertTrue(dbDataset.hasProcessResult());
    AaprocDbDatasetProcessResult dbResult = dbDataset.getProcessResult();
    assertTrue(dbResult.hasTime());
    assertTrue(dbResult.hasFailure());
    AaprocDbDatasetProcessFailure dbFailure = dbResult.getFailure();
    assertNotNull(dbFailure.getMessage());
    assertNotEquals("", dbFailure.getMessage().trim());
  }

  @Test
  public void testProcess_noProcessIfLogFilePresent() throws InterruptedException, IOException {
    Path datasetDir = fs.getPath("/tmp", "data", "blue52");
    Files.createDirectories(datasetDir);
    AapConfiguration conf = new AapConfiguration();
    Path processingDir = datasetDir.resolve(conf.processingDirName());
    Files.createDirectory(processingDir);
    Files.createFile(processingDir.resolve("process.20150730_171231.log"));
    AutoProcDatasetProcessor p =
        new AutoProcDatasetProcessor(new Dataset(datasetDir), conf, "nonexistent");
    DatasetProcessResult r = p.process();
    assertTrue(r.exception() instanceof DatasetProcessFailedException);
  }

  @Test
  public void testProcess_processIfLogFileNotPresent() throws InterruptedException, IOException {
    Path datasetDir = fs.getPath("/tmp", "data", "blue52");
    Files.createDirectories(datasetDir);
    AapConfiguration conf = new AapConfiguration();
    Path processingDir = datasetDir.resolve(conf.processingDirName());
    Files.createDirectory(processingDir);
    AutoProcDatasetProcessor p =
        new AutoProcDatasetProcessor(new Dataset(datasetDir), conf, "nonexistent");
    CountDownLatch processed = new CountDownLatch(1);
    p.processRunnerFactory(newProcessRunnerFactory(processed));
    p.process();
    assertTrue(processed.await(3L, TimeUnit.SECONDS));
  }

  @Test
  public void testProcess_noProcessIfLogFilePresentAndNewerThanImages()
      throws InterruptedException, IOException {
    Path datasetDir = fs.getPath("/tmp", "data", "blue52");
    Files.createDirectories(datasetDir);
    Path image1 = datasetDir.resolve("Image_0001.cbf");
    Files.createFile(image1);
    AapConfiguration conf = new AapConfiguration();
    Path processingDir = datasetDir.resolve(conf.processingDirName());
    Files.createDirectory(processingDir);
    Path log = processingDir.resolve("process.20150730_171231.log");
    Files.createFile(log);
    /*
     * Make the image file have an older last-modified time than the log file.
     */
    FileTime image1Time = FileTime.fromMillis(1000);
    FileTime logTime = FileTime.fromMillis(2000);
    Files.setLastModifiedTime(image1, image1Time);
    Files.setLastModifiedTime(log, logTime);
    AutoProcDatasetProcessor p =
        new AutoProcDatasetProcessor(new Dataset(datasetDir), conf, "nonexistent");
    DatasetProcessResult r = p.process();
    assertTrue(r.exception() instanceof DatasetProcessFailedException);
  }

  @Test
  public void testProcess_processIfLogFilePresentButOlderThanImages()
      throws InterruptedException, IOException {
    Path datasetDir = fs.getPath("/tmp", "data", "blue52");
    Files.createDirectories(datasetDir);
    Path image1 = datasetDir.resolve("Image_0001.cbf");
    Files.createFile(image1);
    AapConfiguration conf = new AapConfiguration();
    Path processingDir = datasetDir.resolve(conf.processingDirName());
    Files.createDirectory(processingDir);
    Path log = processingDir.resolve("process.20150730_171231.log");
    Files.createFile(log);
    /*
     * Make the image file have a newer last-modified time than the log file.
     */
    FileTime image1Time = FileTime.fromMillis(2000);
    FileTime logTime = FileTime.fromMillis(1000);
    Files.setLastModifiedTime(image1, image1Time);
    Files.setLastModifiedTime(log, logTime);
    AutoProcDatasetProcessor p =
        new AutoProcDatasetProcessor(new Dataset(datasetDir), conf, "nonexistent");
    CountDownLatch processed = new CountDownLatch(1);
    p.processRunnerFactory(newProcessRunnerFactory(processed));
    p.process();
    assertTrue(processed.await(3L, TimeUnit.SECONDS));
  }

  private static ProcessRunnerFactory newProcessRunnerFactory(final CountDownLatch processed) {
    return new ProcessRunnerFactory() {
      @Override
      public ProcessRunner newRunner(NioProcessBuilder builder,
          InputStreamConsumerFactory stdoutConsumerFactory,
          InputStreamConsumerFactory stderrConsumerFactory) {
        return new ProcessRunnerStub() {
          @Override
          public ProcessRunnerResult call() {
            processed.countDown();
            return new ProcessRunnerResult(-1, null, null, null);
          }
        };
      };
    };
  }

  private static ProcessRunnerFactory newProcessRunnerFactory(final ProcessRunner r) {
    return new ProcessRunnerFactory() {
      @Override
      public ProcessRunner newRunner(NioProcessBuilder builder,
          InputStreamConsumerFactory stdoutConsumerFactory,
          InputStreamConsumerFactory stderrConsumerFactory) {
        return r;
      };
    };
  }

  private static String newNormalAutoProcXml() {
    /* @formatter:off */
    return
        "<?xml version=\"1.0\"?>\n" +
        "<AutoProcContainer>\n" +
        "  <AutoProcProgramContainer>\n" +
        "    <AutoProcProgram>\n" +
        "      <processingCommandLine>/auto/nfs/software/x86_64/encap/autoPROC_snapshot_20160708/autoPROC/bin/linux64/process -I .. -B -nthreads 6 -xml</processingCommandLine>\n" +
        "      <processingPrograms>autoPROC 1.1.7 (see: http://www.globalphasing.com/autoproc/)</processingPrograms>\n" +
        "      <processingStatus>1</processingStatus>\n" +
        "      <processingMessage></processingMessage>\n" +
        "      <processingStartTime>Mon Jun 12 14:55:22 CDT 2017</processingStartTime>\n" +
        "      <processingEndTime>Mon Jun 12 15:28:13 CDT 2017</processingEndTime>\n" +
        "      <processingEnvironment>\n" +
        "  Host      : r2d2.example.org\n" +
        "  OS        : PRETTY_NAME=\"Red Hat Enterprise Linux\"||REDHAT_BUGZILLA_PRODUCT=\"Red Hat Enterprise Linux 7\"|REDHAT_BUGZILLA_PRODUCT_VERSION=7.3|REDHAT_SUPPORT_PRODUCT=\"Red Hat Enterprise Linux\"|REDHAT_SUPPORT_PRODUCT_VERSION=\"7.3\"|Red Hat Enterprise Linux Workstation release 7.3 (Maipo)|x86_64|3.10.0-514.21.1.el7.x86_64\n" +
        "  User      : luke (group = luke)\n" +
        "  Directory : /data/luke/test/Blue52/processing\n" +
        "  Date      : Mon Jun 12 14:55:22 CDT 2017\n" +
        "  autoPROC  : /imcacat/software/x86_64/encap/autoPROC_snapshot_20160708\n" +
        "      </processingEnvironment>\n" +
        "    </AutoProcProgram>\n" +
        "    <AutoProcProgramAttachment>\n" +
        "      <fileType>Log</fileType>\n" +
        "      <fileName>summary.html</fileName>\n" +
        "      <filePath>/data/luke/test/Blue52/processing</filePath>\n" +
        "    </AutoProcProgramAttachment>\n" +
        "    <AutoProcProgramAttachment>\n" +
        "      <fileType>Result</fileType>\n" +
        "      <fileName>truncate-unique.mtz</fileName>\n" +
        "      <filePath>/data/luke/test/Blue52/processing/./Blue52</filePath>\n" +
        "    </AutoProcProgramAttachment>\n" +
        "    <AutoProcProgramAttachment>\n" +
        "      <fileType>Result</fileType>\n" +
        "      <fileName>truncate_early-unique.mtz</fileName>\n" +
        "      <filePath>/data/luke/test/Blue52/processing/./Blue52</filePath>\n" +
        "    </AutoProcProgramAttachment>\n" +
        "    <AutoProcProgramAttachment>\n" +
        "      <fileType>Result</fileType>\n" +
        "      <fileName>truncate_late-unique.mtz</fileName>\n" +
        "      <filePath>/data/luke/test/Blue52/processing/./Blue52</filePath>\n" +
        "    </AutoProcProgramAttachment>\n" +
        "  </AutoProcProgramContainer>\n" +
        "  <AutoProc>\n" +
        "    <spaceGroup>P 41 21 2</spaceGroup>\n" +
        "    <wavelength>1.00000</wavelength>\n" +
        "    <refinedCell_a>57.849</refinedCell_a>\n" +
        "    <refinedCell_b>57.849</refinedCell_b>\n" +
        "    <refinedCell_c>150.137</refinedCell_c>\n" +
        "    <refinedCell_alpha>90.000</refinedCell_alpha>\n" +
        "    <refinedCell_beta>90.000</refinedCell_beta>\n" +
        "    <refinedCell_gamma>90.000</refinedCell_gamma>\n" +
        "  </AutoProc>\n" +
        "  <AutoProcScalingContainer>\n" +
        "    <AutoProcScaling>\n" +
        "      <recordTimeStamp>2017-06-12 15:22:24</recordTimeStamp>\n" +
        "    </AutoProcScaling>\n" +
        "    <AutoProcScalingStatistics>\n" +
        "      <scalingStatisticsType>overall</scalingStatisticsType>\n" +
        "      <resolutionLimitLow>57.849</resolutionLimitLow>\n" +
        "      <resolutionLimitHigh>1.340</resolutionLimitHigh>\n" +
        "      <rMerge>.057</rMerge>\n" +
        "      <rMeasWithinIPlusIMinus>.062</rMeasWithinIPlusIMinus>\n" +
        "      <rMeasAllIPlusIMinus>.060</rMeasAllIPlusIMinus>\n" +
        "      <rPimWithinIPlusIMinus>.024</rPimWithinIPlusIMinus>\n" +
        "      <rPimAllIPlusIMinus>.017</rPimAllIPlusIMinus>\n" +
        "      <nTotalObservations>691434</nTotalObservations>\n" +
        "      <nTotalUniqueObservations>58223</nTotalUniqueObservations>\n" +
        "      <meanIOverSigI>22.3</meanIOverSigI>\n" +
        "      <completeness>99.9</completeness>\n" +
        "      <multiplicity>11.9</multiplicity>\n" +
        "      <ccHalf>1.000</ccHalf>\n" +
        "      <anomalousCompleteness>99.6</anomalousCompleteness>\n" +
        "      <anomalousMultiplicity>6.3</anomalousMultiplicity>\n" +
        "      <ccAnomalous>.016</ccAnomalous>\n" +
        "      <DanoOverSigDano>.720</DanoOverSigDano>\n" +
        "    </AutoProcScalingStatistics>\n" +
        "    <AutoProcScalingStatistics>\n" +
        "      <scalingStatisticsType>innerShell</scalingStatisticsType>\n" +
        "      <resolutionLimitLow>57.849</resolutionLimitLow>\n" +
        "      <resolutionLimitHigh>3.637</resolutionLimitHigh>\n" +
        "      <rMerge>.031</rMerge>\n" +
        "      <rMeasWithinIPlusIMinus>.032</rMeasWithinIPlusIMinus>\n" +
        "      <rMeasAllIPlusIMinus>.032</rMeasAllIPlusIMinus>\n" +
        "      <rPimWithinIPlusIMinus>.012</rPimWithinIPlusIMinus>\n" +
        "      <rPimAllIPlusIMinus>.009</rPimAllIPlusIMinus>\n" +
        "      <nTotalObservations>35907</nTotalObservations>\n" +
        "      <nTotalUniqueObservations>3237</nTotalUniqueObservations>\n" +
        "      <meanIOverSigI>61.9</meanIOverSigI>\n" +
        "      <completeness>100.0</completeness>\n" +
        "      <multiplicity>11.1</multiplicity>\n" +
        "      <ccHalf>1.000</ccHalf>\n" +
        "      <anomalousCompleteness>99.8</anomalousCompleteness>\n" +
        "      <anomalousMultiplicity>6.6</anomalousMultiplicity>\n" +
        "      <ccAnomalous>-.016</ccAnomalous>\n" +
        "      <DanoOverSigDano>.956</DanoOverSigDano>\n" +
        "    </AutoProcScalingStatistics>\n" +
        "    <AutoProcScalingStatistics>\n" +
        "      <scalingStatisticsType>outerShell</scalingStatisticsType>\n" +
        "      <resolutionLimitLow>1.363</resolutionLimitLow>\n" +
        "      <resolutionLimitHigh>1.341</resolutionLimitHigh>\n" +
        "      <rMerge>.676</rMerge>\n" +
        "      <rMeasWithinIPlusIMinus>.737</rMeasWithinIPlusIMinus>\n" +
        "      <rMeasAllIPlusIMinus>.731</rMeasAllIPlusIMinus>\n" +
        "      <rPimWithinIPlusIMinus>.371</rPimWithinIPlusIMinus>\n" +
        "      <rPimAllIPlusIMinus>.272</rPimAllIPlusIMinus>\n" +
        "      <nTotalObservations>19156</nTotalObservations>\n" +
        "      <nTotalUniqueObservations>2787</nTotalUniqueObservations>\n" +
        "      <meanIOverSigI>2.3</meanIOverSigI>\n" +
        "      <completeness>96.6</completeness>\n" +
        "      <multiplicity>6.9</multiplicity>\n" +
        "      <ccHalf>.861</ccHalf>\n" +
        "      <anomalousCompleteness>94.9</anomalousCompleteness>\n" +
        "      <anomalousMultiplicity>3.6</anomalousMultiplicity>\n" +
        "      <ccAnomalous>-.058</ccAnomalous>\n" +
        "      <DanoOverSigDano>.624</DanoOverSigDano>\n" +
        "    </AutoProcScalingStatistics>\n" +
        "    <AutoProcIntegrationContainer>\n" +
        "      <Image>\n" +
        "        <fileName>Blue52_0001.cbf</fileName>\n" +
        "        <fileLocation>/data/luke/test/Blue52</fileLocation>\n" +
        "      </Image>\n" +
        "      <AutoProcIntegration>\n" +
        "        <cell_a>57.8495</cell_a>\n" +
        "        <cell_b>57.8495</cell_b>\n" +
        "        <cell_c>150.1371</cell_c>\n" +
        "        <cell_alpha>90.000</cell_alpha>\n" +
        "        <cell_beta>90.000</cell_beta>\n" +
        "        <cell_gamma>90.000</cell_gamma>\n" +
        "        <startImageNumber>1</startImageNumber>\n" +
        "        <endImageNumber>720</endImageNumber>\n" +
        "        <refinedDetectorDistance>250.022934</refinedDetectorDistance>\n" +
        "        <refinedXBeam>1252.86</refinedXBeam>\n" +
        "        <refinedYBeam>1288.70</refinedYBeam>\n" +
        "        <rotationAxisX>0.999992</rotationAxisX>\n" +
        "        <rotationAxisY>0.003620</rotationAxisY>\n" +
        "        <rotationAxisZ>-0.001720</rotationAxisZ>\n" +
        "        <beamVectorX>-0.005269</beamVectorX>\n" +
        "        <beamVectorY>0.001129</beamVectorY>\n" +
        "        <beamVectorZ>0.999985</beamVectorZ>\n" +
        "      </AutoProcIntegration>\n" +
        "    </AutoProcIntegrationContainer>\n" +
        "  </AutoProcScalingContainer>\n" +
        "</AutoProcContainer>\n";
    /* @formatter:on */
  }

  private static class ProcessRunnerStub extends ProcessRunner {
    public ProcessRunnerStub() {
      super(null);
    }
  }
}
