/*
 * Copyright (c) 2015 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import java.io.IOException;
import java.nio.file.FileSystem;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.attribute.FileTime;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import com.github.marschall.memoryfilesystem.MemoryFileSystemBuilder;

/**
 * Tests {@link AutoProcResultParser}.
 */
public class AutoProcResultParserTest {
  private FileSystem fs;

  public AutoProcResultParserTest() {
    super();
    fs = null;
  }

  @Before
  public void setUp() throws IOException {
    fs = MemoryFileSystemBuilder.newLinux().build("test");
  }

  @After
  public void tearDown() throws Throwable {
    Utilities.closeSilently(fs);
  }

  @Test
  public void testResultLastModifiedTime_emptyProcessingDir() throws IOException {
    Path tmpDir = fs.getPath("/tmp");
    Files.createDirectory(tmpDir);
    assertNull(AutoProcResultParser.resultLastModifiedTime(tmpDir));
  }

  @Test
  public void testResultLastModifiedTime_noProcessingLog() throws IOException {
    Path tmpDir = fs.getPath("/tmp");
    Files.createDirectory(tmpDir);
    Files.createFile(tmpDir.resolve("process.X0150730_171231.log"));
    assertNull(AutoProcResultParser.resultLastModifiedTime(tmpDir));
  }

  @Test
  public void testResultLastModifiedTime_oneProcessLog() throws IOException {
    Path tmpDir = fs.getPath("/tmp");
    Files.createDirectory(tmpDir);
    Files.createFile(tmpDir.resolve("process.20150730_171231.log"));
    assertNotNull(AutoProcResultParser.resultLastModifiedTime(tmpDir));
  }

  @Test
  public void testResultLastModifiedTime_twoProcessLogs() throws IOException {
    Path tmpDir = fs.getPath("/tmp");
    Files.createDirectory(tmpDir);
    Path log1 = tmpDir.resolve("process.20150730_171231.log");
    Path log2 = tmpDir.resolve("process.20150731_114323.log");
    Files.createFile(log1);
    Files.createFile(log2);
    /* Typical last-modified ordering. */
    FileTime log1Time = FileTime.fromMillis(1000);
    FileTime log2Time = FileTime.fromMillis(2000);
    Files.setLastModifiedTime(log1, log1Time);
    Files.setLastModifiedTime(log2, log2Time);
    assertEquals(log2Time, AutoProcResultParser.resultLastModifiedTime(tmpDir));
    /*
     * The first log is more recently modified, but only the file name is used
     * to select the log, so the second log is still selected.
     */
    log1Time = FileTime.fromMillis(2000);
    log2Time = FileTime.fromMillis(1000);
    Files.setLastModifiedTime(log1, log1Time);
    Files.setLastModifiedTime(log2, log2Time);
    assertEquals(log2Time, AutoProcResultParser.resultLastModifiedTime(tmpDir));
  }
}
