/*
 * Copyright (c) 2017 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.math.BigDecimal;
import java.nio.file.FileSystem;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import com.github.marschall.memoryfilesystem.MemoryFileSystemBuilder;

/**
 * Tests {@link AutoProcXmlParser}.
 */
public class AutoProcXmlParserTest {
  private FileSystem fs;

  public AutoProcXmlParserTest() {
    super();
  }

  @Before
  public void setUp() throws IOException {
    fs = MemoryFileSystemBuilder.newLinux().build("test");
  }

  @After
  public void tearDown() throws Throwable {
    Utilities.closeSilently(fs);
  }

  @Test
  public void testParse_normal() throws IOException {
    Dataset d = new Dataset(fs.getPath("/tmp/data/luke/blue52"));
    DatasetProcessResult result = null;
    try (Reader reader = xml_normal()) {
      result = (new AutoProcXmlParser(d, reader)).parse();
    }
    assertNotNull(result);
    assertEquals(d, result.dataset());
    assertFalse(result.hasException());
    assertTrue(result.succeeded());
    assertEquals(new BigDecimal("57.849"), result.lowResolution());
    assertEquals(new BigDecimal("1.340"), result.highResolution());
    assertEquals(new BigDecimal("0.057"), result.rMerge());
    assertEquals(new BigDecimal("22.3"), result.iOverSigma());
    assertEquals(new BigDecimal("99.9"), result.completeness());
    assertEquals(new BigDecimal("11.9"), result.multiplicity());
    assertEquals(newUnitCell("57.849", "57.849", "150.137", "90.000", "90.000", "90.000"),
        result.unitCell());
    assertEquals("P 41 21 2", result.spaceGroup());
  }

  @Test
  public void testParse_overallScalingStatsNotFirst() throws IOException {
    Dataset d = new Dataset(fs.getPath("/tmp/data/luke/blue52"));
    DatasetProcessResult result = null;
    try (Reader reader = xml_overallScalingStatsNotFirst()) {
      result = (new AutoProcXmlParser(d, reader)).parse();
    }
    assertNotNull(result);
    assertEquals(d, result.dataset());
    assertFalse(result.hasException());
    assertTrue(result.succeeded());
    assertEquals(new BigDecimal("57.849"), result.lowResolution());
    assertEquals(new BigDecimal("1.340"), result.highResolution());
    assertEquals(new BigDecimal("0.057"), result.rMerge());
    assertEquals(new BigDecimal("22.3"), result.iOverSigma());
    assertEquals(new BigDecimal("99.9"), result.completeness());
    assertEquals(new BigDecimal("11.9"), result.multiplicity());
    assertEquals(newUnitCell("57.849", "57.849", "150.137", "90.000", "90.000", "90.000"),
        result.unitCell());
    assertEquals("P 41 21 2", result.spaceGroup());
  }

  @Test
  public void testParse_empty() throws IOException {
    Dataset d = new Dataset(fs.getPath("/tmp/data/luke/blue52"));
    DatasetProcessResult result = null;
    try (Reader reader = xml_empty()) {
      result = (new AutoProcXmlParser(d, reader)).parse();
    }
    assertNotNull(result);
    assertEquals(d, result.dataset());
    assertTrue(result.hasException());
    assertFalse(result.succeeded());
    assertNull(result.lowResolution());
    assertNull(result.highResolution());
    assertNull(result.rMerge());
    assertNull(result.iOverSigma());
    assertNull(result.completeness());
    assertNull(result.multiplicity());
    assertNull(result.unitCell());
    assertNull(result.spaceGroup());
  }

  @Test
  public void testParse_failedToProcess() throws IOException {
    Dataset d = new Dataset(fs.getPath("/tmp/data/luke/blue52"));
    DatasetProcessResult result = null;
    try (Reader reader = xml_failedToProcess()) {
      result = (new AutoProcXmlParser(d, reader)).parse();
    }
    assertNotNull(result);
    assertEquals(d, result.dataset());
    assertTrue(result.hasException());
    assertTrue(result.exception() instanceof DataFormatException);
    assertFalse(result.succeeded());
  }

  @Test
  public void testParse_noOverallScalingStats() throws IOException {
    Dataset d = new Dataset(fs.getPath("/tmp/data/luke/blue52"));
    DatasetProcessResult result = null;
    try (Reader reader = xml_noOverallScalingStats()) {
      result = (new AutoProcXmlParser(d, reader)).parse();
    }
    assertNotNull(result);
    assertEquals(d, result.dataset());
    assertTrue(result.hasException());
    assertTrue(result.exception() instanceof DataFormatException);
    assertFalse(result.succeeded());
  }

  @Test
  public void testParse_corrupt() throws IOException {
    Dataset d = new Dataset(fs.getPath("/tmp/data/luke/blue52"));
    DatasetProcessResult result = null;
    try (Reader reader = xml_corrupt()) {
      result = (new AutoProcXmlParser(d, reader)).parse();
    }
    assertNotNull(result);
    assertEquals(d, result.dataset());
    assertTrue(result.hasException());
    assertTrue(result.exception() instanceof DataFormatException);
    assertFalse(result.succeeded());
  }

  private static UnitCell newUnitCell(String... lengthsAndAngles) {
    BigDecimal[] bigLengthsAndAngles = new BigDecimal[lengthsAndAngles.length];
    for (int i = 0; i < lengthsAndAngles.length; i++) {
      bigLengthsAndAngles[i] = new BigDecimal(lengthsAndAngles[i]);
    }
    return new UnitCell(bigLengthsAndAngles);
  }

  private static Reader xml_normal() {
    /* @formatter:off */
    return new StringReader(
        "<?xml version=\"1.0\"?>\n" +
        "<AutoProcContainer>\n" +
        "  <AutoProcProgramContainer>\n" +
        "    <AutoProcProgram>\n" +
        "      <processingCommandLine>/auto/nfs/software/x86_64/encap/autoPROC_snapshot_20160708/autoPROC/bin/linux64/process -I .. -B -nthreads 6 -xml</processingCommandLine>\n" +
        "      <processingPrograms>autoPROC 1.1.7 (see: http://www.globalphasing.com/autoproc/)</processingPrograms>\n" +
        "      <processingStatus>1</processingStatus>\n" +
        "      <processingMessage></processingMessage>\n" +
        "      <processingStartTime>Mon Jun 12 14:55:22 CDT 2017</processingStartTime>\n" +
        "      <processingEndTime>Mon Jun 12 15:28:13 CDT 2017</processingEndTime>\n" +
        "      <processingEnvironment>\n" +
        "  Host      : r2d2.example.org\n" +
        "  OS        : PRETTY_NAME=\"Red Hat Enterprise Linux\"||REDHAT_BUGZILLA_PRODUCT=\"Red Hat Enterprise Linux 7\"|REDHAT_BUGZILLA_PRODUCT_VERSION=7.3|REDHAT_SUPPORT_PRODUCT=\"Red Hat Enterprise Linux\"|REDHAT_SUPPORT_PRODUCT_VERSION=\"7.3\"|Red Hat Enterprise Linux Workstation release 7.3 (Maipo)|x86_64|3.10.0-514.21.1.el7.x86_64\n" +
        "  User      : luke (group = luke)\n" +
        "  Directory : /data/luke/test/Blue52/processing\n" +
        "  Date      : Mon Jun 12 14:55:22 CDT 2017\n" +
        "  autoPROC  : /imcacat/software/x86_64/encap/autoPROC_snapshot_20160708\n" +
        "      </processingEnvironment>\n" +
        "    </AutoProcProgram>\n" +
        "    <AutoProcProgramAttachment>\n" +
        "      <fileType>Log</fileType>\n" +
        "      <fileName>summary.html</fileName>\n" +
        "      <filePath>/data/luke/test/Blue52/processing</filePath>\n" +
        "    </AutoProcProgramAttachment>\n" +
        "    <AutoProcProgramAttachment>\n" +
        "      <fileType>Result</fileType>\n" +
        "      <fileName>truncate-unique.mtz</fileName>\n" +
        "      <filePath>/data/luke/test/Blue52/processing/./Blue52</filePath>\n" +
        "    </AutoProcProgramAttachment>\n" +
        "    <AutoProcProgramAttachment>\n" +
        "      <fileType>Result</fileType>\n" +
        "      <fileName>truncate_early-unique.mtz</fileName>\n" +
        "      <filePath>/data/luke/test/Blue52/processing/./Blue52</filePath>\n" +
        "    </AutoProcProgramAttachment>\n" +
        "    <AutoProcProgramAttachment>\n" +
        "      <fileType>Result</fileType>\n" +
        "      <fileName>truncate_late-unique.mtz</fileName>\n" +
        "      <filePath>/data/luke/test/Blue52/processing/./Blue52</filePath>\n" +
        "    </AutoProcProgramAttachment>\n" +
        "  </AutoProcProgramContainer>\n" +
        "  <AutoProc>\n" +
        "    <spaceGroup>P 41 21 2</spaceGroup>\n" +
        "    <wavelength>1.00000</wavelength>\n" +
        "    <refinedCell_a>57.849</refinedCell_a>\n" +
        "    <refinedCell_b>57.849</refinedCell_b>\n" +
        "    <refinedCell_c>150.137</refinedCell_c>\n" +
        "    <refinedCell_alpha>90.000</refinedCell_alpha>\n" +
        "    <refinedCell_beta>90.000</refinedCell_beta>\n" +
        "    <refinedCell_gamma>90.000</refinedCell_gamma>\n" +
        "  </AutoProc>\n" +
        "  <AutoProcScalingContainer>\n" +
        "    <AutoProcScaling>\n" +
        "      <recordTimeStamp>2017-06-12 15:22:24</recordTimeStamp>\n" +
        "    </AutoProcScaling>\n" +
        "    <AutoProcScalingStatistics>\n" +
        "      <scalingStatisticsType>overall</scalingStatisticsType>\n" +
        "      <resolutionLimitLow>57.849</resolutionLimitLow>\n" +
        "      <resolutionLimitHigh>1.340</resolutionLimitHigh>\n" +
        "      <rMerge>.057</rMerge>\n" +
        "      <rMeasWithinIPlusIMinus>.062</rMeasWithinIPlusIMinus>\n" +
        "      <rMeasAllIPlusIMinus>.060</rMeasAllIPlusIMinus>\n" +
        "      <rPimWithinIPlusIMinus>.024</rPimWithinIPlusIMinus>\n" +
        "      <rPimAllIPlusIMinus>.017</rPimAllIPlusIMinus>\n" +
        "      <nTotalObservations>691434</nTotalObservations>\n" +
        "      <nTotalUniqueObservations>58223</nTotalUniqueObservations>\n" +
        "      <meanIOverSigI>22.3</meanIOverSigI>\n" +
        "      <completeness>99.9</completeness>\n" +
        "      <multiplicity>11.9</multiplicity>\n" +
        "      <ccHalf>1.000</ccHalf>\n" +
        "      <anomalousCompleteness>99.6</anomalousCompleteness>\n" +
        "      <anomalousMultiplicity>6.3</anomalousMultiplicity>\n" +
        "      <ccAnomalous>.016</ccAnomalous>\n" +
        "      <DanoOverSigDano>.720</DanoOverSigDano>\n" +
        "    </AutoProcScalingStatistics>\n" +
        "    <AutoProcScalingStatistics>\n" +
        "      <scalingStatisticsType>innerShell</scalingStatisticsType>\n" +
        "      <resolutionLimitLow>57.849</resolutionLimitLow>\n" +
        "      <resolutionLimitHigh>3.637</resolutionLimitHigh>\n" +
        "      <rMerge>.031</rMerge>\n" +
        "      <rMeasWithinIPlusIMinus>.032</rMeasWithinIPlusIMinus>\n" +
        "      <rMeasAllIPlusIMinus>.032</rMeasAllIPlusIMinus>\n" +
        "      <rPimWithinIPlusIMinus>.012</rPimWithinIPlusIMinus>\n" +
        "      <rPimAllIPlusIMinus>.009</rPimAllIPlusIMinus>\n" +
        "      <nTotalObservations>35907</nTotalObservations>\n" +
        "      <nTotalUniqueObservations>3237</nTotalUniqueObservations>\n" +
        "      <meanIOverSigI>61.9</meanIOverSigI>\n" +
        "      <completeness>100.0</completeness>\n" +
        "      <multiplicity>11.1</multiplicity>\n" +
        "      <ccHalf>1.000</ccHalf>\n" +
        "      <anomalousCompleteness>99.8</anomalousCompleteness>\n" +
        "      <anomalousMultiplicity>6.6</anomalousMultiplicity>\n" +
        "      <ccAnomalous>-.016</ccAnomalous>\n" +
        "      <DanoOverSigDano>.956</DanoOverSigDano>\n" +
        "    </AutoProcScalingStatistics>\n" +
        "    <AutoProcScalingStatistics>\n" +
        "      <scalingStatisticsType>outerShell</scalingStatisticsType>\n" +
        "      <resolutionLimitLow>1.363</resolutionLimitLow>\n" +
        "      <resolutionLimitHigh>1.341</resolutionLimitHigh>\n" +
        "      <rMerge>.676</rMerge>\n" +
        "      <rMeasWithinIPlusIMinus>.737</rMeasWithinIPlusIMinus>\n" +
        "      <rMeasAllIPlusIMinus>.731</rMeasAllIPlusIMinus>\n" +
        "      <rPimWithinIPlusIMinus>.371</rPimWithinIPlusIMinus>\n" +
        "      <rPimAllIPlusIMinus>.272</rPimAllIPlusIMinus>\n" +
        "      <nTotalObservations>19156</nTotalObservations>\n" +
        "      <nTotalUniqueObservations>2787</nTotalUniqueObservations>\n" +
        "      <meanIOverSigI>2.3</meanIOverSigI>\n" +
        "      <completeness>96.6</completeness>\n" +
        "      <multiplicity>6.9</multiplicity>\n" +
        "      <ccHalf>.861</ccHalf>\n" +
        "      <anomalousCompleteness>94.9</anomalousCompleteness>\n" +
        "      <anomalousMultiplicity>3.6</anomalousMultiplicity>\n" +
        "      <ccAnomalous>-.058</ccAnomalous>\n" +
        "      <DanoOverSigDano>.624</DanoOverSigDano>\n" +
        "    </AutoProcScalingStatistics>\n" +
        "    <AutoProcIntegrationContainer>\n" +
        "      <Image>\n" +
        "        <fileName>Blue52_0001.cbf</fileName>\n" +
        "        <fileLocation>/data/luke/test/Blue52</fileLocation>\n" +
        "      </Image>\n" +
        "      <AutoProcIntegration>\n" +
        "        <cell_a>57.8495</cell_a>\n" +
        "        <cell_b>57.8495</cell_b>\n" +
        "        <cell_c>150.1371</cell_c>\n" +
        "        <cell_alpha>90.000</cell_alpha>\n" +
        "        <cell_beta>90.000</cell_beta>\n" +
        "        <cell_gamma>90.000</cell_gamma>\n" +
        "        <startImageNumber>1</startImageNumber>\n" +
        "        <endImageNumber>720</endImageNumber>\n" +
        "        <refinedDetectorDistance>250.022934</refinedDetectorDistance>\n" +
        "        <refinedXBeam>1252.86</refinedXBeam>\n" +
        "        <refinedYBeam>1288.70</refinedYBeam>\n" +
        "        <rotationAxisX>0.999992</rotationAxisX>\n" +
        "        <rotationAxisY>0.003620</rotationAxisY>\n" +
        "        <rotationAxisZ>-0.001720</rotationAxisZ>\n" +
        "        <beamVectorX>-0.005269</beamVectorX>\n" +
        "        <beamVectorY>0.001129</beamVectorY>\n" +
        "        <beamVectorZ>0.999985</beamVectorZ>\n" +
        "      </AutoProcIntegration>\n" +
        "    </AutoProcIntegrationContainer>\n" +
        "  </AutoProcScalingContainer>\n" +
        "</AutoProcContainer>\n");
    /* @formatter:on */
  }

  private static Reader xml_overallScalingStatsNotFirst() {
    /* @formatter:off */
    return new StringReader(
        "<?xml version=\"1.0\"?>\n" +
        "<AutoProcContainer>\n" +
        "  <AutoProcProgramContainer>\n" +
        "    <AutoProcProgram>\n" +
        "      <processingCommandLine>/auto/nfs/software/x86_64/encap/autoPROC_snapshot_20160708/autoPROC/bin/linux64/process -I .. -B -nthreads 6 -xml</processingCommandLine>\n" +
        "      <processingPrograms>autoPROC 1.1.7 (see: http://www.globalphasing.com/autoproc/)</processingPrograms>\n" +
        "      <processingStatus>1</processingStatus>\n" +
        "      <processingMessage></processingMessage>\n" +
        "      <processingStartTime>Mon Jun 12 14:55:22 CDT 2017</processingStartTime>\n" +
        "      <processingEndTime>Mon Jun 12 15:28:13 CDT 2017</processingEndTime>\n" +
        "      <processingEnvironment>\n" +
        "  Host      : r2d2.example.org\n" +
        "  OS        : PRETTY_NAME=\"Red Hat Enterprise Linux\"||REDHAT_BUGZILLA_PRODUCT=\"Red Hat Enterprise Linux 7\"|REDHAT_BUGZILLA_PRODUCT_VERSION=7.3|REDHAT_SUPPORT_PRODUCT=\"Red Hat Enterprise Linux\"|REDHAT_SUPPORT_PRODUCT_VERSION=\"7.3\"|Red Hat Enterprise Linux Workstation release 7.3 (Maipo)|x86_64|3.10.0-514.21.1.el7.x86_64\n" +
        "  User      : luke (group = luke)\n" +
        "  Directory : /data/luke/test/Blue52/processing\n" +
        "  Date      : Mon Jun 12 14:55:22 CDT 2017\n" +
        "  autoPROC  : /imcacat/software/x86_64/encap/autoPROC_snapshot_20160708\n" +
        "      </processingEnvironment>\n" +
        "    </AutoProcProgram>\n" +
        "    <AutoProcProgramAttachment>\n" +
        "      <fileType>Log</fileType>\n" +
        "      <fileName>summary.html</fileName>\n" +
        "      <filePath>/data/luke/test/Blue52/processing</filePath>\n" +
        "    </AutoProcProgramAttachment>\n" +
        "    <AutoProcProgramAttachment>\n" +
        "      <fileType>Result</fileType>\n" +
        "      <fileName>truncate-unique.mtz</fileName>\n" +
        "      <filePath>/data/luke/test/Blue52/processing/./Blue52</filePath>\n" +
        "    </AutoProcProgramAttachment>\n" +
        "    <AutoProcProgramAttachment>\n" +
        "      <fileType>Result</fileType>\n" +
        "      <fileName>truncate_early-unique.mtz</fileName>\n" +
        "      <filePath>/data/luke/test/Blue52/processing/./Blue52</filePath>\n" +
        "    </AutoProcProgramAttachment>\n" +
        "    <AutoProcProgramAttachment>\n" +
        "      <fileType>Result</fileType>\n" +
        "      <fileName>truncate_late-unique.mtz</fileName>\n" +
        "      <filePath>/data/luke/test/Blue52/processing/./Blue52</filePath>\n" +
        "    </AutoProcProgramAttachment>\n" +
        "  </AutoProcProgramContainer>\n" +
        "  <AutoProc>\n" +
        "    <spaceGroup>P 41 21 2</spaceGroup>\n" +
        "    <wavelength>1.00000</wavelength>\n" +
        "    <refinedCell_a>57.849</refinedCell_a>\n" +
        "    <refinedCell_b>57.849</refinedCell_b>\n" +
        "    <refinedCell_c>150.137</refinedCell_c>\n" +
        "    <refinedCell_alpha>90.000</refinedCell_alpha>\n" +
        "    <refinedCell_beta>90.000</refinedCell_beta>\n" +
        "    <refinedCell_gamma>90.000</refinedCell_gamma>\n" +
        "  </AutoProc>\n" +
        "  <AutoProcScalingContainer>\n" +
        "    <AutoProcScaling>\n" +
        "      <recordTimeStamp>2017-06-12 15:22:24</recordTimeStamp>\n" +
        "    </AutoProcScaling>\n" +
        "    <AutoProcScalingStatistics>\n" +
        "      <scalingStatisticsType>innerShell</scalingStatisticsType>\n" +
        "      <resolutionLimitLow>58.849</resolutionLimitLow>\n" +
        "      <resolutionLimitHigh>4.637</resolutionLimitHigh>\n" +
        "      <rMerge>.131</rMerge>\n" +
        "      <rMeasWithinIPlusIMinus>.132</rMeasWithinIPlusIMinus>\n" +
        "      <rMeasAllIPlusIMinus>.132</rMeasAllIPlusIMinus>\n" +
        "      <rPimWithinIPlusIMinus>.112</rPimWithinIPlusIMinus>\n" +
        "      <rPimAllIPlusIMinus>.109</rPimAllIPlusIMinus>\n" +
        "      <nTotalObservations>35908</nTotalObservations>\n" +
        "      <nTotalUniqueObservations>3238</nTotalUniqueObservations>\n" +
        "      <meanIOverSigI>62.9</meanIOverSigI>\n" +
        "      <completeness>97.1</completeness>\n" +
        "      <multiplicity>11.2</multiplicity>\n" +
        "      <ccHalf>1.100</ccHalf>\n" +
        "      <anomalousCompleteness>98.8</anomalousCompleteness>\n" +
        "      <anomalousMultiplicity>6.7</anomalousMultiplicity>\n" +
        "      <ccAnomalous>-.026</ccAnomalous>\n" +
        "      <DanoOverSigDano>.856</DanoOverSigDano>\n" +
        "    </AutoProcScalingStatistics>\n" +
        "    <AutoProcScalingStatistics>\n" +
        "      <scalingStatisticsType>outerShell</scalingStatisticsType>\n" +
        "      <resolutionLimitLow>1.563</resolutionLimitLow>\n" +
        "      <resolutionLimitHigh>1.541</resolutionLimitHigh>\n" +
        "      <rMerge>.876</rMerge>\n" +
        "      <rMeasWithinIPlusIMinus>.937</rMeasWithinIPlusIMinus>\n" +
        "      <rMeasAllIPlusIMinus>.931</rMeasAllIPlusIMinus>\n" +
        "      <rPimWithinIPlusIMinus>.571</rPimWithinIPlusIMinus>\n" +
        "      <rPimAllIPlusIMinus>.472</rPimAllIPlusIMinus>\n" +
        "      <nTotalObservations>19158</nTotalObservations>\n" +
        "      <nTotalUniqueObservations>2789</nTotalUniqueObservations>\n" +
        "      <meanIOverSigI>2.5</meanIOverSigI>\n" +
        "      <completeness>98.6</completeness>\n" +
        "      <multiplicity>7.1</multiplicity>\n" +
        "      <ccHalf>.971</ccHalf>\n" +
        "      <anomalousCompleteness>95.1</anomalousCompleteness>\n" +
        "      <anomalousMultiplicity>3.8</anomalousMultiplicity>\n" +
        "      <ccAnomalous>-.258</ccAnomalous>\n" +
        "      <DanoOverSigDano>.824</DanoOverSigDano>\n" +
        "    </AutoProcScalingStatistics>\n" +
        "    <AutoProcScalingStatistics>\n" +
        "      <scalingStatisticsType>overall</scalingStatisticsType>\n" +
        "      <resolutionLimitLow>57.849</resolutionLimitLow>\n" +
        "      <resolutionLimitHigh>1.340</resolutionLimitHigh>\n" +
        "      <rMerge>.057</rMerge>\n" +
        "      <rMeasWithinIPlusIMinus>.062</rMeasWithinIPlusIMinus>\n" +
        "      <rMeasAllIPlusIMinus>.060</rMeasAllIPlusIMinus>\n" +
        "      <rPimWithinIPlusIMinus>.024</rPimWithinIPlusIMinus>\n" +
        "      <rPimAllIPlusIMinus>.017</rPimAllIPlusIMinus>\n" +
        "      <nTotalObservations>691434</nTotalObservations>\n" +
        "      <nTotalUniqueObservations>58223</nTotalUniqueObservations>\n" +
        "      <meanIOverSigI>22.3</meanIOverSigI>\n" +
        "      <completeness>99.9</completeness>\n" +
        "      <multiplicity>11.9</multiplicity>\n" +
        "      <ccHalf>1.000</ccHalf>\n" +
        "      <anomalousCompleteness>99.6</anomalousCompleteness>\n" +
        "      <anomalousMultiplicity>6.3</anomalousMultiplicity>\n" +
        "      <ccAnomalous>.016</ccAnomalous>\n" +
        "      <DanoOverSigDano>.720</DanoOverSigDano>\n" +
        "    </AutoProcScalingStatistics>\n" +
        "    <AutoProcIntegrationContainer>\n" +
        "      <Image>\n" +
        "        <fileName>Blue52_0001.cbf</fileName>\n" +
        "        <fileLocation>/data/luke/test/Blue52</fileLocation>\n" +
        "      </Image>\n" +
        "      <AutoProcIntegration>\n" +
        "        <cell_a>57.8495</cell_a>\n" +
        "        <cell_b>57.8495</cell_b>\n" +
        "        <cell_c>150.1371</cell_c>\n" +
        "        <cell_alpha>90.000</cell_alpha>\n" +
        "        <cell_beta>90.000</cell_beta>\n" +
        "        <cell_gamma>90.000</cell_gamma>\n" +
        "        <startImageNumber>1</startImageNumber>\n" +
        "        <endImageNumber>720</endImageNumber>\n" +
        "        <refinedDetectorDistance>250.022934</refinedDetectorDistance>\n" +
        "        <refinedXBeam>1252.86</refinedXBeam>\n" +
        "        <refinedYBeam>1288.70</refinedYBeam>\n" +
        "        <rotationAxisX>0.999992</rotationAxisX>\n" +
        "        <rotationAxisY>0.003620</rotationAxisY>\n" +
        "        <rotationAxisZ>-0.001720</rotationAxisZ>\n" +
        "        <beamVectorX>-0.005269</beamVectorX>\n" +
        "        <beamVectorY>0.001129</beamVectorY>\n" +
        "        <beamVectorZ>0.999985</beamVectorZ>\n" +
        "      </AutoProcIntegration>\n" +
        "    </AutoProcIntegrationContainer>\n" +
        "  </AutoProcScalingContainer>\n" +
        "</AutoProcContainer>\n");
    /* @formatter:on */
  }

  private static Reader xml_failedToProcess() {
    /* @formatter:off */
    return new StringReader(
        "<?xml version=\"1.0\"?>\n" +
        "<AutoProcContainer>\n" +
        "  <AutoProcProgramContainer>\n" +
        "    <AutoProcProgram>\n" +
        "      <processingCommandLine>/auto/nfs/software/x86_64/encap/autoPROC_snapshot_20160708/autoPROC/bin/linux64/process -I .. -B -nthreads 6 -xml</processingCommandLine>\n" +
        "      <processingPrograms>autoPROC 1.1.7 (see: http://www.globalphasing.com/autoproc/)</processingPrograms>\n" +
        "      <processingStatus>1</processingStatus>\n" +
        "      <processingMessage></processingMessage>\n" +
        "      <processingStartTime>Mon Jun 12 14:55:22 CDT 2017</processingStartTime>\n" +
        "      <processingEndTime>Mon Jun 12 15:28:13 CDT 2017</processingEndTime>\n" +
        "      <processingEnvironment>\n" +
        "  Host      : r2d2.example.org\n" +
        "  OS        : PRETTY_NAME=\"Red Hat Enterprise Linux\"||REDHAT_BUGZILLA_PRODUCT=\"Red Hat Enterprise Linux 7\"|REDHAT_BUGZILLA_PRODUCT_VERSION=7.3|REDHAT_SUPPORT_PRODUCT=\"Red Hat Enterprise Linux\"|REDHAT_SUPPORT_PRODUCT_VERSION=\"7.3\"|Red Hat Enterprise Linux Workstation release 7.3 (Maipo)|x86_64|3.10.0-514.21.1.el7.x86_64\n" +
        "  User      : luke (group = luke)\n" +
        "  Directory : /data/luke/test/Blue52/processing\n" +
        "  Date      : Mon Jun 12 14:55:22 CDT 2017\n" +
        "  autoPROC  : /imcacat/software/x86_64/encap/autoPROC_snapshot_20160708\n" +
        "      </processingEnvironment>\n" +
        "    </AutoProcProgram>\n" +
        "    <AutoProcProgramAttachment>\n" +
        "      <fileType>Log</fileType>\n" +
        "      <fileName>summary.html</fileName>\n" +
        "      <filePath>/data/luke/test/Blue52/processing</filePath>\n" +
        "    </AutoProcProgramAttachment>\n" +
        "  </AutoProcProgramContainer>\n" +
        "</AutoProcContainer>\n");
    /* @formatter:on */
  }

  private static Reader xml_noOverallScalingStats() {
    /* @formatter:off */
    return new StringReader(
        "<?xml version=\"1.0\"?>\n" +
        "<AutoProcContainer>\n" +
        "  <AutoProcProgramContainer>\n" +
        "    <AutoProcProgram>\n" +
        "      <processingCommandLine>/auto/nfs/software/x86_64/encap/autoPROC_snapshot_20160708/autoPROC/bin/linux64/process -I .. -B -nthreads 6 -xml</processingCommandLine>\n" +
        "      <processingPrograms>autoPROC 1.1.7 (see: http://www.globalphasing.com/autoproc/)</processingPrograms>\n" +
        "      <processingStatus>1</processingStatus>\n" +
        "      <processingMessage></processingMessage>\n" +
        "      <processingStartTime>Mon Jun 12 14:55:22 CDT 2017</processingStartTime>\n" +
        "      <processingEndTime>Mon Jun 12 15:28:13 CDT 2017</processingEndTime>\n" +
        "      <processingEnvironment>\n" +
        "  Host      : r2d2.example.org\n" +
        "  OS        : PRETTY_NAME=\"Red Hat Enterprise Linux\"||REDHAT_BUGZILLA_PRODUCT=\"Red Hat Enterprise Linux 7\"|REDHAT_BUGZILLA_PRODUCT_VERSION=7.3|REDHAT_SUPPORT_PRODUCT=\"Red Hat Enterprise Linux\"|REDHAT_SUPPORT_PRODUCT_VERSION=\"7.3\"|Red Hat Enterprise Linux Workstation release 7.3 (Maipo)|x86_64|3.10.0-514.21.1.el7.x86_64\n" +
        "  User      : luke (group = luke)\n" +
        "  Directory : /data/luke/test/Blue52/processing\n" +
        "  Date      : Mon Jun 12 14:55:22 CDT 2017\n" +
        "  autoPROC  : /imcacat/software/x86_64/encap/autoPROC_snapshot_20160708\n" +
        "      </processingEnvironment>\n" +
        "    </AutoProcProgram>\n" +
        "    <AutoProcProgramAttachment>\n" +
        "      <fileType>Log</fileType>\n" +
        "      <fileName>summary.html</fileName>\n" +
        "      <filePath>/data/luke/test/Blue52/processing</filePath>\n" +
        "    </AutoProcProgramAttachment>\n" +
        "    <AutoProcProgramAttachment>\n" +
        "      <fileType>Result</fileType>\n" +
        "      <fileName>truncate-unique.mtz</fileName>\n" +
        "      <filePath>/data/luke/test/Blue52/processing/./Blue52</filePath>\n" +
        "    </AutoProcProgramAttachment>\n" +
        "    <AutoProcProgramAttachment>\n" +
        "      <fileType>Result</fileType>\n" +
        "      <fileName>truncate_early-unique.mtz</fileName>\n" +
        "      <filePath>/data/luke/test/Blue52/processing/./Blue52</filePath>\n" +
        "    </AutoProcProgramAttachment>\n" +
        "    <AutoProcProgramAttachment>\n" +
        "      <fileType>Result</fileType>\n" +
        "      <fileName>truncate_late-unique.mtz</fileName>\n" +
        "      <filePath>/data/luke/test/Blue52/processing/./Blue52</filePath>\n" +
        "    </AutoProcProgramAttachment>\n" +
        "  </AutoProcProgramContainer>\n" +
        "  <AutoProc>\n" +
        "    <spaceGroup>P 41 21 2</spaceGroup>\n" +
        "    <wavelength>1.00000</wavelength>\n" +
        "    <refinedCell_a>57.849</refinedCell_a>\n" +
        "    <refinedCell_b>57.849</refinedCell_b>\n" +
        "    <refinedCell_c>150.137</refinedCell_c>\n" +
        "    <refinedCell_alpha>90.000</refinedCell_alpha>\n" +
        "    <refinedCell_beta>90.000</refinedCell_beta>\n" +
        "    <refinedCell_gamma>90.000</refinedCell_gamma>\n" +
        "  </AutoProc>\n" +
        "  <AutoProcScalingContainer>\n" +
        "    <AutoProcScaling>\n" +
        "      <recordTimeStamp>2017-06-12 15:22:24</recordTimeStamp>\n" +
        "    </AutoProcScaling>\n" +
        "    <AutoProcScalingStatistics>\n" +
        "      <scalingStatisticsType>innerShell</scalingStatisticsType>\n" +
        "      <resolutionLimitLow>57.849</resolutionLimitLow>\n" +
        "      <resolutionLimitHigh>3.637</resolutionLimitHigh>\n" +
        "      <rMerge>.031</rMerge>\n" +
        "      <rMeasWithinIPlusIMinus>.032</rMeasWithinIPlusIMinus>\n" +
        "      <rMeasAllIPlusIMinus>.032</rMeasAllIPlusIMinus>\n" +
        "      <rPimWithinIPlusIMinus>.012</rPimWithinIPlusIMinus>\n" +
        "      <rPimAllIPlusIMinus>.009</rPimAllIPlusIMinus>\n" +
        "      <nTotalObservations>35907</nTotalObservations>\n" +
        "      <nTotalUniqueObservations>3237</nTotalUniqueObservations>\n" +
        "      <meanIOverSigI>61.9</meanIOverSigI>\n" +
        "      <completeness>100.0</completeness>\n" +
        "      <multiplicity>11.1</multiplicity>\n" +
        "      <ccHalf>1.000</ccHalf>\n" +
        "      <anomalousCompleteness>99.8</anomalousCompleteness>\n" +
        "      <anomalousMultiplicity>6.6</anomalousMultiplicity>\n" +
        "      <ccAnomalous>-.016</ccAnomalous>\n" +
        "      <DanoOverSigDano>.956</DanoOverSigDano>\n" +
        "    </AutoProcScalingStatistics>\n" +
        "    <AutoProcScalingStatistics>\n" +
        "      <scalingStatisticsType>outerShell</scalingStatisticsType>\n" +
        "      <resolutionLimitLow>1.363</resolutionLimitLow>\n" +
        "      <resolutionLimitHigh>1.341</resolutionLimitHigh>\n" +
        "      <rMerge>.676</rMerge>\n" +
        "      <rMeasWithinIPlusIMinus>.737</rMeasWithinIPlusIMinus>\n" +
        "      <rMeasAllIPlusIMinus>.731</rMeasAllIPlusIMinus>\n" +
        "      <rPimWithinIPlusIMinus>.371</rPimWithinIPlusIMinus>\n" +
        "      <rPimAllIPlusIMinus>.272</rPimAllIPlusIMinus>\n" +
        "      <nTotalObservations>19156</nTotalObservations>\n" +
        "      <nTotalUniqueObservations>2787</nTotalUniqueObservations>\n" +
        "      <meanIOverSigI>2.3</meanIOverSigI>\n" +
        "      <completeness>96.6</completeness>\n" +
        "      <multiplicity>6.9</multiplicity>\n" +
        "      <ccHalf>.861</ccHalf>\n" +
        "      <anomalousCompleteness>94.9</anomalousCompleteness>\n" +
        "      <anomalousMultiplicity>3.6</anomalousMultiplicity>\n" +
        "      <ccAnomalous>-.058</ccAnomalous>\n" +
        "      <DanoOverSigDano>.624</DanoOverSigDano>\n" +
        "    </AutoProcScalingStatistics>\n" +
        "    <AutoProcIntegrationContainer>\n" +
        "      <Image>\n" +
        "        <fileName>Blue52_0001.cbf</fileName>\n" +
        "        <fileLocation>/data/luke/test/Blue52</fileLocation>\n" +
        "      </Image>\n" +
        "      <AutoProcIntegration>\n" +
        "        <cell_a>57.8495</cell_a>\n" +
        "        <cell_b>57.8495</cell_b>\n" +
        "        <cell_c>150.1371</cell_c>\n" +
        "        <cell_alpha>90.000</cell_alpha>\n" +
        "        <cell_beta>90.000</cell_beta>\n" +
        "        <cell_gamma>90.000</cell_gamma>\n" +
        "        <startImageNumber>1</startImageNumber>\n" +
        "        <endImageNumber>720</endImageNumber>\n" +
        "        <refinedDetectorDistance>250.022934</refinedDetectorDistance>\n" +
        "        <refinedXBeam>1252.86</refinedXBeam>\n" +
        "        <refinedYBeam>1288.70</refinedYBeam>\n" +
        "        <rotationAxisX>0.999992</rotationAxisX>\n" +
        "        <rotationAxisY>0.003620</rotationAxisY>\n" +
        "        <rotationAxisZ>-0.001720</rotationAxisZ>\n" +
        "        <beamVectorX>-0.005269</beamVectorX>\n" +
        "        <beamVectorY>0.001129</beamVectorY>\n" +
        "        <beamVectorZ>0.999985</beamVectorZ>\n" +
        "      </AutoProcIntegration>\n" +
        "    </AutoProcIntegrationContainer>\n" +
        "  </AutoProcScalingContainer>\n" +
        "</AutoProcContainer>\n");
    /* @formatter:on */
  }

  private static Reader xml_empty() {
    return new StringReader("");
  }

  private static Reader xml_corrupt() {
    return new StringReader("corrupt");
  }
}
