/*
 * Copyright (c) 2017 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNull;

import java.io.IOException;
import java.nio.file.FileSystem;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.attribute.FileTime;
import java.util.Arrays;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import org.imca_cat.procrun.InputStreamConsumerFactory;
import org.imca_cat.procrun.ProcessRunner;
import org.imca_cat.procrun.ProcessRunnerResult;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import com.github.marschall.memoryfilesystem.MemoryFileSystemBuilder;

/**
 * Tests {@link DatasetProcessService}.
 */
public class DatasetProcessServiceTest {
  private AaprocSystem sys;
  private FileSystem fs;
  private DatasetProcessorFactorySpy processorFactory;
  private ProcessRunnerFactorySpy runnerFactory;
  private DatasetProcessService service;

  public DatasetProcessServiceTest() {
    super();
    sys = null;
    fs = null;
    processorFactory = null;
    runnerFactory = null;
    service = null;
  }

  @Before
  public void setUp() throws IOException {
    sys = newAaprocSystem();
    fs = newFileSystem(sys);
    processorFactory = new DatasetProcessorFactorySpy();
    runnerFactory = new ProcessRunnerFactorySpy();
  }

  @After
  public void tearDown() throws Throwable {
    Utilities.closeSilently(service);
    Utilities.closeSilently(fs);
  }

  @Test
  public void testPostProcessHook_oneDataset() throws InterruptedException, IOException {
    Path datasetDir = fs.getPath("/tmp");
    Files.createDirectory(datasetDir);
    Path datasetFile1 = datasetDir.resolve("lys_0001.cbf");
    Files.createFile(datasetFile1);
    Files.createFile(datasetDir.resolve("lys_0002.cbf"));
    Files.createFile(datasetDir.resolve("lys_0003.cbf"));
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(3));
    b.postProcessHook(Arrays.asList("my-hook", "$dataset"));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetProcessService(c, sys, processorFactory, runnerFactory);
    service.start();
    service.put(new Dataset(datasetDir));
    service.take();
    ProcessRunnerSpy hook = runnerFactory.runners().take();
    hook.ran().await();
    assertEquals(Arrays.asList("my-hook", datasetFile1.toString()), hook.builder().command());
  }

  @Test
  public void testPostProcessHook_oneDatasetWithOtherArgs()
      throws InterruptedException, IOException {
    Path datasetDir = fs.getPath("/tmp");
    Files.createDirectory(datasetDir);
    Path datasetFile1 = datasetDir.resolve("lys_0001.cbf");
    Files.createFile(datasetFile1);
    Files.createFile(datasetDir.resolve("lys_0002.cbf"));
    Files.createFile(datasetDir.resolve("lys_0003.cbf"));
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(3));
    b.postProcessHook(Arrays.asList("my-hook", "--dataset", "$dataset", "post-process"));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetProcessService(c, sys, processorFactory, runnerFactory);
    service.start();
    service.put(new Dataset(datasetDir));
    service.take();
    ProcessRunnerSpy hook = runnerFactory.runners().take();
    hook.ran().await();
    assertEquals(Arrays.asList("my-hook", "--dataset", datasetFile1.toString(), "post-process"),
        hook.builder().command());
  }

  @Test
  public void testPostProcessHook_oneDatasetSortedByFilenameNotTimestamp()
      throws InterruptedException, IOException {
    Path datasetDir = fs.getPath("/tmp");
    Files.createDirectory(datasetDir);
    Path datasetFile3 = datasetDir.resolve("lys_0003.cbf");
    Files.createFile(datasetFile3);
    long lastModified = 1000L;
    Files.setLastModifiedTime(datasetFile3, FileTime.fromMillis(lastModified));
    Path datasetFile2 = datasetDir.resolve("lys_0002.cbf");
    Files.createFile(datasetFile2);
    lastModified += 1000L;
    Files.setLastModifiedTime(datasetFile2, FileTime.fromMillis(lastModified));
    Path datasetFile1 = datasetDir.resolve("lys_0001.cbf");
    Files.createFile(datasetFile1);
    lastModified += 1000L;
    Files.setLastModifiedTime(datasetFile1, FileTime.fromMillis(lastModified));
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(3));
    b.postProcessHook(Arrays.asList("my-hook", "$dataset"));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetProcessService(c, sys, processorFactory, runnerFactory);
    service.start();
    service.put(new Dataset(datasetDir));
    service.take();
    ProcessRunnerSpy hook = runnerFactory.runners().take();
    hook.ran().await();
    assertEquals(Arrays.asList("my-hook", datasetFile1.toString()), hook.builder().command());
  }

  @Test
  public void testPostProcessHook_twoDatasetsLastModifiedWins()
      throws InterruptedException, IOException {
    Path datasetDir = fs.getPath("/tmp");
    Files.createDirectory(datasetDir);

    Path datasetFileA1_1 = datasetDir.resolve("A1-lys_0001.cbf");
    Files.createFile(datasetFileA1_1);
    long lastModified = 1000L;
    Files.setLastModifiedTime(datasetFileA1_1, FileTime.fromMillis(lastModified));
    Path datasetFileA1_2 = datasetDir.resolve("A1-lys_0002.cbf");
    Files.createFile(datasetFileA1_2);
    lastModified += 1000L;
    Files.setLastModifiedTime(datasetFileA1_2, FileTime.fromMillis(lastModified));
    Path datasetFileA1_3 = datasetDir.resolve("A1-lys_0003.cbf");
    Files.createFile(datasetFileA1_3);
    lastModified += 1000L;
    Files.setLastModifiedTime(datasetFileA1_3, FileTime.fromMillis(lastModified));

    Path datasetFileA2_1 = datasetDir.resolve("A2-lys_0001.cbf");
    Files.createFile(datasetFileA2_1);
    lastModified += 1000L;
    Files.setLastModifiedTime(datasetFileA2_1, FileTime.fromMillis(lastModified));
    Path datasetFileA2_2 = datasetDir.resolve("A2-lys_0002.cbf");
    Files.createFile(datasetFileA2_2);
    lastModified += 1000L;
    Files.setLastModifiedTime(datasetFileA2_2, FileTime.fromMillis(lastModified));
    Path datasetFileA2_3 = datasetDir.resolve("A2-lys_0003.cbf");
    Files.createFile(datasetFileA2_3);
    lastModified += 1000L;
    Files.setLastModifiedTime(datasetFileA2_3, FileTime.fromMillis(lastModified));

    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(3));
    b.postProcessHook(Arrays.asList("my-hook", "$dataset"));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetProcessService(c, sys, processorFactory, runnerFactory);
    service.start();
    service.put(new Dataset(datasetDir));
    service.take();
    ProcessRunnerSpy hook = runnerFactory.runners().take();
    hook.ran().await();
    assertEquals(Arrays.asList("my-hook", datasetFileA2_1.toString()), hook.builder().command());
  }

  @Test
  public void testPostProcessHook_oneDatasetNoHook() throws InterruptedException, IOException {
    Path datasetDir = fs.getPath("/tmp");
    Files.createDirectory(datasetDir);
    Files.createFile(datasetDir.resolve("lys_0001.cbf"));
    Files.createFile(datasetDir.resolve("lys_0002.cbf"));
    Files.createFile(datasetDir.resolve("lys_0003.cbf"));
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(3));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetProcessService(c, sys, processorFactory, runnerFactory);
    service.start();
    service.put(new Dataset(datasetDir));
    service.take();
    assertNull(runnerFactory.runners().poll(100L, TimeUnit.MILLISECONDS));
  }

  @Test
  public void testPostProcessHook_noDatasetDueToNoFiles()
      throws InterruptedException, IOException, ExecutionException {
    Path datasetDir = fs.getPath("/tmp");
    Files.createDirectory(datasetDir);
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(3));
    b.postProcessHook(Arrays.asList("my-hook", "$dataset"));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetProcessService(c, sys, processorFactory, runnerFactory);
    service.start();
    service.put(new Dataset(datasetDir));
    service.take();
    assertNull(runnerFactory.runners().poll(100L, TimeUnit.MILLISECONDS));
  }

  @Test
  public void testPostProcessHook_noDatasetDueToNoMatchingFiles()
      throws InterruptedException, IOException, ExecutionException {
    Path datasetDir = fs.getPath("/tmp");
    Files.createDirectory(datasetDir);
    Files.createFile(datasetDir.resolve("lys_0001.txt"));
    Files.createFile(datasetDir.resolve("lys_0002.txt"));
    Files.createFile(datasetDir.resolve("lys_0003.txt"));
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(3));
    b.postProcessHook(Arrays.asList("my-hook", "$dataset"));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetProcessService(c, sys, processorFactory, runnerFactory);
    service.start();
    service.put(new Dataset(datasetDir));
    service.take();
    assertNull(runnerFactory.runners().poll(100L, TimeUnit.MILLISECONDS));
  }

  @Test
  public void testPostProcessHook_noDatasetDueToNotFiles()
      throws InterruptedException, IOException, ExecutionException {
    Path datasetDir = fs.getPath("/tmp");
    Files.createDirectory(datasetDir);
    Files.createDirectory(datasetDir.resolve("lys_0001.cbf"));
    Files.createDirectory(datasetDir.resolve("lys_0002.cbf"));
    Files.createDirectory(datasetDir.resolve("lys_0003.cbf"));
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(3));
    b.postProcessHook(Arrays.asList("my-hook", "$dataset"));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetProcessService(c, sys, processorFactory, runnerFactory);
    service.start();
    service.put(new Dataset(datasetDir));
    service.take();
    assertNull(runnerFactory.runners().poll(100L, TimeUnit.MILLISECONDS));
  }

  @Test
  public void testPostProcessHook_noDatasetDueToNoCommonBaseName()
      throws InterruptedException, IOException, ExecutionException {
    Path datasetDir = fs.getPath("/tmp");
    Files.createDirectory(datasetDir);
    Files.createFile(datasetDir.resolve("lys-a.cbf"));
    Files.createFile(datasetDir.resolve("lys-b.cbf"));
    Files.createFile(datasetDir.resolve("lys-c.cbf"));
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(3));
    b.postProcessHook(Arrays.asList("my-hook", "$dataset"));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetProcessService(c, sys, processorFactory, runnerFactory);
    service.start();
    service.put(new Dataset(datasetDir));
    service.take();
    assertNull(runnerFactory.runners().poll(100L, TimeUnit.MILLISECONDS));
  }

  @Test
  public void testPostProcessHook_noDatasetDueToNotEnoughFiles()
      throws InterruptedException, IOException {
    Path datasetDir = fs.getPath("/tmp");
    Files.createDirectory(datasetDir);
    Files.createFile(datasetDir.resolve("lys_0001.cbf"));
    Files.createFile(datasetDir.resolve("lys_0002.cbf"));
    Files.createFile(datasetDir.resolve("lys_0003.cbf"));
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(4));
    b.postProcessHook(Arrays.asList("my-hook", "$dataset"));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetProcessService(c, sys, processorFactory, runnerFactory);
    service.start();
    service.put(new Dataset(datasetDir));
    service.take();
    assertNull(runnerFactory.runners().poll(100L, TimeUnit.MILLISECONDS));
  }

  @Test
  public void testPostProcessHook_noDatasetDueToDiffCenterImages()
      throws InterruptedException, IOException {
    Path datasetDir = fs.getPath("/tmp");
    Files.createDirectory(datasetDir);
    Files.createFile(datasetDir.resolve("cell_zx_01_01.cbf"));
    Files.createFile(datasetDir.resolve("cell_zx_01_02.cbf"));
    Files.createFile(datasetDir.resolve("cell_zx_01_03.cbf"));
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(3));
    b.postProcessHook(Arrays.asList("my-hook", "$dataset"));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetProcessService(c, sys, processorFactory, runnerFactory);
    service.start();
    service.put(new Dataset(datasetDir));
    service.take();
    assertNull(runnerFactory.runners().poll(100L, TimeUnit.MILLISECONDS));
  }

  private static AaprocSystem newAaprocSystem() {
    return new AaprocSystem() {
      @Override
      public String aapAutoprocHelperPath() {
        return "/nonexistent";
      }

      @Override
      public String sysConfFilePath() {
        return "/nonexistent";
      }

      @Override
      public String userHomeDirectory() {
        return "/home/luke";
      }

      @Override
      public String userCurrentWorkingDirectory() {
        return userHomeDirectory();
      }

      @Override
      public String version() {
        return "0.0.0";
      }
    };
  }

  private static FileSystem newFileSystem(AaprocSystem sys) throws IOException {
    FileSystem result = MemoryFileSystemBuilder.newLinux().build("test");
    Path cwd = result.getPath(sys.userCurrentWorkingDirectory());
    Files.createDirectories(cwd);
    return result;
  }

  private static DatasetProcessService newDatasetProcessService(AapConfiguration c,
      AaprocSystem sys, DatasetProcessorFactory processorFactory,
      ProcessRunnerFactory runnerFactory) {
    DatasetProcessService result =
        new DatasetProcessService(c, sys, new DatasetProcessServiceExceptionHandler() {
          @Override
          public void exception(Throwable t) {
            t.printStackTrace();
          }
        });
    result.datasetProcessorFactory(processorFactory);
    result.processRunnerFactory(runnerFactory);
    return result;
  }

  private static class DatasetProcessorFactorySpy implements DatasetProcessorFactory {
    private final BlockingQueue<DatasetProcessorSpy> processors;

    public DatasetProcessorFactorySpy() {
      super();
      processors = new LinkedBlockingQueue<>();
    }

    @Override
    public DatasetProcessor newProcessor(Dataset dataset, AapConfiguration conf, AaprocSystem sys) {
      DatasetProcessorSpy result = new DatasetProcessorSpy(dataset, conf, sys);
      processors.add(result);
      return result;
    }

    @SuppressWarnings("unused")
    public BlockingQueue<DatasetProcessorSpy> processors() {
      return processors;
    }
  }

  private static class DatasetProcessorSpy extends DatasetProcessor {
    private final Dataset dataset;
    private CountDownLatch ran;

    public DatasetProcessorSpy(Dataset dataset, AapConfiguration conf, AaprocSystem sys) {
      super(dataset, conf, sys);
      this.dataset = dataset;
      ran = new CountDownLatch(1);
    }

    @Override
    public DatasetProcessResult process() throws InterruptedException {
      ran.countDown();
      return new DatasetProcessResult(dataset, null);
    }

    @SuppressWarnings("unused")
    public Dataset dataset() {
      return dataset;
    }

    @SuppressWarnings("unused")
    public CountDownLatch ran() {
      return ran;
    }
  }

  private static class ProcessRunnerFactorySpy implements ProcessRunnerFactory {
    private final BlockingQueue<ProcessRunnerSpy> runners;

    public ProcessRunnerFactorySpy() {
      super();
      runners = new LinkedBlockingQueue<>();
    }

    @Override
    public ProcessRunner newRunner(NioProcessBuilder builder,
        InputStreamConsumerFactory stdoutConsumerFactory,
        InputStreamConsumerFactory stderrConsumerFactory) {
      ProcessRunnerSpy result =
          new ProcessRunnerSpy(builder, stdoutConsumerFactory, stderrConsumerFactory);
      runners.add(result);
      return result;
    }

    public BlockingQueue<ProcessRunnerSpy> runners() {
      return runners;
    }
  }

  private static class ProcessRunnerSpy extends ProcessRunner {
    private final NioProcessBuilder builder;
    private CountDownLatch ran;

    public ProcessRunnerSpy(NioProcessBuilder builder,
        InputStreamConsumerFactory stdoutConsumerFactory,
        InputStreamConsumerFactory stderrConsumerFactory) {
      super(builder.toProcessBuilder(), stdoutConsumerFactory, stderrConsumerFactory);
      this.builder = builder;
      ran = new CountDownLatch(1);
    }

    @Override
    public ProcessRunnerResult call() {
      ran.countDown();
      return new ProcessRunnerResult(0, null, null, null);
    }

    public NioProcessBuilder builder() {
      return builder;
    }

    public CountDownLatch ran() {
      return ran;
    }
  }
}
