/*
 * Copyright (c) 2014, 2015, 2016 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import static org.junit.Assert.assertEquals;

import java.io.IOException;
import java.nio.file.FileSystem;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.attribute.FileTime;
import java.util.Locale;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.regex.Pattern;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import com.github.marschall.memoryfilesystem.MemoryFileSystemBuilder;

/**
 * Tests {@link DatasetWatchService}.
 */
public class DatasetWatchServiceTest {
  private FileSystem fs;
  private DatasetWatchService service;
  private ExecutorService execService;

  public DatasetWatchServiceTest() {
    super();
    fs = null;
    service = null;
    execService = null;
  }

  @Before
  public void setUp() throws IOException {
    fs = MemoryFileSystemBuilder.newLinux().build("test");
  }

  @After
  public void tearDown() throws Throwable {
    Utilities.shutdownNowSilently(execService);
    Utilities.closeSilently(service);
    Utilities.closeSilently(fs);
  }

  @Test
  public void testClose_nothingRegistered() throws InterruptedException, IOException {
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(5));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetWatchService(c);
    service.start();
    service.close();
  }

  @Test
  public void testTake_preexistingDataset() throws InterruptedException, IOException {
    Path tmpDir = fs.getPath("/tmp");
    Files.createDirectory(tmpDir);
    Files.createFile(tmpDir.resolve("image_0001.cbf"));
    Files.createFile(tmpDir.resolve("image_0002.cbf"));
    Files.createFile(tmpDir.resolve("image_0003.cbf"));
    Files.createFile(tmpDir.resolve("image_0004.cbf"));
    Files.createFile(tmpDir.resolve("image_0005.cbf"));
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(5));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetWatchService(c);
    service.register(tmpDir);
    service.start();
    DatasetWatchResult r = service.take();
    Dataset d = r.dataset();
    assertEquals(d.directory(), tmpDir);
  }

  @Test
  public void testTake_preexistingDatasetAddedToWithoutMetadata()
      throws InterruptedException, IOException {
    Path tmpDir = fs.getPath("/tmp");
    Files.createDirectory(tmpDir);
    Files.createFile(tmpDir.resolve("image_0001.cbf"));
    Files.createFile(tmpDir.resolve("image_0002.cbf"));
    Files.createFile(tmpDir.resolve("image_0003.cbf"));
    Files.createFile(tmpDir.resolve("image_0004.cbf"));
    Files.createFile(tmpDir.resolve("image_0005.cbf"));
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.fsPollMetadataEnabled(false);
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(5));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetWatchService(c);
    service.register(tmpDir);
    service.start();
    DatasetWatchResult r = service.take();
    Dataset d = r.dataset();
    assertEquals(d.directory(), tmpDir);
    Files.createFile(tmpDir.resolve("image_0006.cbf"));
    r = service.take();
    d = r.dataset();
    assertEquals(d.directory(), tmpDir);
  }

  @Test
  public void testTake_preexistingDatasetInSubdir() throws InterruptedException, IOException {
    Path tmpDir = fs.getPath("/tmp");
    Files.createDirectory(tmpDir);
    Path tmpBlueDir = tmpDir.resolve("blue");
    Files.createDirectory(tmpBlueDir);
    Files.createFile(tmpBlueDir.resolve("image_0001.cbf"));
    Files.createFile(tmpBlueDir.resolve("image_0002.cbf"));
    Files.createFile(tmpBlueDir.resolve("image_0003.cbf"));
    Files.createFile(tmpBlueDir.resolve("image_0004.cbf"));
    Files.createFile(tmpBlueDir.resolve("image_0005.cbf"));
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(5));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetWatchService(c);
    service.register(tmpDir);
    service.start();
    DatasetWatchResult r = service.take();
    Dataset d = r.dataset();
    assertEquals(d.directory(), tmpBlueDir);
  }

  @Test(expected = TimeoutException.class)
  public void testTake_preexistingDatasetInExcludedSubdir()
      throws InterruptedException, IOException, ExecutionException, TimeoutException {
    Path tmpDir = fs.getPath("/tmp");
    Files.createDirectory(tmpDir);
    Path tmpBlueDir = tmpDir.resolve("blue");
    Files.createDirectory(tmpBlueDir);
    Files.createFile(tmpBlueDir.resolve("image_0001.cbf"));
    Files.createFile(tmpBlueDir.resolve("image_0002.cbf"));
    Files.createFile(tmpBlueDir.resolve("image_0003.cbf"));
    Files.createFile(tmpBlueDir.resolve("image_0004.cbf"));
    Files.createFile(tmpBlueDir.resolve("image_0005.cbf"));
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(5));
    b.excludeDirPat(Pattern.compile("camera.*|blue.*"));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetWatchService(c);
    service.register(tmpDir);
    service.start();
    execService = Executors.newSingleThreadExecutor();
    Future<DatasetWatchResult> f = execService.submit(new Callable<DatasetWatchResult>() {
      @Override
      public DatasetWatchResult call() throws Exception {
        return service.take();
      }
    });
    f.get(500L, TimeUnit.MILLISECONDS);
  }

  @Test(expected = TimeoutException.class)
  public void testTake_processingDirDeleted()
      throws InterruptedException, IOException, ExecutionException, TimeoutException {
    Path tmpDir = fs.getPath("/tmp");
    Files.createDirectory(tmpDir);
    Files.createFile(tmpDir.resolve("image_0001.cbf"));
    Files.createFile(tmpDir.resolve("image_0002.cbf"));
    Files.createFile(tmpDir.resolve("image_0003.cbf"));
    Files.createFile(tmpDir.resolve("image_0004.cbf"));
    Files.createFile(tmpDir.resolve("image_0005.cbf"));
    Path processingDir = tmpDir.resolve("processing");
    Files.createDirectory(processingDir);
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(5));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetWatchService(c);
    service.register(tmpDir);
    service.start();
    DatasetWatchResult r = service.take();
    Dataset d = r.dataset();
    assertEquals(d.directory(), tmpDir);
    Files.delete(processingDir);
    Thread.sleep(2 * c.fileSystemPollPeriod().to(TimeUnit.MILLISECONDS).amount());
    Files.createDirectory(processingDir);
    execService = Executors.newSingleThreadExecutor();
    Future<DatasetWatchResult> f = execService.submit(new Callable<DatasetWatchResult>() {
      @Override
      public DatasetWatchResult call() throws Exception {
        return service.take();
      }
    });
    f.get(500L, TimeUnit.MILLISECONDS);
  }

  @Test(expected = TimeoutException.class)
  public void testTake_ignoreDiffractionCenterImages()
      throws InterruptedException, IOException, TimeoutException, ExecutionException {
    Path tmpDir = fs.getPath("/tmp");
    Files.createDirectory(tmpDir);
    Files.createFile(tmpDir.resolve("cell_zx_01_01.cbf"));
    Files.createFile(tmpDir.resolve("cell_zx_01_02.cbf"));
    Files.createFile(tmpDir.resolve("cell_zx_01_03.cbf"));
    Files.createFile(tmpDir.resolve("cell_zx_fly_01_0001.cbf"));
    Files.createFile(tmpDir.resolve("cell_zx_fly_01_0002.cbf"));
    Files.createFile(tmpDir.resolve("cell_zx_fly_01_0003.cbf"));
    Files.createFile(tmpDir.resolve("cell_zx_fly_01_0001.log"));
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(1));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetWatchService(c);
    service.register(tmpDir);
    service.start();
    execService = Executors.newSingleThreadExecutor();
    Future<DatasetWatchResult> f = execService.submit(new Callable<DatasetWatchResult>() {
      @Override
      public DatasetWatchResult call() throws Exception {
        return service.take();
      }
    });
    f.get(500L, TimeUnit.MILLISECONDS);
  }

  @Test(timeout = 5000)
  public void testTake_updatedDataset() throws InterruptedException, IOException {
    Path tmpDir = fs.getPath("/tmp");
    Files.createDirectory(tmpDir);
    Path[] images = new Path[5];
    for (int i = 0; i < images.length; i++) {
      images[i] = tmpDir.resolve(format("image_%04d.cbf", i + 1));
      Files.createFile(images[i]);
    }
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(5));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetWatchService(c);
    service.register(tmpDir);
    service.start();
    DatasetWatchResult r = service.take();
    Dataset d = r.dataset();
    assertEquals(d.directory(), tmpDir);
    Files.setLastModifiedTime(images[0],
        FileTime.fromMillis(Files.getLastModifiedTime(images[0]).toMillis() + 1L));
    r = service.take();
    d = r.dataset();
    assertEquals(d.directory(), tmpDir);
  }

  @Test(expected = TimeoutException.class)
  public void testTake_updatedDatasetWithoutMetadata()
      throws InterruptedException, IOException, TimeoutException, ExecutionException {
    Path tmpDir = fs.getPath("/tmp");
    Files.createDirectory(tmpDir);
    Path[] images = new Path[5];
    for (int i = 0; i < images.length; i++) {
      images[i] = tmpDir.resolve(format("image_%04d.cbf", i + 1));
      Files.createFile(images[i]);
    }
    AapConfigurationBuilder b = new AapConfigurationBuilder();
    b.fsPollEnabled(Boolean.TRUE);
    b.fsPollPeriod(TimeDuration.milliseconds(100L));
    b.fsPollMetadataEnabled(Boolean.FALSE);
    b.datasetSettleTime(TimeDuration.milliseconds(1L));
    b.datasetMinNumImages(Integer.valueOf(5));
    AapConfiguration c = b.toAapConfiguration();
    service = newDatasetWatchService(c);
    service.register(tmpDir);
    service.start();
    DatasetWatchResult r = service.take();
    Dataset d = r.dataset();
    assertEquals(d.directory(), tmpDir);
    Files.setLastModifiedTime(images[0],
        FileTime.fromMillis(Files.getLastModifiedTime(images[0]).toMillis() + 1L));
    execService = Executors.newSingleThreadExecutor();
    Future<DatasetWatchResult> f = execService.submit(new Callable<DatasetWatchResult>() {
      @Override
      public DatasetWatchResult call() throws Exception {
        return service.take();
      }
    });
    f.get(500L, TimeUnit.MILLISECONDS);
  }

  private DatasetWatchService newDatasetWatchService(AapConfiguration c) throws IOException {
    return new DatasetWatchService(c, fs, new DatasetWatchServiceExceptionHandler() {
      @Override
      public void exception(Throwable t) {
        t.printStackTrace();
      }
    }, new DatasetWatchServiceInfoHandler() {
      @Override
      public void pollTime(long t) {
        /* IGNORED */
      }
    });
  }

  private static String format(String format, Object... args) {
    return String.format(Locale.US, format, args);
  }
}
