/*
 * Copyright (c) 2014, 2015 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertTrue;

import java.math.BigDecimal;
import java.util.Arrays;

import org.junit.Test;

/**
 * Tests {@link UnitCell}.
 */
public class UnitCellTest {
  public UnitCellTest() {
    super();
  }

  @Test
  public void testEquals_equal() {
    /* @formatter:off */
    String[] u1 = new String[] {"71.55", "72", "53.31", "73", "78", "75"};
    /* @formatter:on */
    assertEquals(newUnitCell(u1), newUnitCell(u1));
  }

  @Test
  public void testEquals_lengthNotEqual() {
    /* @formatter:off */
    String[] u1 = new String[] {"71.55", "72", "53.31", "73", "78", "75"};
    /* @formatter:on */
    String[] u2 = Arrays.copyOf(u1, u1.length);
    u2[2] = "54.31";
    assertNotEquals(u1, u2);
  }

  @Test
  public void testEquals_angleNotEqual() {
    /* @formatter:off */
    String[] u1 = new String[] {"71.55", "72", "53.31", "73", "78", "75"};
    /* @formatter:on */
    String[] u2 = Arrays.copyOf(u1, u1.length);
    u2[5] = "76";
    assertNotEquals(u1, u2);
  }

  @Test
  public void testEquals_lengthScaleNotEqual() {
    /* @formatter:off */
    String[] u1 = new String[] {"71.55", "72", "53.31", "73", "78", "75"};
    /* @formatter:on */
    String[] u2 = Arrays.copyOf(u1, u1.length);
    u2[0] = u1[0] + "0";
    assertNotEquals(u1, u2);
  }

  @Test
  public void testEquals_angleScaleNotEqual() {
    /* @formatter:off */
    String[] u1 = new String[] {"71.55", "72", "53.31", "73", "78", "75"};
    /* @formatter:on */
    String[] u2 = Arrays.copyOf(u1, u1.length);
    u2[3] = u1[3] + ".0";
    assertNotEquals(u1, u2);
  }

  @Test
  public void testEqualsIgnoreScale_equal() {
    /* @formatter:off */
    String[] u1 = new String[] {"71.55", "72", "53.31", "73", "78", "75"};
    /* @formatter:on */
    assertTrue(newUnitCell(u1).equalsIgnoreScale(newUnitCell(u1)));
  }

  @Test
  public void testEqualsIgnoreScale_lengthNotEqual() {
    /* @formatter:off */
    String[] u1 = new String[] {"71.55", "72", "53.31", "73", "78", "75"};
    /* @formatter:on */
    String[] u2 = Arrays.copyOf(u1, u1.length);
    u2[2] = "54.31";
    assertFalse(newUnitCell(u1).equalsIgnoreScale(newUnitCell(u2)));
  }

  @Test
  public void testEqualsIgnoreScale_angleNotEqual() {
    /* @formatter:off */
    String[] u1 = new String[] {"71.55", "72", "53.31", "73", "78", "75"};
    /* @formatter:on */
    String[] u2 = Arrays.copyOf(u1, u1.length);
    u2[5] = "76";
    assertFalse(newUnitCell(u1).equalsIgnoreScale(newUnitCell(u2)));
  }

  @Test
  public void testEqualsIgnoreScale_lengthScaleNotEqual() {
    /* @formatter:off */
    String[] u1 = new String[] {"71.55", "72", "53.31", "73", "78", "75"};
    /* @formatter:on */
    String[] u2 = Arrays.copyOf(u1, u1.length);
    u2[0] = u1[0] + "0";
    assertTrue(newUnitCell(u1).equalsIgnoreScale(newUnitCell(u2)));
  }

  @Test
  public void testEqualsIgnoreScale_angleScaleNotEqual() {
    /* @formatter:off */
    String[] u1 = new String[] {"71.55", "72", "53.31", "73", "78", "75"};
    /* @formatter:on */
    String[] u2 = Arrays.copyOf(u1, u1.length);
    u2[3] = u1[3] + ".0";
    assertTrue(newUnitCell(u1).equalsIgnoreScale(newUnitCell(u2)));
  }

  @Test
  public void testEqualsIgnoreScale_null() {
    /* @formatter:off */
    String[] u1 = new String[] {"71.55", "72", "53.31", "73", "78", "75"};
    /* @formatter:on */
    assertFalse(newUnitCell(u1).equalsIgnoreScale(null));
  }

  private static UnitCell newUnitCell(String[] lengthsAndAngles) {
    BigDecimal[] bigLengthsAndAngles = new BigDecimal[lengthsAndAngles.length];
    for (int i = 0; i < lengthsAndAngles.length; i++) {
      bigLengthsAndAngles[i] = new BigDecimal(lengthsAndAngles[i]);
    }
    return new UnitCell(bigLengthsAndAngles);
  }
}
