/*
 * Copyright (c) 2015 J. Lewis Muir <jlmuir@imca-cat.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.imca_cat.aaproc;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.math.BigDecimal;

import org.junit.Test;

/**
 * Tests {@link XdsIntegrateLogParser}.
 */
public class XdsIntegrateLogParserTest {
  public XdsIntegrateLogParserTest() {
    super();
  }

  @Test
  public void testParse_normal() throws IOException {
    XdsIntegrateResult result = null;
    try (BufferedReader reader = newNormalLog()) {
      result = (new XdsIntegrateLogParser(reader)).parse();
    }
    assertFalse(result.hasException());
    assertFalse(result.failedToRefine());
    assertEquals(new BigDecimal("250.003"), result.detectorDistance());
    assertEquals(new BigDecimal("250.20"), result.refinedDetectorDistance());
    assertEquals(new BigDecimal("0.197"), result.refinedDetectorDistanceDelta());
  }

  @Test
  public void testParse_failedToRefine() throws IOException {
    XdsIntegrateResult result = null;
    try (BufferedReader reader = newFailedToRefineLog()) {
      result = (new XdsIntegrateLogParser(reader)).parse();
    }
    assertFalse(result.hasException());
    assertTrue(result.failedToRefine());
    assertEquals(new BigDecimal("250.003"), result.detectorDistance());
    assertNull(result.refinedDetectorDistance());
    assertNull(result.refinedDetectorDistanceDelta());
  }

  @Test
  public void testParse_preserveFirstLineOfNextSection() throws IOException {
    XdsIntegrateResult result = null;
    try (BufferedReader reader = newLogNeedingPreservationOfFirstLineOfNextSection()) {
      result = (new XdsIntegrateLogParser(reader)).parse();
    }
    assertFalse(result.hasException());
    assertTrue(result.failedToRefine());
    assertEquals(new BigDecimal("250.003"), result.detectorDistance());
    assertNull(result.refinedDetectorDistance());
    assertNull(result.refinedDetectorDistanceDelta());
  }

  @Test
  public void testParse_empty() throws IOException {
    XdsIntegrateResult result = null;
    try (BufferedReader reader = newEmptyLog()) {
      result = (new XdsIntegrateLogParser(reader)).parse();
    }
    assertFalse(result.hasException());
    assertFalse(result.failedToRefine());
    assertNull(result.detectorDistance());
    assertNull(result.refinedDetectorDistance());
    assertNull(result.refinedDetectorDistanceDelta());
  }

  @Test
  public void testParse_partial() throws IOException {
    XdsIntegrateResult result = null;
    try (BufferedReader reader = newPartialLog()) {
      result = (new XdsIntegrateLogParser(reader)).parse();
    }
    assertFalse(result.hasException());
    assertFalse(result.failedToRefine());
    assertEquals(new BigDecimal("250.003"), result.detectorDistance());
    assertNull(result.refinedDetectorDistance());
    assertNull(result.refinedDetectorDistanceDelta());
  }

  private static BufferedReader newNormalLog() {
    /* @formatter:off */
    String[] output = new String[] {
        "",
        " ***** XDS ***** (VERSION January 10, 2014  BUILT=20140317)   9-Sep-2014",
        "",
        " ***** INTEGRATE ***** (VERSION January 10, 2014)   9-Sep-2014",
        "",
        " NAME_TEMPLATE_OF_DATA_FRAMES=Blue52_????.cbf CBF",
        " DATA_RANGE=       1     720",
        " MINIMUM_VALID_PIXEL_VALUE=     0    OVERLOAD=    998776",
        " BACKGROUND_PIXEL=    6.00      SIGNAL_PIXEL=    3.00",
        " MAXIMUM_ERROR_OF_SPOT_POSITION=   2.0",
        " MAXIMUM_ERROR_OF_SPINDLE_POSITION=   2.0",
        " MINPK=  75.00000",
        " PROFILE_FITTING= TRUE ",
        " SPACE_GROUP_NUMBER=    1",
        " UNIT_CELL_CONSTANTS=    57.83    57.88   150.17  89.928  90.024  89.980",
        " ROTATION_AXIS=  0.99999  0.00371 -0.00192",
        " OSCILLATION_RANGE=  0.250000 DEGREES",
        " STARTING_ANGLE=   0.000 STARTING_FRAME=       1",
        " X-RAY_WAVELENGTH=   1.00000 ANGSTROM",
        " INCIDENT_BEAM_DIRECTION= -0.005167  0.001551  0.999986",
        " NUMBER OF DETECTOR SEGMENTS   1",
        " NX=  2463 NY=  2527 QX= 0.17200 QY= 0.17200",
        " ORGX=   1260.44  ORGY=   1286.40",
        " DETECTOR_DISTANCE=   250.003",
        " DIRECTION_OF_DETECTOR_X-AXIS=   1.00000   0.00000   0.00000",
        " DIRECTION_OF_DETECTOR_Y-AXIS=   0.00000   1.00000   0.00000",
        " SEGMENT=    1 2463    1 2527  ! x1,x2,y1,y2 (pixel units)",
        "    SEGMENT_ORGX=      0.00   SEGMENT_ORGY=      0.00",
        "    SEGMENT_DISTANCE=     0.00",
        "    DIRECTION_OF_SEGMENT_X-AXIS=  1.000000  0.000000  0.000000",
        "    DIRECTION_OF_SEGMENT_Y-AXIS=  0.000000  1.000000  0.000000",
        " RESOLUTION RANGE RECORDED BY DETECTOR (ANGSTROM)   915.893     1.165",
        " NUMBER OF TRUSTED DETECTOR PIXELS    5622878",
        " MEAN CONTENTS OF TRUSTED PIXELS IN BACKGROUND TABLE    583.534",
        " MEAN VALUE OF NON-XRAY BACKGROUND (OFFSET)=    0.00",
        " NUMBER OF X-RAY COUNTS EQUIVALENT TO PIXEL CONTENTS IN A DATA IMAGE",
        " (PIXEL VALUE - OFFSET)/GAIN ,    WITH GAIN=    0.99",
        "",
        "",
        "",
        " ******************************************************************************",
        "                     PROCESSING OF IMAGES        1 ...      20",
        " ******************************************************************************",
        "",
        " USING  20 PROCESSORS",
        "",
        " *** DEFINITION OF SYMBOLS ***",
        " IER     = ERROR CODE AFTER ACCESSING DATA IMAGE ",
        "              0: NO ERROR",
        "             -1: CANNOT OPEN OR READ IMAGE FILE",
        "             -3: WRONG DATA FORMAT",
        " SCALE   = SCALING FACTOR FOR THIS DATA IMAGE",
        " NBKG    = NUMBER OF BACKGROUND PIXELS ON DATA IMAGE",
        " NOVL    = NUMBER OF OVERLOADED REFLECTIONS ON DATA IMAGE",
        " NEWALD  = NUMBER OF REFLECTIONS CLOSE TO THE EWALD SPHERE",
        " NSTRONG = NUMBER OF STRONG REFLECTIONS ON DATA IMAGE",
        " NREJ    = NUMBER OF UNEXPECTED REFLECTIONS",
        " SIGMAB  = BEAM_DIVERGENCE_E.S.D.=SIGMAB ",
        " SIGMAR  = REFLECTING_RANGE_E.S.D.=SIGMAR (MOSAICITY)",
        "",
        " IMAGE IER  SCALE     NBKG NOVL NEWALD NSTRONG  NREJ  SIGMAB  SIGMAR",
        "     1   0  0.964  5555332    0   6237     807    42  0.0391  0.1066",
        "     2   0  0.967  5551684    0   6314     843    38  0.0388  0.1077",
        "     3   0  0.969  5551502    0   6381     863    35  0.0384  0.1140",
        "     4   0  0.968  5551775    0   6443     858    34  0.0387  0.1209",
        "     5   0  0.972  5548814    0   6502     890    40  0.0383  0.1093",
        "     6   0  0.971  5549806    0   6507     881    42  0.0387  0.1152",
        "     7   0  0.970  5552620    0   6429     850    38  0.0386  0.1136",
        "     8   0  0.971  5551573    0   6317     857    33  0.0387  0.1151",
        "     9   0  0.973  5552442    0   6267     848    37  0.0387  0.1170",
        "    10   0  0.974  5550786    0   6287     874    29  0.0384  0.1216",
        "    11   0  0.979  5549826    0   6358     879    35  0.0392  0.1169",
        "    12   0  0.976  5550265    0   6411     880    36  0.0389  0.1171",
        "    13   0  0.980  5550658    0   6381     880    41  0.0387  0.1085",
        "    14   0  0.979  5552784    0   6446     857    35  0.0384  0.1190",
        "    15   0  0.980  5549219    0   6364     893    38  0.0377  0.1123",
        "    16   0  0.983  5552887    0   6324     833    46  0.0391  0.1148",
        "    17   0  0.979  5555253    0   6314     810    32  0.0387  0.1299",
        "    18   0  0.982  5553733    0   6209     842    37  0.0388  0.1183",
        "    19   0  0.986  5554680    0   6287     831    70  0.0385  0.1042",
        "    20   0  0.984  5552608    0   6274     856    41  0.0390  0.1186",
        "",
        "  7161 OUT OF   7359 REFLECTIONS ACCEPTED FOR REFINEMENT",
        " REFINED PARAMETERS:   DISTANCE BEAM ORIENTATION CELL                        ",
        " STANDARD DEVIATION OF SPOT    POSITION (PIXELS)     0.42",
        " STANDARD DEVIATION OF SPINDLE POSITION (DEGREES)    0.05",
        " SPACE GROUP NUMBER      1",
        " UNIT CELL PARAMETERS     57.861    57.906   150.257  90.033  89.982  90.067",
        " REC. CELL PARAMETERS   0.017283  0.017269  0.006655  89.967  90.018  89.933",
        " COORDINATES OF UNIT CELL A-AXIS   -52.103   -18.401    17.162",
        " COORDINATES OF UNIT CELL B-AXIS    -6.427   -28.316   -50.099",
        " COORDINATES OF UNIT CELL C-AXIS    63.106  -121.984    60.949",
        " CRYSTAL ROTATION OFF FROM INITIAL ORIENTATION    -0.022    -0.004     0.018",
        " shown as x,y,z components of rotation axis X angle (degrees)",
        " CRYSTAL MOSAICITY (DEGREES)     0.115",
        " LAB COORDINATES OF ROTATION AXIS  0.999991  0.003708 -0.001924",
        " DIRECT BEAM COORDINATES (REC. ANGSTROEM)  -0.005108  0.002000  0.999985",
        " DETECTOR COORDINATES (PIXELS) OF DIRECT BEAM    1252.95   1288.58",
        " DETECTOR ORIGIN (PIXELS) AT                     1260.38   1285.67",
        " CRYSTAL TO DETECTOR DISTANCE (mm)       250.40",
        " LAB COORDINATES OF DETECTOR X-AXIS  1.000000  0.000000  0.000000",
        " LAB COORDINATES OF DETECTOR Y-AXIS  0.000000  1.000000  0.000000",
        "",
        "",
        " STANDARD DEVIATIONS OF BEAM DIVERGENCE AND REFLECTING RANGE OBTAINED",
        " FROM    5276 REFLECTION PROFILES AT 9 POSITIONS ON THE DETECTOR SURFACE.",
        " POSITION NUMBER        1      2      3      4      5      6      7      8      9",
        " X-COORDINATE (pixel) 1231.0 2148.9 1880.1 1231.0  582.0  313.1  581.9 1231.0 1880.0",
        " Y-COORDINATE (pixel) 1263.0 1263.0 1928.9 2204.8 1928.9 1263.0  597.1  321.2  597.0",
        " NUMBER                  918    421    598    740    642    359    444    580    561",
        " SIGMAB (degree)       0.043  0.049  0.047  0.042  0.041  0.048  0.051  0.049  0.048",
        " SIGMAR (degree)       0.090  0.082  0.090  0.087  0.084  0.090  0.092  0.089  0.082",
        "",
        "",
        "  26929 REFLECTION INTENSITIES INTEGRATED",
        "",
        "",
        " ******************************************************************************",
        "                     PROCESSING OF IMAGES       21 ...      40",
        " ******************************************************************************",
        "",
        " USING  20 PROCESSORS",
        "",
        " *** DEFINITION OF SYMBOLS ***",
        " IER     = ERROR CODE AFTER ACCESSING DATA IMAGE ",
        "              0: NO ERROR",
        "             -1: CANNOT OPEN OR READ IMAGE FILE",
        "             -3: WRONG DATA FORMAT",
        " SCALE   = SCALING FACTOR FOR THIS DATA IMAGE",
        " NBKG    = NUMBER OF BACKGROUND PIXELS ON DATA IMAGE",
        " NOVL    = NUMBER OF OVERLOADED REFLECTIONS ON DATA IMAGE",
        " NEWALD  = NUMBER OF REFLECTIONS CLOSE TO THE EWALD SPHERE",
        " NSTRONG = NUMBER OF STRONG REFLECTIONS ON DATA IMAGE",
        " NREJ    = NUMBER OF UNEXPECTED REFLECTIONS",
        " SIGMAB  = BEAM_DIVERGENCE_E.S.D.=SIGMAB ",
        " SIGMAR  = REFLECTING_RANGE_E.S.D.=SIGMAR (MOSAICITY)",
        "",
        " IMAGE IER  SCALE     NBKG NOVL NEWALD NSTRONG  NREJ  SIGMAB  SIGMAR",
        "    21   0  0.986  5551336    0   6823     990    54  0.0385  0.1185",
        "    22   0  0.981  5550188    0   6941    1003    31  0.0397  0.1300",
        "    23   0  0.984  5550595    0   7004    1013    36  0.0383  0.1317",
        "    24   0  0.984  5549364    0   6969    1012    45  0.0397  0.1201",
        "    25   0  0.986  5548048    0   6926    1046    46  0.0392  0.1212",
        "    26   0  0.991  5553456    0   6927     974    40  0.0384  0.1311",
        "    27   0  0.993  5552463    0   6861     980    41  0.0385  0.1268",
        "    28   0  0.994  5550521    0   6731    1007    59  0.0388  0.1155",
        "    29   0  1.000  5551758    0   6795     998    36  0.0388  0.1432",
        "    30   0  0.996  5553418    0   6807     969    45  0.0389  0.1356",
        "    31   0  1.000  5552511    0   6801     984    54  0.0388  0.1272",
        "    32   0  1.003  5553179    0   6832     983    49  0.0387  0.1273",
        "    33   0  1.002  5552769    0   6965     975    37  0.0394  0.1291",
        "    34   0  1.000  5548725    0   6930    1037    46  0.0400  0.1243",
        "    35   0  1.000  5547518    0   6887    1051    45  0.0394  0.1231",
        "    36   0  0.999  5547207    0   6874    1063    41  0.0392  0.1263",
        "    37   0  0.996  5550457    0   6864    1008    43  0.0389  0.1331",
        "    38   0  1.000  5553686    0   6801     960    48  0.0389  0.1263",
        "    39   0  0.997  5551398    0   6748     996    45  0.0388  0.1302",
        "    40   0  1.002  5552698    0   6802     973    47  0.0390  0.1244",
        "",
        "  7856 OUT OF   8082 REFLECTIONS ACCEPTED FOR REFINEMENT",
        " REFINED PARAMETERS:   DISTANCE BEAM ORIENTATION CELL                        ",
        " STANDARD DEVIATION OF SPOT    POSITION (PIXELS)     0.46",
        " STANDARD DEVIATION OF SPINDLE POSITION (DEGREES)    0.05",
        " SPACE GROUP NUMBER      1",
        " UNIT CELL PARAMETERS     57.815    57.840   150.149  90.019  89.986  90.064",
        " REC. CELL PARAMETERS   0.017297  0.017289  0.006660  89.981  90.014  89.936",
        " COORDINATES OF UNIT CELL A-AXIS   -52.059   -18.382    17.161",
        " COORDINATES OF UNIT CELL B-AXIS    -6.438   -28.283   -50.041",
        " COORDINATES OF UNIT CELL C-AXIS    63.068  -121.920    60.851",
        " CRYSTAL ROTATION OFF FROM INITIAL ORIENTATION    -0.013    -0.003     0.029",
        " shown as x,y,z components of rotation axis X angle (degrees)",
        " CRYSTAL MOSAICITY (DEGREES)     0.127",
        " LAB COORDINATES OF ROTATION AXIS  0.999991  0.003708 -0.001924",
        " DIRECT BEAM COORDINATES (REC. ANGSTROEM)  -0.004964  0.001753  0.999986",
        " DETECTOR COORDINATES (PIXELS) OF DIRECT BEAM    1252.94   1288.62",
        " DETECTOR ORIGIN (PIXELS) AT                     1260.16   1286.07",
        " CRYSTAL TO DETECTOR DISTANCE (mm)       250.20",
        " LAB COORDINATES OF DETECTOR X-AXIS  1.000000  0.000000  0.000000",
        " LAB COORDINATES OF DETECTOR Y-AXIS  0.000000  1.000000  0.000000",
        "",
        "",
        " STANDARD DEVIATIONS OF BEAM DIVERGENCE AND REFLECTING RANGE OBTAINED",
        " FROM    5551 REFLECTION PROFILES AT 9 POSITIONS ON THE DETECTOR SURFACE.",
        " POSITION NUMBER        1      2      3      4      5      6      7      8      9",
        " X-COORDINATE (pixel) 1231.0 2148.9 1880.1 1231.0  582.0  313.1  581.9 1231.0 1880.0",
        " Y-COORDINATE (pixel) 1263.0 1263.0 1928.9 2204.8 1928.9 1263.0  597.1  321.2  597.0",
        " NUMBER                  915    409    589    750    653    381    527    682    626",
        " SIGMAB (degree)       0.043  0.049  0.047  0.043  0.042  0.049  0.052  0.050  0.049",
        " SIGMAR (degree)       0.095  0.088  0.098  0.092  0.088  0.097  0.099  0.093  0.086",
        "",
        "",
        "  27496 REFLECTION INTENSITIES INTEGRATED"
    };
    /* @formatter:on */
    StringBuilder result = new StringBuilder();
    for (String each : output) {
      result.append(each);
      result.append("\n");
    }
    return new BufferedReader(new StringReader(result.toString()));
  }

  private static BufferedReader newFailedToRefineLog() {
    /* @formatter:off */
    String[] output = new String[] {
        "",
        " ***** XDS ***** (VERSION January 10, 2014  BUILT=20140317)   9-Sep-2014",
        "",
        " ***** INTEGRATE ***** (VERSION January 10, 2014)   9-Sep-2014",
        "",
        " NAME_TEMPLATE_OF_DATA_FRAMES=Blue52_????.cbf CBF",
        " DATA_RANGE=       1     720",
        " MINIMUM_VALID_PIXEL_VALUE=     0    OVERLOAD=    998776",
        " BACKGROUND_PIXEL=    6.00      SIGNAL_PIXEL=    3.00",
        " MAXIMUM_ERROR_OF_SPOT_POSITION=   2.0",
        " MAXIMUM_ERROR_OF_SPINDLE_POSITION=   2.0",
        " MINPK=  75.00000",
        " PROFILE_FITTING= TRUE ",
        " SPACE_GROUP_NUMBER=    1",
        " UNIT_CELL_CONSTANTS=    57.83    57.88   150.17  89.928  90.024  89.980",
        " ROTATION_AXIS=  0.99999  0.00371 -0.00192",
        " OSCILLATION_RANGE=  0.250000 DEGREES",
        " STARTING_ANGLE=   0.000 STARTING_FRAME=       1",
        " X-RAY_WAVELENGTH=   1.00000 ANGSTROM",
        " INCIDENT_BEAM_DIRECTION= -0.005167  0.001551  0.999986",
        " NUMBER OF DETECTOR SEGMENTS   1",
        " NX=  2463 NY=  2527 QX= 0.17200 QY= 0.17200",
        " ORGX=   1260.44  ORGY=   1286.40",
        " DETECTOR_DISTANCE=   250.003",
        " DIRECTION_OF_DETECTOR_X-AXIS=   1.00000   0.00000   0.00000",
        " DIRECTION_OF_DETECTOR_Y-AXIS=   0.00000   1.00000   0.00000",
        " SEGMENT=    1 2463    1 2527  ! x1,x2,y1,y2 (pixel units)",
        "    SEGMENT_ORGX=      0.00   SEGMENT_ORGY=      0.00",
        "    SEGMENT_DISTANCE=     0.00",
        "    DIRECTION_OF_SEGMENT_X-AXIS=  1.000000  0.000000  0.000000",
        "    DIRECTION_OF_SEGMENT_Y-AXIS=  0.000000  1.000000  0.000000",
        " RESOLUTION RANGE RECORDED BY DETECTOR (ANGSTROM)   915.893     1.165",
        " NUMBER OF TRUSTED DETECTOR PIXELS    5622878",
        " MEAN CONTENTS OF TRUSTED PIXELS IN BACKGROUND TABLE    583.534",
        " MEAN VALUE OF NON-XRAY BACKGROUND (OFFSET)=    0.00",
        " NUMBER OF X-RAY COUNTS EQUIVALENT TO PIXEL CONTENTS IN A DATA IMAGE",
        " (PIXEL VALUE - OFFSET)/GAIN ,    WITH GAIN=    0.99",
        "",
        "",
        "",
        " ******************************************************************************",
        "                     PROCESSING OF IMAGES        1 ...      20",
        " ******************************************************************************",
        "",
        " USING  20 PROCESSORS",
        "",
        " *** DEFINITION OF SYMBOLS ***",
        " IER     = ERROR CODE AFTER ACCESSING DATA IMAGE ",
        "              0: NO ERROR",
        "             -1: CANNOT OPEN OR READ IMAGE FILE",
        "             -3: WRONG DATA FORMAT",
        " SCALE   = SCALING FACTOR FOR THIS DATA IMAGE",
        " NBKG    = NUMBER OF BACKGROUND PIXELS ON DATA IMAGE",
        " NOVL    = NUMBER OF OVERLOADED REFLECTIONS ON DATA IMAGE",
        " NEWALD  = NUMBER OF REFLECTIONS CLOSE TO THE EWALD SPHERE",
        " NSTRONG = NUMBER OF STRONG REFLECTIONS ON DATA IMAGE",
        " NREJ    = NUMBER OF UNEXPECTED REFLECTIONS",
        " SIGMAB  = BEAM_DIVERGENCE_E.S.D.=SIGMAB ",
        " SIGMAR  = REFLECTING_RANGE_E.S.D.=SIGMAR (MOSAICITY)",
        "",
        " IMAGE IER  SCALE     NBKG NOVL NEWALD NSTRONG  NREJ  SIGMAB  SIGMAR",
        "     1   0  0.964  5555332    0   6237     807    42  0.0391  0.1066",
        "     2   0  0.967  5551684    0   6314     843    38  0.0388  0.1077",
        "     3   0  0.969  5551502    0   6381     863    35  0.0384  0.1140",
        "     4   0  0.968  5551775    0   6443     858    34  0.0387  0.1209",
        "     5   0  0.972  5548814    0   6502     890    40  0.0383  0.1093",
        "     6   0  0.971  5549806    0   6507     881    42  0.0387  0.1152",
        "     7   0  0.970  5552620    0   6429     850    38  0.0386  0.1136",
        "     8   0  0.971  5551573    0   6317     857    33  0.0387  0.1151",
        "     9   0  0.973  5552442    0   6267     848    37  0.0387  0.1170",
        "    10   0  0.974  5550786    0   6287     874    29  0.0384  0.1216",
        "    11   0  0.979  5549826    0   6358     879    35  0.0392  0.1169",
        "    12   0  0.976  5550265    0   6411     880    36  0.0389  0.1171",
        "    13   0  0.980  5550658    0   6381     880    41  0.0387  0.1085",
        "    14   0  0.979  5552784    0   6446     857    35  0.0384  0.1190",
        "    15   0  0.980  5549219    0   6364     893    38  0.0377  0.1123",
        "    16   0  0.983  5552887    0   6324     833    46  0.0391  0.1148",
        "    17   0  0.979  5555253    0   6314     810    32  0.0387  0.1299",
        "    18   0  0.982  5553733    0   6209     842    37  0.0388  0.1183",
        "    19   0  0.986  5554680    0   6287     831    70  0.0385  0.1042",
        "    20   0  0.984  5552608    0   6274     856    41  0.0390  0.1186",
        "",
        "  7161 OUT OF   7359 REFLECTIONS ACCEPTED FOR REFINEMENT",
        " REFINED PARAMETERS:   DISTANCE BEAM ORIENTATION CELL                        ",
        " STANDARD DEVIATION OF SPOT    POSITION (PIXELS)     0.42",
        " STANDARD DEVIATION OF SPINDLE POSITION (DEGREES)    0.05",
        " SPACE GROUP NUMBER      1",
        " UNIT CELL PARAMETERS     57.861    57.906   150.257  90.033  89.982  90.067",
        " REC. CELL PARAMETERS   0.017283  0.017269  0.006655  89.967  90.018  89.933",
        " COORDINATES OF UNIT CELL A-AXIS   -52.103   -18.401    17.162",
        " COORDINATES OF UNIT CELL B-AXIS    -6.427   -28.316   -50.099",
        " COORDINATES OF UNIT CELL C-AXIS    63.106  -121.984    60.949",
        " CRYSTAL ROTATION OFF FROM INITIAL ORIENTATION    -0.022    -0.004     0.018",
        " shown as x,y,z components of rotation axis X angle (degrees)",
        " CRYSTAL MOSAICITY (DEGREES)     0.115",
        " LAB COORDINATES OF ROTATION AXIS  0.999991  0.003708 -0.001924",
        " DIRECT BEAM COORDINATES (REC. ANGSTROEM)  -0.005108  0.002000  0.999985",
        " DETECTOR COORDINATES (PIXELS) OF DIRECT BEAM    1252.95   1288.58",
        " DETECTOR ORIGIN (PIXELS) AT                     1260.38   1285.67",
        " CRYSTAL TO DETECTOR DISTANCE (mm)       250.40",
        " LAB COORDINATES OF DETECTOR X-AXIS  1.000000  0.000000  0.000000",
        " LAB COORDINATES OF DETECTOR Y-AXIS  0.000000  1.000000  0.000000",
        "",
        "",
        " STANDARD DEVIATIONS OF BEAM DIVERGENCE AND REFLECTING RANGE OBTAINED",
        " FROM    5276 REFLECTION PROFILES AT 9 POSITIONS ON THE DETECTOR SURFACE.",
        " POSITION NUMBER        1      2      3      4      5      6      7      8      9",
        " X-COORDINATE (pixel) 1231.0 2148.9 1880.1 1231.0  582.0  313.1  581.9 1231.0 1880.0",
        " Y-COORDINATE (pixel) 1263.0 1263.0 1928.9 2204.8 1928.9 1263.0  597.1  321.2  597.0",
        " NUMBER                  918    421    598    740    642    359    444    580    561",
        " SIGMAB (degree)       0.043  0.049  0.047  0.042  0.041  0.048  0.051  0.049  0.048",
        " SIGMAR (degree)       0.090  0.082  0.090  0.087  0.084  0.090  0.092  0.089  0.082",
        "",
        "",
        "  26929 REFLECTION INTENSITIES INTEGRATED",
        "",
        "",
        " ******************************************************************************",
        "                     PROCESSING OF IMAGES       21 ...      40",
        " ******************************************************************************",
        "",
        " USING  20 PROCESSORS",
        "",
        " *** DEFINITION OF SYMBOLS ***",
        " IER     = ERROR CODE AFTER ACCESSING DATA IMAGE ",
        "              0: NO ERROR",
        "             -1: CANNOT OPEN OR READ IMAGE FILE",
        "             -3: WRONG DATA FORMAT",
        " SCALE   = SCALING FACTOR FOR THIS DATA IMAGE",
        " NBKG    = NUMBER OF BACKGROUND PIXELS ON DATA IMAGE",
        " NOVL    = NUMBER OF OVERLOADED REFLECTIONS ON DATA IMAGE",
        " NEWALD  = NUMBER OF REFLECTIONS CLOSE TO THE EWALD SPHERE",
        " NSTRONG = NUMBER OF STRONG REFLECTIONS ON DATA IMAGE",
        " NREJ    = NUMBER OF UNEXPECTED REFLECTIONS",
        " SIGMAB  = BEAM_DIVERGENCE_E.S.D.=SIGMAB ",
        " SIGMAR  = REFLECTING_RANGE_E.S.D.=SIGMAR (MOSAICITY)",
        "",
        " IMAGE IER  SCALE     NBKG NOVL NEWALD NSTRONG  NREJ  SIGMAB  SIGMAR",
        "    21   0  0.986  5551336    0   6823     990    54  0.0385  0.1185",
        "    22   0  0.981  5550188    0   6941    1003    31  0.0397  0.1300",
        "    23   0  0.984  5550595    0   7004    1013    36  0.0383  0.1317",
        "    24   0  0.984  5549364    0   6969    1012    45  0.0397  0.1201",
        "    25   0  0.986  5548048    0   6926    1046    46  0.0392  0.1212",
        "    26   0  0.991  5553456    0   6927     974    40  0.0384  0.1311",
        "    27   0  0.993  5552463    0   6861     980    41  0.0385  0.1268",
        "    28   0  0.994  5550521    0   6731    1007    59  0.0388  0.1155",
        "    29   0  1.000  5551758    0   6795     998    36  0.0388  0.1432",
        "    30   0  0.996  5553418    0   6807     969    45  0.0389  0.1356",
        "    31   0  1.000  5552511    0   6801     984    54  0.0388  0.1272",
        "    32   0  1.003  5553179    0   6832     983    49  0.0387  0.1273",
        "    33   0  1.002  5552769    0   6965     975    37  0.0394  0.1291",
        "    34   0  1.000  5548725    0   6930    1037    46  0.0400  0.1243",
        "    35   0  1.000  5547518    0   6887    1051    45  0.0394  0.1231",
        "    36   0  0.999  5547207    0   6874    1063    41  0.0392  0.1263",
        "    37   0  0.996  5550457    0   6864    1008    43  0.0389  0.1331",
        "    38   0  1.000  5553686    0   6801     960    48  0.0389  0.1263",
        "    39   0  0.997  5551398    0   6748     996    45  0.0388  0.1302",
        "    40   0  1.002  5552698    0   6802     973    47  0.0390  0.1244",
        "",
        " !!! WARNING !!! UNABLE TO CARRY OUT REFINEMENT. RETURN CODE  IER=       0",
        " XDS WILL CONTINUE WITH ORIGINAL DIFFRACTION PARAMETER VALUES.",
        "",
        "",
        " STANDARD DEVIATIONS OF BEAM DIVERGENCE AND REFLECTING RANGE OBTAINED",
        " FROM    5551 REFLECTION PROFILES AT 9 POSITIONS ON THE DETECTOR SURFACE.",
        " POSITION NUMBER        1      2      3      4      5      6      7      8      9",
        " X-COORDINATE (pixel) 1231.0 2148.9 1880.1 1231.0  582.0  313.1  581.9 1231.0 1880.0",
        " Y-COORDINATE (pixel) 1263.0 1263.0 1928.9 2204.8 1928.9 1263.0  597.1  321.2  597.0",
        " NUMBER                  915    409    589    750    653    381    527    682    626",
        " SIGMAB (degree)       0.043  0.049  0.047  0.043  0.042  0.049  0.052  0.050  0.049",
        " SIGMAR (degree)       0.095  0.088  0.098  0.092  0.088  0.097  0.099  0.093  0.086",
        "",
        "",
        "  27496 REFLECTION INTENSITIES INTEGRATED"
    };
    /* @formatter:on */
    StringBuilder result = new StringBuilder();
    for (String each : output) {
      result.append(each);
      result.append("\n");
    }
    return new BufferedReader(new StringReader(result.toString()));
  }

  private static BufferedReader newLogNeedingPreservationOfFirstLineOfNextSection() {
    /* @formatter:off */
    String[] output = new String[] {
        " ***** INTEGRATE ***** (VERSION January 10, 2014)   9-Sep-2014",
        " DETECTOR_DISTANCE=   250.003",
        " !!! WARNING !!! UNABLE TO CARRY OUT REFINEMENT. RETURN CODE  IER=       0",
        " XDS WILL CONTINUE WITH ORIGINAL DIFFRACTION PARAMETER VALUES."
    };
    /* @formatter:on */
    StringBuilder result = new StringBuilder();
    for (String each : output) {
      result.append(each);
      result.append("\n");
    }
    return new BufferedReader(new StringReader(result.toString()));
  }

  private static BufferedReader newPartialLog() {
    /* @formatter:off */
    String[] output = new String[] {
        "",
        " ***** XDS ***** (VERSION January 10, 2014  BUILT=20140317)   9-Sep-2014",
        "",
        " ***** INTEGRATE ***** (VERSION January 10, 2014)   9-Sep-2014",
        "",
        " NAME_TEMPLATE_OF_DATA_FRAMES=Blue52_????.cbf CBF",
        " DATA_RANGE=       1     720",
        " MINIMUM_VALID_PIXEL_VALUE=     0    OVERLOAD=    998776",
        " BACKGROUND_PIXEL=    6.00      SIGNAL_PIXEL=    3.00",
        " MAXIMUM_ERROR_OF_SPOT_POSITION=   2.0",
        " MAXIMUM_ERROR_OF_SPINDLE_POSITION=   2.0",
        " MINPK=  75.00000",
        " PROFILE_FITTING= TRUE ",
        " SPACE_GROUP_NUMBER=    1",
        " UNIT_CELL_CONSTANTS=    57.83    57.88   150.17  89.928  90.024  89.980",
        " ROTATION_AXIS=  0.99999  0.00371 -0.00192",
        " OSCILLATION_RANGE=  0.250000 DEGREES",
        " STARTING_ANGLE=   0.000 STARTING_FRAME=       1",
        " X-RAY_WAVELENGTH=   1.00000 ANGSTROM",
        " INCIDENT_BEAM_DIRECTION= -0.005167  0.001551  0.999986",
        " NUMBER OF DETECTOR SEGMENTS   1",
        " NX=  2463 NY=  2527 QX= 0.17200 QY= 0.17200",
        " ORGX=   1260.44  ORGY=   1286.40",
        " DETECTOR_DISTANCE=   250.003",
        " DIRECTION_OF_DETECTOR_X-AXIS=   1.00000   0.00000   0.00000",
        " DIRECTION_OF_DETECTOR_Y-AXIS=   0.00000   1.00000   0.00000",
        " SEGMENT=    1 2463    1 2527  ! x1,x2,y1,y2 (pixel units)",
        "    SEGMENT_ORGX=      0.00   SEGMENT_ORGY=      0.00",
        "    SEGMENT_DISTANCE=     0.00",
        "    DIRECTION_OF_SEGMENT_X-AXIS=  1.000000  0.000000  0.000000",
        "    DIRECTION_OF_SEGMENT_Y-AXIS=  0.000000  1.000000  0.000000",
        " RESOLUTION RANGE RECORDED BY DETECTOR (ANGSTROM)   915.893     1.165",
        " NUMBER OF TRUSTED DETECTOR PIXELS    5622878",
        " MEAN CONTENTS OF TRUSTED PIXELS IN BACKGROUND TABLE    583.534",
        " MEAN VALUE OF NON-XRAY BACKGROUND (OFFSET)=    0.00",
        " NUMBER OF X-RAY COUNTS EQUIVALENT TO PIXEL CONTENTS IN A DATA IMAGE",
        " (PIXEL VALUE - OFFSET)/GAIN ,    WITH GAIN=    0.99",
        "",
        "",
        "",
        " ******************************************************************************",
        "                     PROCESSING OF IMAGES        1 ...      20",
        " ******************************************************************************",
        "",
        " USING  20 PROCESSORS",
        "",
        " *** DEFINITION OF SYMBOLS ***",
        " IER     = ERROR CODE AFTER ACCESSING DATA IMAGE ",
        "              0: NO ERROR",
        "             -1: CANNOT OPEN OR READ IMAGE FILE",
        "             -3: WRONG DATA FORMAT",
        " SCALE   = SCALING FACTOR FOR THIS DATA IMAGE",
        " NBKG    = NUMBER OF BACKGROUND PIXELS ON DATA IMAGE",
        " NOVL    = NUMBER OF OVERLOADED REFLECTIONS ON DATA IMAGE",
        " NEWALD  = NUMBER OF REFLECTIONS CLOSE TO THE EWALD SPHERE",
        " NSTRONG = NUMBER OF STRONG REFLECTIONS ON DATA IMAGE",
        " NREJ    = NUMBER OF UNEXPECTED REFLECTIONS",
        " SIGMAB  = BEAM_DIVERGENCE_E.S.D.=SIGMAB ",
        " SIGMAR  = REFLECTING_RANGE_E.S.D.=SIGMAR (MOSAICITY)",
        "",
        " IMAGE IER  SCALE     NBKG NOVL NEWALD NSTRONG  NREJ  SIGMAB  SIGMAR",
        "     1   0  0.964  5555332    0   6237     807    42  0.0391  0.1066",
        "     2   0  0.967  5551684    0   6314     843    38  0.0388  0.1077",
        "     3   0  0.969  5551502    0   6381     863    35  0.0384  0.1140"
    };
    /* @formatter:on */
    StringBuilder result = new StringBuilder();
    for (String each : output) {
      result.append(each);
      result.append("\n");
    }
    return new BufferedReader(new StringReader(result.toString()));
  }

  private static BufferedReader newEmptyLog() {
    return new BufferedReader(new StringReader(""));
  }
}
